/***********************************************************************
 *
 * linux/arch/arm/mach-ns9750/core.c
 *
 * Copyright (C) 2004-2005 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: ns9750.c,v 1.9 2006-10-31 10:40:28 tkuhlman Exp $
 * @Author: Markus Pietrek
 * @References: [1] NS9750 Hardware Reference/December 2003
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *     2005/11/29 : moved to kernel 2.6.14
 *     2005/08/29 : changed phys to virt mapping for CS2 from 0xf5000000
 *                  to 0xf6000000
 *     2005/08/29 : changed UART IRQ to IRQ_FPGA_UART
 *                           added touch and CF support
 *     2005/03/09 : inserted stuff from arch/arm/arch-ns9750/serial.h
 *                          for ST16C654 support
 *     2005/02/18 : PCI added
 *     2004/11/26 : USB Host added
 *     2004/08/20 : ported to kernel 2.6.8
 *     2004/08/17 : removed addr_*
 *     2004/07/20 : allocating BBUS on block
 *                  exporting ns_sys_clock_freq
 *     2004/06/16 : requesting resources
 ***********************************************************************/

#include <linux/init.h>		/* arch_initcall */
#include <linux/kernel.h>	/* ARRAY_SIZE */
#include <linux/ioport.h>	/* request_resource */
#include <linux/module.h>	/* EXPORT_SYMBOL */
#include <linux/platform_device.h>

#include <asm/io.h>		/* writel */

#include <asm/mach-types.h>	/* MACH_TYPE_NS9750 */
#include <asm/mach/arch.h>	/* MACHINE_START */
#include <asm/mach/map.h>	/* map_desc */

#include <asm/arch/ns9750_config.h> /* NS_SYS_CLK_FREQ */
#include <asm/arch/ns9750_sys.h> /* NS_SYS_MISC */
#include <asm/arch/ns9750_bbus.h> /* ns_gpio_lock */
#include <asm/arch/ns9750_bbus_bridge.h> /* NS_BBUS_BRIDGE_MODULE_BASE */
#include <asm/arch/ns9750_mem.h>
#include <asm/arch/ns9750_usb.h>
#include <asm/arch/irqs.h>
#include <asm/arch/a9m9xx0dev.h>

#include <asm/arch/time.h> /* ns_time_init */

#ifdef CONFIG_A9M9750DEV_BOARD
#include <linux/serial_8250.h>
#endif /*CONFIG_A9M9750DEV_BOARD*/


#include <asm/arch/a9m9xx0_ts.h>
#include <asm/arch/a9m9xx0_cf.h>

#include "devices.h"

/* Some drivers for PCI devices that originated in x86 architectures
 * depend on having their I/O space in the first 64k bytes of the
 * system's emulated x86 I/O port region.  In the NS9750, this is at
 * IO_BASE, defined in .../include/asm-arm/arch-ns9750/hardware.h
 * and currently set to 0xF0000000.  Accordingly, we position the PCI
 * I/O region in the first 64k byte region of that I/O space.
 *
 * BBus and AHB control registers are defined in multiple different
 * header files, so it is probably unwise to shift the positions of
 * BBus and AHB regions.
 */
static struct map_desc standard_io_desc[] __initdata = {
	/* virtual     physical    length      type */
	{ /* PCI I/O */
		.virtual	= 0xf0000000,
		.pfn		= __phys_to_pfn(0xa0000000),
		.length		= 0x00010000,
		.type		= MT_DEVICE
	}, { /* CS0 */
		.virtual	= 0xf4000000,
		.pfn		= __phys_to_pfn(0x40000000),
		.length		= 0x00001000,
		.type		= MT_DEVICE
	}, { /* CS2 */
		.virtual	= 0xf6000000,
		.pfn		= __phys_to_pfn(0x60000000),
		.length		= 0x000e0000,
		.type		= MT_DEVICE
	}, { /* PCI mem */
		.virtual	= 0xf8000000,
		.pfn		= __phys_to_pfn(0x80000000),
		.length		= 0x01000000,
		.type		= MT_DEVICE
	}, { /* BBus */
		.virtual	= 0xf9000000,
		.pfn		= __phys_to_pfn(0x90000000),
		.length		= 0x00700000,
		.type		= MT_DEVICE
	}, { /* AHB */
		.virtual	= 0xfa100000,
		.pfn		= __phys_to_pfn(0xa0100000),
		.length		= 0x00900000,
		.type		= MT_DEVICE
	}
};

static struct resource resources[] = {
	{
		.name	= "bbus_utility",
		.start	= NS_BBUS_UTILITY_BASE,
		.end	= (NS_BBUS_UTILITY_BASE + NS_BBUS_ARM_WAKE_UP + 3)
	},
/* could be used for serial ports, but is currently not used
	   {
		.name	= "dma_0",
		.start	= (NS_BBUS_DMA_DESCR( 0, 0 )),
		.end	= (NS_BBUS_DMA_STAT( 0, 15 ) + 3)
	},
*/
	   { /* used by USB: */
		.name	= "dma_1",
		.start	= (NS_BBUS_DMA_DESCR( 1, 0 )),
		.end	= (NS_BBUS_DMA_STAT( 1, 15 ) + 3)
	}, {
		.name = NULL
	}
};


#ifdef CONFIG_A9M9750DEV_BOARD
static struct plat_serial8250_port serial_platform_data[] = {
	{
		.iobase		= BASE_EXT_UARTA,
		.membase	= (unsigned char*)BASE_EXT_UARTA,
		.mapbase	= BASE_EXT_UARTA,
		.irq		= IRQ_FPGA_UARTA,
		.iotype		= UPIO_MEM,
		.uartclk	= 18432000,
		.regshift	= 0,
		.flags		= UPF_BOOT_AUTOCONF | UPF_SHARE_IRQ,
	},
	{
		.iobase		= BASE_EXT_UARTB,
		.membase	= (unsigned char*)BASE_EXT_UARTB,
		.mapbase	= BASE_EXT_UARTB,
		.irq		= IRQ_FPGA_UARTB,
		.iotype		= UPIO_MEM,
		.uartclk	= 18432000,
		.regshift	= 0,
		.flags		= UPF_BOOT_AUTOCONF | UPF_SHARE_IRQ,
	},
	{
		.iobase		= BASE_EXT_UARTC,
		.membase	= (unsigned char*)BASE_EXT_UARTC,
		.mapbase	= BASE_EXT_UARTC,
		.irq		= IRQ_FPGA_UARTC,
		.iotype		= UPIO_MEM,
		.uartclk	= 18432000,
		.regshift	= 0,
		.flags		= UPF_BOOT_AUTOCONF | UPF_SHARE_IRQ,
	},
	{
		.iobase		= BASE_EXT_UARTD,
		.membase	= (unsigned char*)BASE_EXT_UARTD,
		.mapbase	= BASE_EXT_UARTD,
		.irq		= IRQ_FPGA_UARTD,
		.iotype		= UPIO_MEM,
		.uartclk	= 18432000,
		.regshift	= 0,
		.flags		= UPF_BOOT_AUTOCONF | UPF_SHARE_IRQ,
	},
	{ },
};

static struct platform_device serial_device = {
	.name			= "serial8250",
	.id			= 0,
	.dev			= {
		.platform_data	= serial_platform_data,
	},
};
#endif /*CONFIG_A9M9750DEV_BOARD*/

#if defined(CONFIG_TOUCHSCREEN_A9M9XX0) || defined(CONFIG_TOUCHSCREEN_A9M9XX0_MODULE)
static struct a9m9xx0_ts_data a9m9xx0_ts_cfg __initdata = {
	.samples_ms = 20,
	.samples_avg = 1,
	.port = 0,
	.bitrate = 200000,
	.y_top = 3740,
	.y_bottom = 225,
	.x_right = 210,
	.x_left = 3870,
};
#endif /*defined(CONFIG_TOUCHSCREEN_A9M9XX0) || defined(CONFIG_TOUCHSCREEN_A9M9XX0_MODULE)*/

#if defined(CONFIG_A9M9XX0_CF) || defined(CONFIG_A9M9XX0_CF_MODULE)
static struct a9m9xx0_cf_data a9m9xx0_cf_cfg __initdata = {
	.irq = -1,
	.detect = -1,
	.reset = -1,
};
#endif /*defined(CONFIG_A9M9XX0_CF) || defined(CONFIG_A9M9XX0_CF_MODULE) */

static struct resource ns9750_usbhost_resources [] = {
        [0] = {
                .start  = (NS_USB_HCR),
                .end    = (NS_USB_HCR_PORT_SAT),
                .flags  = IORESOURCE_MEM,
        },
	[1] = {
		.start = (IRQ_USB_HOST),
		.end = (IRQ_USB_HOST),
		.flags = IORESOURCE_IRQ,
	},
};

static u64 ns9750_usbhost_dma_mask = 0xffffffffUL;

static struct platform_device ns9750_usbhost = {
        .name           = "ns9750-ohci",
        .id             = 0,
        .dev            = {
                .dma_mask = &ns9750_usbhost_dma_mask,
                .coherent_dma_mask = 0xffffffffUL,
        },
        .num_resources  = ARRAY_SIZE (ns9750_usbhost_resources),
        .resource       = ns9750_usbhost_resources,
};


static struct platform_device *usb_devices[] __initdata = {
        &ns9750_usbhost
};


static unsigned int _ns_sys_clock_freq;

spinlock_t ns_gpio_lock;	/* use it to lock all GPIO accesses */

extern void __init ns_init_irq( void );
static void __init ns_map_io( void );
void __init ns_init( void );

static void __init ns_map_io(void)
{
	/* called before ns_init */
	_ns_sys_clock_freq = NS_SYS_CLK_FREQ;

	/* @TODO */
	iotable_init( standard_io_desc, ARRAY_SIZE( standard_io_desc ) );

	printk( "NS9750 Rev. %i, running at %u MHz\n",
		(readl( NS_SYS_MISC ) >> 24) & 0xff,
		_ns_sys_clock_freq/2000000 );
}

void __init ns_init( void )
{
	int i = 0;

	/* @TODO platform_add_devices */


	spin_lock_init( &ns_gpio_lock );
#ifdef CONFIG_A9M9750DEV_BOARD
	/* start setup CS0 */
 	writel( 0x40000000, NS_SYS_CS_STATIC_BASE( 0 ) );
  	writel( 0xfffff000, NS_SYS_CS_STATIC_MASK( 0 ) );
	/* Interrupt Edge-sensitive */
  	writel( NS_SYS_EXT_INT_CTRL_PLTY, NS_SYS_EXT_INT_CTRL(2) );

	writel( NS_MEM_STAT_CFG_MW_8 | NS_MEM_STAT_CFG_PB, NS_MEM_STAT_CFG( 0 ) );
	writel(0x2 , NS_MEM_STAT_WAIT_WEN( 0 ) );
	writel( 0x2, NS_MEM_STAT_WAIT_OEN( 0 ) );
	writel( 0x6, NS_MEM_STAT_RD( 0 ) );
	writel( 0x6, NS_MEM_STAT_WR( 0 ) );
	/* end setup CS0 */

	/* start setup CS2 */
 	writel( 0x60000000, NS_SYS_CS_STATIC_BASE( 2 ) );
  	rmwl( NS_SYS_CS_STATIC_MASK(0), | 0x6FFFF001 );

	writel( NS_MEM_STAT_CFG_MW_16 | NS_MEM_STAT_CFG_PB | NS_MEM_STAT_CFG_EW, NS_MEM_STAT_CFG( 2 ) );
	writel( 0x10, NS_MEM_STAT_EXT_WAIT );
	writel(0x2 , NS_MEM_STAT_WAIT_WEN( 2 ) );
	writel( 0x2, NS_MEM_STAT_WAIT_OEN( 2 ) );
	writel( 0x6, NS_MEM_STAT_RD( 2 ) );
	writel( 0x6, NS_MEM_STAT_WR( 2 ) );
	/* end setup CS2 */

/*use IRQ2, GPIO11 */
	request_gpio( 11, 1, "serial");
	set_gpio_cfg( 11, NS_GPIO_CFG_FUNC_1 );
	platform_device_register(&serial_device);
#endif /*CONFIG_A9M9750DEV_BOARD*/
	while( resources[ i ].name != NULL ) {
		request_resource( &iomem_resource, &resources[ i++ ] );
	}
	platform_add_devices(usb_devices, 1);

#if defined(CONFIG_TOUCHSCREEN_A9M9XX0) || defined(CONFIG_TOUCHSCREEN_A9M9XX0_MODULE)
	set_a9m9xx0ts_data( &a9m9xx0_ts_cfg );
	platform_device_register( &a9m9xx0_device_ts );
#endif
#if defined(CONFIG_A9M9XX0_CF) || defined(CONFIG_A9M9XX0_CF_MODULE)
	set_a9m9xx0cf_data( &a9m9xx0_cf_cfg );
	platform_device_register( &a9m9xx0_devide_cf );
#endif
}

/***********************************************************************
 * @Function:
 * @Return:
 * @Descr:  provides system clock for drivers
 ***********************************************************************/

unsigned int ns_sys_clock_freq( void )
{
	return _ns_sys_clock_freq;
}

EXPORT_SYMBOL( ns_sys_clock_freq );


/* look in debug.S when adjusting BOOT_MEM */
MACHINE_START(A9M9750, "A9M9750")
	/* Maintainer: Jonas Dietsche <jonas_dietsche@digi.com> */
	.phys_io		= 0x90000000,
	.io_pg_offst	= ((0xf9000000) >> 18) & 0xfffc,
	.boot_params	= 0x00000100,
	.map_io		= ns_map_io,
	.init_irq		= ns_init_irq,
	.timer 		= &ns_timer,
	.init_machine	= ns_init,
MACHINE_END
