***************
*** 0 ****
--- 1,306 ----
+ /*
+  * Serial Peripheral Interface (SPI) driver for the Atmel AT91RM9200 (Thunder)
+  *
+  *  Copyright (C) SAN People (Pty) Ltd
+  *
+  * This program is free software; you can redistribute it and/or
+  * modify it under the terms of the GNU General Public License
+  * as published by the Free Software Foundation; either version
+  * 2 of the License, or (at your option) any later version.
+  */
+ 
+ #include <linux/init.h>
+ #include <linux/dma-mapping.h>
+ #include <linux/module.h>
+ #include <linux/sched.h>
+ #include <linux/completion.h>
+ #include <linux/interrupt.h>
+ #include <asm/io.h>
+ #include <asm/semaphore.h>
+ 
+ #include <asm/arch/at91rm9200_spi.h>
+ #include <asm/arch/at91rm9200_pdc.h>
+ #include <asm/arch/board.h>
+ #include <asm/arch/pio.h>
+ #include <asm/arch/at91_spi.h>
+ 
+ #undef DEBUG_SPI
+ 
+ static struct spi_local spi_dev[NR_SPI_DEVICES];	/* state of the SPI devices */
+ static int spi_enabled = 0;
+ static struct semaphore spi_lock;			/* protect access to SPI bus */
+ static int current_device = -1;				/* currently selected SPI device */
+ 
+ DECLARE_COMPLETION(transfer_complete);
+ 
+ 
+ /* ......................................................................... */
+ 
+ /*
+  * Read from a SPI register.
+  */
+ static inline unsigned long at91_spi_read(unsigned int reg)
+ {
+ 	void __iomem *spi_base = (void __iomem *)AT91_VA_BASE_SPI;
+ 	
+ 	return __raw_readl(spi_base + reg);
+ }
+ 
+ /*
+  * Write to a SPI register.
+  */
+ static inline void at91_spi_write(unsigned int reg, unsigned long value)
+ {
+ 	void __iomem *spi_base = (void __iomem *)AT91_VA_BASE_SPI;
+ 	
+ 	__raw_writel(value, spi_base + reg);
+ }
+ 
+ /* ......................................................................... */
+ 
+ /*
+  * Access and enable the SPI bus.
+  * This MUST be called before any transfers are performed.
+  */
+ void spi_access_bus(short device)
+ {
+ 	/* Ensure that requested device is valid */
+ 	if ((device < 0) || (device >= NR_SPI_DEVICES))
+ 		panic("at91_spi: spi_access_bus called with invalid device");
+ 
+ 	if (spi_enabled == 0) {
+ 		at91_sys_write(AT91_PMC_PCER, 1 << AT91_ID_SPI);/* Enable Peripheral clock */
+ 		at91_spi_write(AT91_SPI_CR, AT91_SPI_SPIEN);	/* Enable SPI */
+ #ifdef DEBUG_SPI
+ 		printk("SPI on\n");
+ #endif
+ 	}
+ 	spi_enabled++;
+ 
+ 	/* Lock the SPI bus */
+ 	down(&spi_lock);
+ 	current_device = device;
+ 
+ 	/* Enable PIO */
+ 	if (!spi_dev[device].pio_enabled) {
+ 		switch (device) {
+ 			case 0: AT91_CfgPIO_SPI_CS0();  break;
+ 			case 1: AT91_CfgPIO_SPI_CS1();  break;
+ 			case 2: AT91_CfgPIO_SPI_CS2();  break;
+ 			case 3: AT91_CfgPIO_SPI_CS3();  break;
+ 		}
+ 		spi_dev[device].pio_enabled = 1;
+ #ifdef DEBUG_SPI
+ 		printk("SPI CS%i enabled\n", device);
+ #endif
+ 	}
+ 
+ 	/* Configure SPI bus for device */
+ 	at91_spi_write(AT91_SPI_MR, AT91_SPI_MSTR | AT91_SPI_MODFDIS | (spi_dev[device].pcs << 16));
+ }
+ 
+ /*
+  * Relinquish control of the SPI bus.
+  */
+ void spi_release_bus(short device)
+ {
+ 	if (device != current_device)
+ 		panic("at91_spi: spi_release called with invalid device");
+ 
+ 	/* Release the SPI bus */
+ 	current_device = -1;
+ 	up(&spi_lock);
+ 
+ 	spi_enabled--;
+ 	if (spi_enabled == 0) {
+ 		at91_spi_write(AT91_SPI_CR, AT91_SPI_SPIDIS);	/* Disable SPI */
+ 		at91_sys_write(AT91_PMC_PCDR, 1 << AT91_ID_SPI);/* Disable Peripheral clock */
+ #ifdef DEBUG_SPI
+ 		printk("SPI off\n");
+ #endif
+ 	}
+ }
+ 
+ /*
+  * Perform a data transfer over the SPI bus
+  */
+ int spi_transfer(struct spi_transfer_list* list)
+ {
+ 	struct spi_local *device = (struct spi_local *) &spi_dev[current_device];
+ 	int tx_size;
+ 
+ 	if (!list)
+ 		panic("at91_spi: spi_transfer called with NULL transfer list");
+ 	if (current_device == -1)
+ 		panic("at91_spi: spi_transfer called without acquiring bus");
+ 
+ #ifdef DEBUG_SPI
+ 	printk("SPI transfer start [%i]\n", list->nr_transfers);
+ #endif
+ 
+ 	/* If we are in 16-bit mode, we need to modify what we pass to the PDC */
+ 	tx_size = (at91_spi_read(AT91_SPI_CSR(current_device)) & AT91_SPI_BITS_16) ? 2 : 1;
+ 
+ 	/* Store transfer list */
+ 	device->xfers = list;
+ 	list->curr = 0;
+ 
+ 	/* Assume there must be at least one transfer */
+ 	device->tx = dma_map_single(NULL, list->tx[0], list->txlen[0], DMA_TO_DEVICE);
+ 	device->rx = dma_map_single(NULL, list->rx[0], list->rxlen[0], DMA_FROM_DEVICE);
+ 
+ 	/* Program PDC registers */
+ 	at91_spi_write(AT91_PDC_TPR, device->tx);
+ 	at91_spi_write(AT91_PDC_RPR, device->rx);
+ 	at91_spi_write(AT91_PDC_TCR, list->txlen[0] / tx_size);
+ 	at91_spi_write(AT91_PDC_RCR, list->rxlen[0] / tx_size);
+ 
+ 	/* Is there a second transfer? */
+ 	if (list->nr_transfers > 1) {
+ 		device->txnext = dma_map_single(NULL, list->tx[1], list->txlen[1], DMA_TO_DEVICE);
+ 		device->rxnext = dma_map_single(NULL, list->rx[1], list->rxlen[1], DMA_FROM_DEVICE);
+ 
+ 		/* Program Next PDC registers */
+ 		at91_spi_write(AT91_PDC_TNPR, device->txnext);
+ 		at91_spi_write(AT91_PDC_RNPR, device->rxnext);
+ 		at91_spi_write(AT91_PDC_TNCR, list->txlen[1] / tx_size);
+ 		at91_spi_write(AT91_PDC_RNCR, list->rxlen[1] / tx_size);
+ 	}
+ 	else {
+ 		device->txnext = 0;
+ 		device->rxnext = 0;
+ 		at91_spi_write(AT91_PDC_TNCR, 0);
+ 		at91_spi_write(AT91_PDC_RNCR, 0);
+ 	}
+ 
+ 	// TODO: If we are doing consecutive transfers (at high speed, or
+ 	//   small buffers), then it might be worth modifying the 'Delay between
+ 	//   Consecutive Transfers' in the CSR registers.
+ 	//   This is an issue if we cannot chain the next buffer fast enough
+ 	//   in the interrupt handler.
+ 
+ 	/* Enable transmitter and receiver */
+ 	at91_spi_write(AT91_PDC_PTCR, AT91_PDC_RXTEN | AT91_PDC_TXTEN);
+ 
+ 	at91_spi_write(AT91_SPI_IER, AT91_SPI_ENDRX);		/* enable buffer complete interrupt */
+ 	wait_for_completion(&transfer_complete);
+ 
+ #ifdef DEBUG_SPI
+ 	printk("SPI transfer end\n");
+ #endif
+ 
+ 	return 0;
+ }
+ 
+ /* ......................................................................... */
+ 
+ /*
+  * Handle interrupts from the SPI controller.
+  */
+ static irqreturn_t spi_interrupt(int irq, void *dev_id, struct pt_regs *regs)
+ {
+  	unsigned int status;
+ 	struct spi_local *device = (struct spi_local *) &spi_dev[current_device];
+ 	struct spi_transfer_list *list = device->xfers;
+ 
+ #ifdef DEBUG_SPI
+ 	printk("SPI interrupt %i\n", current_device);
+ #endif
+ 
+ 	if (!list)
+ 		panic("at91_spi: spi_interrupt with a NULL transfer list");
+ 
+        	status = at91_spi_read(AT91_SPI_SR) & at91_spi_read(AT91_SPI_IMR);	/* read status */
+ 
+ 	dma_unmap_single(NULL, device->tx, list->txlen[list->curr], DMA_TO_DEVICE);
+ 	dma_unmap_single(NULL, device->rx, list->rxlen[list->curr], DMA_FROM_DEVICE);
+ 
+ 	device->tx = device->txnext;	/* move next transfer to current transfer */
+ 	device->rx = device->rxnext;
+ 
+ 	list->curr = list->curr + 1;
+ 	if (list->curr == list->nr_transfers) {		/* all transfers complete */
+ 		at91_spi_write(AT91_SPI_IDR, AT91_SPI_ENDRX);		/* disable interrupt */
+ 
+ 		/* Disable transmitter and receiver */
+ 		at91_spi_write(AT91_PDC_PTCR, AT91_PDC_RXTDIS | AT91_PDC_TXTDIS);
+ 
+ 		device->xfers = NULL;
+ 		complete(&transfer_complete);
+ 	}
+ 	else if (list->curr+1 == list->nr_transfers) {	/* no more next transfers */
+ 		device->txnext = 0;
+ 		device->rxnext = 0;
+ 		at91_spi_write(AT91_PDC_TNCR, 0);
+ 		at91_spi_write(AT91_PDC_RNCR, 0);
+ 	}
+ 	else {
+ 		int i = (list->curr)+1;
+ 		
+ 		/* If we are in 16-bit mode, we need to modify what we pass to the PDC */
+ 		int tx_size = (at91_spi_read(AT91_SPI_CSR(current_device)) & AT91_SPI_BITS_16) ? 2 : 1;
+ 
+ 		device->txnext = dma_map_single(NULL, list->tx[i], list->txlen[i], DMA_TO_DEVICE);
+ 		device->rxnext = dma_map_single(NULL, list->rx[i], list->rxlen[i], DMA_FROM_DEVICE);
+ 		at91_spi_write(AT91_PDC_TNPR, device->txnext);
+ 		at91_spi_write(AT91_PDC_RNPR, device->rxnext);
+ 		at91_spi_write(AT91_PDC_TNCR, list->txlen[i] / tx_size);
+ 		at91_spi_write(AT91_PDC_RNCR, list->rxlen[i] / tx_size);
+ 	}
+ 	return IRQ_HANDLED;
+ }
+ 
+ /* ......................................................................... */
+ 
+ /*
+  * Initialize the SPI controller
+  */
+ static int __init at91_spi_init(void)
+ {
+ 	int i;
+ 
+ 	init_MUTEX(&spi_lock);
+ 
+ 	AT91_CfgPIO_SPI();
+ 
+ 	at91_spi_write(AT91_SPI_CR, AT91_SPI_SWRST);	/* software reset of SPI controller */
+ 
+ 	/* Set Chip Select registers to good defaults */
+ 	for (i = 0; i < 4; i++) {
+ 		at91_spi_write(AT91_SPI_CSR(i), AT91_SPI_CPOL | AT91_SPI_BITS_8 | (16 << 16) | (DEFAULT_SPI_BAUD << 8));
+ 	}
+ 
+ 	at91_spi_write(AT91_PDC_PTCR, AT91_PDC_RXTDIS | AT91_PDC_TXTDIS);
+ 
+ 	memset(&spi_dev, 0, sizeof(spi_dev));
+ 	spi_dev[0].pcs = 0xE;
+ 	spi_dev[1].pcs = 0xD;
+ 	spi_dev[2].pcs = 0xB;
+ 	spi_dev[3].pcs = 0x7;
+ 
+ 	if (request_irq(AT91_ID_SPI, spi_interrupt, 0, "spi", NULL))
+ 		return -EBUSY;
+ 
+ 	at91_spi_write(AT91_SPI_CR, AT91_SPI_SPIEN);		/* Enable SPI */
+ 
+ 	return 0;
+ }
+ 
+ static void at91_spi_exit(void)
+ {
+ 	at91_spi_write(AT91_SPI_CR, AT91_SPI_SPIDIS);		/* Disable SPI */
+ 
+ 	free_irq(AT91_ID_SPI, 0);
+ }
+ 
+ 
+ EXPORT_SYMBOL(spi_access_bus);
+ EXPORT_SYMBOL(spi_release_bus);
+ EXPORT_SYMBOL(spi_transfer);
+ 
+ module_init(at91_spi_init);
+ module_exit(at91_spi_exit);
+ 
+ MODULE_LICENSE("GPL")
+ MODULE_AUTHOR("Andrew Victor")
+ MODULE_DESCRIPTION("SPI driver for Atmel AT91RM9200")
