/***********************************************************************
 *
 *  Copyright 2005 by Sistemas Embebidos S.A.
 *  A Digi International Company
 *  All rights reserved.
 *
 *  $Id: irq_latency_unc90.c,v 1.1 2005-10-05 11:54:45 pperez Exp $
 *  @Author: Pedro Perez de Heredia
 *  @Descr: This tests interrupt latency by toggleing an acknowledge bit
 *          each time an interrupt is received. The latency can be
 *          measured with scope from input pulse to output pulse.
 *          An external interrupt generator is needed.
 *          UNC90 PORTC0 is used as interrupt input and PORTA6 as 
 *          external output.
 ***********************************************************************
 *  @History:
 *     2005/04/14 : Initial version.
 ***********************************************************************/
 
#include <linux/module.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>

// Some constant values
#define	UNC90_IRQ_LAT_VERSION	"V1.0"
#define	UNC90_IRQ_LAT_DRV_NAME	"irq_latency_unc90"


/***********************************************************************
 * @Function: unc90_irq_latency_handler
 * @Return: IRQ_HANDLED
 * @Descr: minimal interrupt handler, setting IRQ ACK asap.
 ***********************************************************************/
static irqreturn_t unc90_irq_latency_handler( int irq, void *dev_id, struct pt_regs *regs )
{
	int loop;
	
	// Switch the gpio line inmediatly
	AT91_SYS->PIOD_CODR = 0x02000000;
	
	for( loop = 100; loop > 0; loop-- )
		;
	AT91_SYS->PIOD_SODR = 0x02000000;
	
	return IRQ_HANDLED;
}


/***********************************************************************
 * @Function: unc90_irq_latency_exit
 * @Return:   Nothing.
 * @Descr:    Module unload function. 
 ***********************************************************************/
static void unc90_irq_latency_exit(void)
{
	unsigned int mask;
	
	// Configure UNC90 PORTA6 -> AT91RM9200 PORTD25 as input
	mask = 0x1 << 25;
	AT91_SYS->PIOD_ODR = mask;	// Set as input
	
	free_irq( AT91C_ID_IRQ0, NULL );

	// Configure UNC90 PORTC0 -> AT91RM9200 PORTB6 as input
	mask = 0x1 << 6;
	AT91_SYS->PIOB_ODR = mask;	// Set as input
}


/***********************************************************************
 * @Function: unc90_irq_latency_init
 * @Return:   0 on success, error code otherwise.
 * @Descr:    Module load function. 
 ***********************************************************************/
static int unc90_irq_latency_init( void )
{	
	int ret;
	unsigned int mask;
	
	// Enable PIO controller clk (not for PIOC)
	AT91_SYS->PMC_PCER = (1 << AT91C_ID_PIOB) |
	                     (1 << AT91C_ID_PIOD);

	// Configure UNC90 PORTA6 -> AT91RM9200 PORTD25 as output
	mask = 0x1 << 25;
	AT91_SYS->PIOD_PER = mask;	// Enable PIO pin
	AT91_SYS->PIOD_OER = mask;	// Set as output
	
	// Configure UNC90 PORTC0 -> AT91RM9200 PORTB6 as interrupt
	// and request the irq
	if( ( ret = request_irq( AT91C_ID_IRQ0, unc90_irq_latency_handler,
                                 SA_INTERRUPT, UNC90_IRQ_LAT_DRV_NAME, NULL ) ) != 0 ) {
	    printk( KERN_ERR "Unable to request irq %d, ret %d\n", AT91C_ID_IRQ0, ret );
	    goto error_irq;
	}

	// All ok...
	printk( KERN_INFO UNC90_IRQ_LAT_DRV_NAME ": UNC90 GPIO driver " UNC90_IRQ_LAT_VERSION "\n" );
	return 0;
	
error_irq: 
	unc90_irq_latency_exit();
	return ret;
}



MODULE_AUTHOR( "Pedro Perez de Heredia <pperez@embebidos.com>" );
MODULE_DESCRIPTION( "Irq latency driver for UNC90" );
MODULE_LICENSE( "GPL" );

module_init(unc90_irq_latency_init);
module_exit(unc90_irq_latency_exit);

