/***********************************************************************
 *
 *  Copyright 2003 by FS Forth-Systeme GmbH.
 *  All rights reserved.
 *
 *  $Id: lm202.c,v 1.1 2005-09-08 13:42:41 pperez Exp $
 *  @Author: Markus Pietrek
 *  @Descr: LCD display driver for the Sharp LM202KS1 providing character
 *          output on UNC20. Requires that CS3 has been initialized.
 * 	    Performs initialization as described in
 *          description taken from HD44780_Instruction_Set.txt
 *          write() behaves as noted below.
 *           o on an \n, the output is directed to the next line
 *             if already at last line of display, the first one is used
 *           o on an \a, the display is cleared
 *           o on an \r, the cursor is moved to the beginning of the line
 *           o on an \b, the display is turned off
 *           o on an \v, the display is turned on
 *           o all other characters are written to the buffer/display
 *           o if more data on the actual line is written than can be
 *             displayed, the current line only is scrolled.
 *          It is accessable via /dev/misc/lcd
 ***********************************************************************/
/*********************************************************************** *
 *  @History:
 *     2005/04/04 : Ported to kernel 2.6 by Pedro Perez de Heredia.
 *     2005/02/03 : Ported to UNC90 by Pedro Perez de Heredia.
 *     2003/04/03 : using auto minor mode of misc device
 *                  fixed some symbol sections
 *     2003/03/11 : first timeout will be used for (un-)detection of display
 ********************************************************************** */

#include <linux/module.h>	    // MODULE_AUTHOR ..
#include <linux/init.h>		    // __init
#include <linux/ioport.h>	    // request_region
#include <linux/miscdevice.h>	    // misc_register
#include <asm/io.h>		    // writeb, readb
#include <asm/uaccess.h>	    // get_user
#include <asm/delay.h>		    // udelay
#include <linux/delay.h>
#include <linux/fs.h>               // inode...
#include <asm/arch/hardware.h>    // UNC90_CS3_BASE_ADDR
#include <asm/arch/unc90utils.h>    // unc90_config_smc_cs
#include <linux/lm202.h>	    // driver definitions


#define DEBUG_INIT		1
#define DEBUG_MAJOR_FUNC	2
#define DEBUG_MINOR_FUNC	3
#define DEBUG_LOWEST_FUNC	4
#define DEBUG_ALL		DEBUG_LOWEST_FUNC

//#define DEBUG_VERBOSE		DEBUG_ALL

#ifdef DEBUG_VERBOSE
# define DEBUG_ARGS( n, args... ) \
	if( n <= DEBUG_VERBOSE ) \
		printk( KERN_INFO args )
# define DEBUG_FN( n ) \
	if( n <= DEBUG_VERBOSE ) \
		printk( KERN_INFO "%s:line %d\n", __FUNCTION__,__LINE__ );
# define ASSERT(expr, func) \
	if(!(expr)) { \
        	printk( "Assertion failed! %s:%s:line %d %s\n", \
        	__FILE__,__FUNCTION__,__LINE__,(#expr));  \
        	func }
#else
# define DEBUG_ARGS( n, expr, args... )
# define DEBUG_FN( n )
# define ASSERT(expr, func)
#endif // DEBUG

#define DRIVER_NAME			__FILE__

#define LM202_BASE_ADDRESS		NETARM_MMAP_CS3_BASE
#define LM202_REGION_LENGTH     	0x4  // two address lines
#define LM202_MAX_ADDR          	0x7f // 7 address bits
#define LM202_MAX_VIS_CHARS_IN_LINE	20   // visible characters in line
#define LM202_MAX_PHYS_CHARS_IN_LINE	64   // characters in line until wrap
#define LM202_MAX_LINES			2    // lines the control is capable off


// these offsets have to be added for get_lm202_reg_addr
// for W instructions only write is possible, for R instructions only
// read is possible
#define WINST			0 // write instruction
#define WDATA			1 // write data
#define RINST			2 // read instruction
#define RDATA			3 // read data

static loff_t currentAddr = 0;	  // current position in LCD display buffer

static int lm202_init_module( void );
static void lm202_cleanup_module( void );

static loff_t lm202_lseek( struct file* pFile, loff_t offset, int nOrig );
static ssize_t lm202_write( struct file* pFile, const char* pcBuf, size_t count,
			    loff_t* pPos );
static int lm202_open( struct inode* pInode, struct file* pFile );
static int lm202_release( struct inode* pInode, struct file* pFile );

static int lm202_init( void );
static void lm202_clearScreen( void );
static void lm202_displayOnOff( unsigned char ucDisplay,
				unsigned char ucCursor,
				unsigned char ucBlink );
static void lm202_displayChar( unsigned int unAddr, unsigned char ucChar );
static int lm202_writeCmd( unsigned int unAddr, unsigned char ucChar );
static int lm202_waitReady( void );

static struct file_operations lm202_fops =
{
	.owner		= THIS_MODULE,
	.llseek		= lm202_lseek,
	.write		= lm202_write,
	.open		= lm202_open,
	.release	= lm202_release,
};

// don't pollute major number space, use misc device structure
static struct miscdevice lm202_devs = 
{
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= LM202_CHRDEV,
	.fops		= &lm202_fops
};

// characters are stored for scrolling write-through 
unsigned char aucLm202ScreenBuffer[ LM202_MAX_ADDR + 1 ];
unsigned int unBaseAddr;

#if 0

/***********************************************************************
 * @Function: at91rm9200_config_smc_cs
 * @Return:   0 on success, other value on error.
 * @Descr:    Configures the Static Memory Controller Chip Select
 *            register. 
 ***********************************************************************/
static int at91rm9200_config_smc_cs( int cs, int wait_states, 
                                     int waits_enable, int tdf, 
                                     int bat, int dbw, int drp,
                                     int acss, int rwsetup, int rwhold )
{
	unsigned int mask;
	
	// Check parameters
	if( ( (cs < 0) || (cs > 7) ) &&
            ( (wait_states < 0) || (wait_states > 127) ) &&
            ( (tdf < 0) || (tdf > 127) ) &&
            ( (dbw < 0) || (dbw > 2) ) &&
            ( (acss < 0) || (acss > 3) ) &&
            ( (rwsetup < 0) || (rwsetup > 7) ) &&
            ( (rwhold < 0) || (rwhold > 7) ) )
	{   
	    return -1;
	}
	    
	// Configure GPIOs, if necessary
	if( cs > 3 )
	{
	    switch( cs )
	    {
	        case 4:	mask = AT91C_PC10_NCS4_CFCS;	break;
	        case 5:	mask = AT91C_PC11_NCS5_CFCE1;	break;
	        case 6:	mask = AT91C_PC12_NCS6_CFCE2;	break;
	        case 7:	mask = AT91C_PC13_NCS7;		break;
	        default: mask = 0;
	    }
	    
	    AT91_SYS->PIOC_ASR = mask;	// Select Peripheral A function
	    AT91_SYS->PIOC_PDR = mask;	// Disable PIO controller and enable peripheral
	}
	
	// Write to SMC Chip Select register the new configuration
	AT91_SYS->EBI_SMC2_CSR[cs] = (AT91C_SMC2_NWS & wait_states) | 
	                             (waits_enable ? AT91C_SMC2_WSEN : 0) | 
                                     ((0x0f & tdf) << 8) | (bat ? AT91C_SMC2_BAT : 0) | 
	                             ((0x03 & dbw) << 13) | (drp ? AT91C_SMC2_DRP : 0) | 
	                             ((0x03 & acss) << 16) | ((0x07 & rwsetup) << 24) | 
	                             ((0x07 & rwhold) << 28);

	return 0;
}

/***********************************************************************
 * @Function: unc90_config_smc_cs
 * @Return:   0 on success, other value on error.
 * @Descr:    Configures the CS signals CS3 and CS4. 
 ***********************************************************************/
int unc90_config_smc_cs( int cs, int wait_states, 
                         int waits_enable, int tdf, 
                         int bat, int dbw, int drp,
                         int acss, int rwsetup, int rwhold )
{
	// Check parameters
	if( ( (cs < 3) || (cs > 4) ) )
	    return -1;

	return at91rm9200_config_smc_cs( cs, wait_states, waits_enable,
                                         tdf, bat, dbw, drp, acss, 
                                         rwsetup, rwhold );
}    
#endif

/***********************************************************************
 * @Function: lm202_reg_write
 * @Return: Nothing.
 * @Descr: Writes a byte to the selected LM202 register.
 ***********************************************************************/
static void lm202_reg_write( unsigned char ucVal, unsigned char ucRegister )
{
	DEBUG_FN( DEBUG_MAJOR_FUNC );

	writeb( ucVal, unBaseAddr + ucRegister );
}

/***********************************************************************
 * @Function: lm202_reg_read
 * @Return: The byte read.
 * @Descr: Reads a byte from the selected LM202 register.
 ***********************************************************************/
static unsigned char lm202_reg_read( unsigned char ucRegister )
{
	DEBUG_FN( DEBUG_MAJOR_FUNC );

	return readb( unBaseAddr + ucRegister );
}


/***********************************************************************
 * @Function: lm202_lseek
 * @Return: new position from beginning
 * @Descr: set address of next write. Only SEEK_SET is supported
 ***********************************************************************/

static loff_t lm202_lseek( struct file* pFile, loff_t offset, int nOrig )
{
	DEBUG_FN( DEBUG_MAJOR_FUNC );

	if( offset != 0 ) // SEEK_SET not defined globally
	    return -EINVAL;
	
	if( nOrig >= LM202_MAX_ADDR )
	    return -EINVAL;
	else
	    return ( currentAddr = nOrig );
}


/***********************************************************************
 * @Function: lm202_write
 * @Return: number characters written
 * @Descr: writes the characters to the current position
 ***********************************************************************/

static ssize_t lm202_write( struct file* pFile,
			    const char* pcBuf,
			    size_t count,
			    loff_t* pPos )
{
	unsigned char ucUser;
	size_t charsToWrite = count;
	
	DEBUG_FN( DEBUG_MAJOR_FUNC );

	while( count-- )
	{
	    if( currentAddr >= LM202_MAX_ADDR )
		return charsToWrite-count;	// ignore the rest as of end of
						// display 

	    get_user( ucUser, pcBuf++ );

	    // check for control code
	    switch( ucUser )
	    {
		case LM202_CMD_CLEAR_SCREEN:
		    currentAddr = 0;
		    lm202_clearScreen();
		    break;

		case LM202_CMD_CARRIAGE_RET:
		    // select begin of line
		    currentAddr = (currentAddr / LM202_MAX_PHYS_CHARS_IN_LINE) *
			LM202_MAX_PHYS_CHARS_IN_LINE;
		    break;

		case LM202_CMD_NEWLINE:
		    // calc address of begin of next line
		    currentAddr = (currentAddr / LM202_MAX_PHYS_CHARS_IN_LINE) * 
			(LM202_MAX_PHYS_CHARS_IN_LINE) +
			LM202_MAX_PHYS_CHARS_IN_LINE;

		    if( currentAddr / LM202_MAX_PHYS_CHARS_IN_LINE >=
			LM202_MAX_LINES )
			// already beneath the end of display, go the begin
			// without scrolling
			currentAddr = 0;
		    break;

		case LM202_CMD_DISPLAY_OFF:
		    lm202_displayOnOff( 0, 0, 0 );
		    break;

		case LM202_CMD_DISPLAY_ON:
		    lm202_displayOnOff( 1, 0, 0 );
		    break;

		default:
		    if( currentAddr % (LM202_MAX_PHYS_CHARS_IN_LINE) >=
			LM202_MAX_VIS_CHARS_IN_LINE ) 
		    {
			// no more horizontal space, so scroll the line
			// not very efficient for repeated calling
			unsigned char ucBeginOfLine;
			unsigned int i;

			ucBeginOfLine=(currentAddr/LM202_MAX_PHYS_CHARS_IN_LINE)*
			    LM202_MAX_PHYS_CHARS_IN_LINE;

			// scroll current line 1 character left
			memcpy( aucLm202ScreenBuffer + ucBeginOfLine,
				aucLm202ScreenBuffer + ucBeginOfLine + 1,
				LM202_MAX_VIS_CHARS_IN_LINE - 1 );

			// update display
			for( i = ucBeginOfLine;
			     i < ucBeginOfLine + LM202_MAX_VIS_CHARS_IN_LINE - 1;
			     i++ )
			    lm202_displayChar( i, aucLm202ScreenBuffer[ i ] );

			currentAddr--;
		    }
		    
		    // copy character to be printed in buffer (for scrolling)
		    aucLm202ScreenBuffer[ currentAddr ] = ucUser;
		    // display character
		    lm202_displayChar( currentAddr, ucUser );
		    // position to next character
		    currentAddr++;
	    }
	}
	
	return charsToWrite;
}


/***********************************************************************
 * @Function: lm202_open
 * @Return: 0
 * @Descr: opens and initializes the LCD and clears the screen
 ***********************************************************************/

static int lm202_open( struct inode* pInode, struct file* pFile )
{
	DEBUG_FN( DEBUG_MAJOR_FUNC );

	currentAddr = 0;

	// screen may contain garbage from last usage or may be even turned off
	lm202_clearScreen();
	lm202_displayOnOff( 1, 0, 0 );

	return 0;
}


/***********************************************************************
 * @Function: lm202_release
 * @Return: 0
 * @Descr: doesn't turn off the display to enable updates just by cat'ing or
 *         echo'ing 
 ***********************************************************************/

static int lm202_release( struct inode* pInode, struct file* pFile )
{
	DEBUG_FN( DEBUG_MAJOR_FUNC );

	return 0;
}


/***********************************************************************
 * @Function: lm202_init
 * @Return: 0 on failure otherwise 1
 * @Descr: resets the LCD controller
 ***********************************************************************/

static int lm202_init( void )
{
	DEBUG_FN( DEBUG_INIT );

	lm202_reg_write( 0x30, WINST );
	//udelay( 150000 );
	mdelay( 150 );
	lm202_reg_write( 0x30, WINST );
	//udelay( 150000 );
	mdelay( 150 );
	lm202_reg_write( 0x30, WINST );
	//udelay( 150000 );
	mdelay( 150 );
	if( !lm202_writeCmd( WINST, 0x38 ) ) // 8bit, 1 line, 5x7 dot character
	    // lm202 not detected
	    return 0;
	
	// turn off display
	lm202_displayOnOff( 0, 0, 0 );
	// turn on display
	lm202_displayOnOff( 1, 0, 0 );

	lm202_writeCmd( WINST, 0x6 ); // entry mode increment, no shift

	return 1;
}


/***********************************************************************
 * @Function: lm202_clearScreen
 * @Return: nothing
 * @Descr: erases screen and screen buffer
 ***********************************************************************/

static void lm202_clearScreen( void )
{
	DEBUG_FN( DEBUG_MINOR_FUNC );

	lm202_writeCmd( WINST, 0x01 );

	// reset buffer, too
	memset( aucLm202ScreenBuffer, ' ', sizeof( aucLm202ScreenBuffer ) );
}


/***********************************************************************
 * @Function: lm202_displayOnOff
 * @Return: nothing
 * @Descr: enables or disables the display
 ***********************************************************************/

static void lm202_displayOnOff( unsigned char ucDisplay,
				unsigned char ucCursor,
				unsigned char ucBlink )
{
	DEBUG_FN( DEBUG_MINOR_FUNC );

	lm202_writeCmd( WINST, 0x8 + (ucDisplay<<2) + (ucCursor<<1) + ucBlink );
}


/***********************************************************************
 * @Function: lm202_displayChar
 * @Return: nothing
 * @Descr: writes a character into the LCD controller memory. unAddr may be out
 *         of visible range. No address checking is being performed.
 ***********************************************************************/

static void lm202_displayChar( unsigned int unAddr, unsigned char ucChar )
{
	DEBUG_FN( DEBUG_LOWEST_FUNC );

	lm202_writeCmd( WINST, 0x80 | unAddr );
	lm202_writeCmd( WDATA, ucChar );
}


/***********************************************************************
 * @Function: lm202_writeCmd
 * @Return: remaining time for timeout
 * @Descr: writes a command or a value to the LCD controller and waits for
 *         ready
 ***********************************************************************/

static int lm202_writeCmd( unsigned int unAddr, unsigned char ucChar )
{
	DEBUG_FN( DEBUG_LOWEST_FUNC );

	lm202_reg_write( ucChar, unAddr );

	return lm202_waitReady();
}

/***********************************************************************
 * @Function: lm202_waitReady
 * @Return: remaining time for timeout
 * @Descr: waits until LCD is ready or timeout occurs.
 ***********************************************************************/

static int lm202_waitReady( void )
{
	int unTimeout = 100000;
	
	DEBUG_FN( DEBUG_LOWEST_FUNC );

	// hope that timeout is long enough as I don't have any values
	while( ( lm202_reg_read( RINST ) & 0x80 ) == 0x80 && --unTimeout )
	{
	    // does nothing but wait
	}

	if( !unTimeout )
	    printk( KERN_WARNING DRIVER_NAME ": lm202_waitReady timed out\n" );

	return unTimeout;
}

// lm202_init_module, lm202_cleanup_module at end as of linux convention

/***********************************************************************
 * @Function: lm202_init_module
 * @Return: 0 if successfull otherwise error
 * @Descr: initialized the LM202KS1 LCD interface. Doesn't perform
 *         any hardware detection. Requires that CS3 has been initialized.
 *         Enables LCD
 ***********************************************************************/

static int __init lm202_init_module( void )
{
	int nResult = 0;
	
	DEBUG_FN( DEBUG_INIT );
	
	// Configure CS3 signal
	if( unc90_config_smc_cs( 3, 127, 1, 0, 0, 2, 0, 0, 0, 0 ) != 0 )
	{
	    printk( KERN_ERR "Unable to configure CS3 signal\n" );
            return -EIO;
	}
	  
	// Request memory for device access
	if( ( nResult = check_mem_region( UNC90_CS3_BASE_ADDR, LM202_REGION_LENGTH ) ) )
	{
	    printk( KERN_ERR "%s: can't get mem address 0x%08x - 0x%08x\n", "lm202",
	            UNC90_CS3_BASE_ADDR, UNC90_CS3_BASE_ADDR + LM202_REGION_LENGTH );
            
	    nResult = -EBUSY;
	    goto error_region;
	}
	
	// Request mem space
	request_mem_region( UNC90_CS3_BASE_ADDR, LM202_REGION_LENGTH, "lm202" );

	// Try to get IO space...
	unBaseAddr = (unsigned int)ioremap( UNC90_CS3_BASE_ADDR, LM202_REGION_LENGTH );

	if( misc_register( &lm202_devs ) )
	{
		printk( KERN_WARNING DRIVER_NAME \
			": unable to register\n" );
		nResult = -EIO;
		goto error;
	}
	
	printk( KERN_INFO "LM202KS1 LCD Driver Initializing\n" );

	// perform hardware initialization
	if( !lm202_init() )
	{
	    printk( KERN_ERR "LM202KS1 not detected\n" );
	    nResult = -ENODEV;
	    goto error_hw;
	}
	

	// after a software reset some text might be displayed
	lm202_clearScreen();
	
	return 0;

  error_hw:
	misc_deregister( &lm202_devs );
  error:
	release_mem_region( UNC90_CS3_BASE_ADDR, LM202_REGION_LENGTH );

  error_region:
	return nResult;
}


/***********************************************************************
 * @Function: lm202_cleanup_module
 * @Return: nothing
 * @Descr: deinitializes module and shuts off LCD
 ***********************************************************************/

static void __exit lm202_cleanup_module( void )
{
	DEBUG_FN( DEBUG_INIT );

	printk( KERN_INFO "LM202KS1 LCD Driver removing\n" );

	iounmap( (void *)UNC90_CS3_BASE_ADDR );
	release_mem_region( UNC90_CS3_BASE_ADDR, LM202_REGION_LENGTH );
	misc_deregister( &lm202_devs );

	// turn off display
	lm202_displayOnOff( 0, 0, 0 );
}

module_init( lm202_init_module );
module_exit( lm202_cleanup_module );

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>");
MODULE_DESCRIPTION("LM202KS1 LCD Driver On UNC20 From FS Forth-Systeme GmbH");
MODULE_LICENSE("GPL");

