/***********************************************************************
 *
 *  Copyright 2005 by Sistemas Embebidos S.A.
 *  All rights reserved.
 *
 *  $Id: unc90_gpio.c,v 1.1 2005-10-05 11:41:51 pperez Exp $
 *  @Author: Pedro Perez de Heredia
 *  @Descr:  GPIO driver for the UNC90 module. Can be compiled as 
 *           module or be statically linked with the kernel. Allows to
 *           handle individually all UNC90 (PORTA/PORTC) gpio pins.
 *  
 ***********************************************************************
 *  @History:
 *     2005/01/14 : PPH : Initial version.
 *     2005/06/01 : PPH : Updated to kernel version 2.6.12.
 ***********************************************************************/
 
#include <linux/fs.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>
#include <linux/poll.h>

#include <asm/uaccess.h>		// copy_to_user and friends...
#include <asm/delay.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>			// AT91_PIN_PA0

#ifdef CONFIG_DEVFS_FS
# include <linux/devfs_fs_kernel.h>
#endif

#include <asm/arch/unc90_gpio.h>	// specific data for this driver


// Some constant values
#define	UNC90_GPIO_VERSION	"V1.0"
#define	UNC90_GPIO_DRV_NAME	"unc90_gpio"


// Function prototypes
static void unc90_gpio_per( pio_t port, unsigned int val );
static void unc90_gpio_set_ifr( pio_t port, unsigned int mask, int ifr );
static void unc90_gpio_oer_odr( pio_t port, unsigned int val, int inp );
static void unc90_gpio_set_out( pio_t port, unsigned int mask, int val );
static int unc90_gpio_get_pin_val( pio_t port, unsigned int mask );
static int unc90_gpio_config_pin( unc90_gpio_conf_t *conf );
static int unc90_gpio_ioctl( struct inode *inode, struct file *file,
                             unsigned int cmd, unsigned long arg );
static unsigned int unc90_gpio_poll( struct file *file, poll_table *wait );
static ssize_t unc90_gpio_write( struct file * file, const char * buf, size_t count,
                                 loff_t * ppos );
static ssize_t unc90_gpio_read( struct file * file, char * buf, size_t count,
                                loff_t * ppos );
static int unc90_gpio_open( struct inode * inode, struct file * file );
static int unc90_gpio_release( struct inode * inode, struct file * file );




// Driver file functions
static struct file_operations unc90_gpio_fops = {
	.owner		= THIS_MODULE,
	.read		= unc90_gpio_read,
	.write		= unc90_gpio_write,
	.ioctl		= unc90_gpio_ioctl,
	.poll		= unc90_gpio_poll,
	.open		= unc90_gpio_open,
	.release	= unc90_gpio_release,
};

// Gpio settings
struct gpio_conf {
	int			dev_id;		// minor
	int			port_num;	// port number 0-porta, 1-portc
	int			irq_num;	// irq number assigned
	unc90_gpio_conf_t 	config;		// port settings
	wait_queue_head_t	wait_q;		// A nice place to sleep a bit...
	int			triggered;	// Just to avoid race conditions
};

// Device structure
struct gpio_dev {
	struct cdev		cdev;			// Char device structure
	struct gpio_conf	gpio[UNC90_MAX_GPIOS];	// gpio structs
};

static char *unc90_gpio_dev_names[] = { "gpio/porta", "gpio/portc" };
static int unc90_gpio_major = UNC90_GPIO_MAJOR;
static int unc90_gpio_minor = 0;
static int unc90_gpio_nr_devs = UNC90_MAX_GPIOS;
static struct gpio_dev *gpio_device;


/***********************************************************************
 * @Function: unc90_gpio_per
 * @Return:   Nothing.
 * @Descr:    Sets the selected port pin in PIO mode.
 ***********************************************************************/
static void unc90_gpio_per( pio_t port, unsigned int val )
{
	switch( port ) {
	    case PIO_A:	AT91_SYS->PIOA_PER = val;	break;
	    case PIO_B:	AT91_SYS->PIOB_PER = val;	break;
	    case PIO_C:	AT91_SYS->PIOC_PER = val;	break;
	    case PIO_D:	AT91_SYS->PIOD_PER = val;	break;
	}
}

/***********************************************************************
 * @Function: unc90_gpio_set_ifr
 * @Return:   Nothing.
 * @Descr:    Sets or clears the input filter register for the selected
 *            port pin.
 ***********************************************************************/
static void unc90_gpio_set_ifr( pio_t port, unsigned int mask, int ifr )
{
	if( ifr ) {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_IFER = mask;	break;
	        case PIO_B:	AT91_SYS->PIOB_IFER = mask;	break;
	        case PIO_C:	AT91_SYS->PIOC_IFER = mask;	break;
	        case PIO_D:	AT91_SYS->PIOD_IFER = mask;	break;
	    }
	} else {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_IFDR = mask;	break;
	        case PIO_B:	AT91_SYS->PIOB_IFDR = mask;	break;
	        case PIO_C:	AT91_SYS->PIOC_IFDR = mask;	break;
	        case PIO_D:	AT91_SYS->PIOD_IFDR = mask;	break;
	    }
	}
}

/***********************************************************************
 * @Function: unc90_gpio_set_pullup
 * @Return:   Nothing.
 * @Descr:    Sets or clears the port pull ups for the selected port pin.
 ***********************************************************************/
static void unc90_gpio_set_pullup( pio_t port, unsigned int mask, int onoff )
{
	if( onoff ) {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_PPUER = mask;	break;
	        case PIO_B:	AT91_SYS->PIOB_PPUER = mask;	break;
	        case PIO_C:	AT91_SYS->PIOC_PPUER = mask;	break;
	        case PIO_D:	AT91_SYS->PIOD_PPUER = mask;	break;
	    }
	} else {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_PPUDR = mask;	break;
	        case PIO_B:	AT91_SYS->PIOB_PPUDR = mask;	break;
	        case PIO_C:	AT91_SYS->PIOC_PPUDR = mask;	break;
	        case PIO_D:	AT91_SYS->PIOD_PPUDR = mask;	break;
	    }
	}
}

/***********************************************************************
 * @Function: unc90_gpio_oer_odr
 * @Return:   Nothing.
 * @Descr:    Sets the selected port pin as output or input, depending
 *            on the value passed through inp (0->out/1->inp).
 ***********************************************************************/
static void unc90_gpio_oer_odr( pio_t port, unsigned int val, int inp )
{
	if( inp ) {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_ODR = val;	break;
	        case PIO_B:	AT91_SYS->PIOB_ODR = val;	break;
	        case PIO_C:	AT91_SYS->PIOC_ODR = val;	break;
	        case PIO_D:	AT91_SYS->PIOD_ODR = val;	break;
	    }
	} else {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_OER = val;	break;
	        case PIO_B:	AT91_SYS->PIOB_OER = val;	break;
	        case PIO_C:	AT91_SYS->PIOC_OER = val;	break;
	        case PIO_D:	AT91_SYS->PIOD_OER = val;	break;
	    }
	}
}

/***********************************************************************
 * @Function: unc90_gpio_oer
 * @Return:   Nothing.
 * @Descr:    Sets the selected port pin as output.
 ***********************************************************************/
static void unc90_gpio_set_out( pio_t port, unsigned int mask, int val )
{
	if( val ) {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_SODR = mask;	break;
	        case PIO_B:	AT91_SYS->PIOB_SODR = mask;	break;
	        case PIO_C:	AT91_SYS->PIOC_SODR = mask;	break;
	        case PIO_D:	AT91_SYS->PIOD_SODR = mask;	break;
	    }
	} else {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_CODR = mask;	break;
	        case PIO_B:	AT91_SYS->PIOB_CODR = mask;	break;
	        case PIO_C:	AT91_SYS->PIOC_CODR = mask;	break;
	        case PIO_D:	AT91_SYS->PIOD_CODR = mask;	break;
	    }
	}
}

/***********************************************************************
 * @Function: unc90_gpio_irq_reg
 * @Return:   Nothing.
 * @Descr:    Enables or disables the selected irq port pin.
 ***********************************************************************/
static void unc90_gpio_irq_reg( pio_t port, unsigned int val, int enab )
{
	if( enab ) {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_IER = val;	break;
	        case PIO_B:	AT91_SYS->PIOB_IER = val;	break;
	        case PIO_C:	AT91_SYS->PIOC_IER = val;	break;
	        case PIO_D:	AT91_SYS->PIOD_IER = val;	break;
	    }
	} else {
	    switch( port ) {
	        case PIO_A:	AT91_SYS->PIOA_IDR = val;	break;
	        case PIO_B:	AT91_SYS->PIOB_IDR = val;	break;
	        case PIO_C:	AT91_SYS->PIOC_IDR = val;	break;
	        case PIO_D:	AT91_SYS->PIOD_IDR = val;	break;
	    }
	}
}

/***********************************************************************
 * @Function: unc90_gpio_get_pin_val
 * @Return:   Nothing.
 * @Descr:    Gets the value set on a the selected port pin. Following 
 *            function can be used with port pin configured as outputs
 *            and inputs
 ***********************************************************************/
static int unc90_gpio_get_pin_val( pio_t port, unsigned int mask )
{
	int retval = 1;
	
	switch( port ) {
	    case PIO_A:	retval = AT91_SYS->PIOA_PDSR & mask;	break;
	    case PIO_B:	retval = AT91_SYS->PIOB_PDSR & mask;	break;
	    case PIO_C:	retval = AT91_SYS->PIOC_PDSR & mask;	break;
	    case PIO_D:	retval = AT91_SYS->PIOD_PDSR & mask;	break;
	}

	return ( retval != 0 );
}

/***********************************************************************
 * @Function: latencyInterruptHandler
 * @Return: IRQ_HANDLED
 * @Descr: minimal interrupt handler, setting IRQ ACK asap.
 ***********************************************************************/
static irqreturn_t unc90_gpio_irq_handler( int irq, void *dev_id, struct pt_regs *regs )
{
	struct gpio_conf *gpio = (struct gpio_conf *)dev_id;
	
	// Sanity check
	if( gpio != NULL) {
	    // Awake those processes waiting...
	    gpio->triggered = 1;
	    wake_up_interruptible( &gpio->wait_q );
	}

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: unc90_gpio_config_pin
 * @Return:   0 on success, error code otherwise.
 * @Descr:    Configures the selected pin.
 ***********************************************************************/
static int unc90_gpio_config_pin( unc90_gpio_conf_t *conf )
{
	int ret = 0;
	unsigned int mask = 0x1 << conf->bit;
	
	// Set the port pin in PIO mode
	if( conf->mode != GPIO_UNCONFIG ) {
	    unc90_gpio_per( conf->port, mask );
	}
	// Set the port pin as input/output
	if( conf->mode == GPIO_OUTPUT || conf->mode == GPIO_INPUT ) {
	    unc90_gpio_oer_odr( conf->port, mask, (conf->mode == GPIO_INPUT) ? 1 : 0 );
	} 
	
	// Pull-up enable/disable funtionality depending on flags
	unc90_gpio_set_pullup( conf->port, mask, conf->flags & GPIO_PULL_UP );
	   
	// Input filter enable/disable funtionality depending on flags
	unc90_gpio_set_ifr( conf->port, mask, conf->flags & GPIO_INP_GLITCH_FILTER );
	
	// Configure the port pin as interrupt and enable it
	unc90_gpio_irq_reg( conf->port, mask, conf->mode == GPIO_IRQ ? 1 : 0 );
	
	return ret;
}

/***********************************************************************
 * @Function: unc90_gpio_ioctl
 * @Return:   0 on success, error code otherwise.
 * @Descr:    Ioctl system call.
 ***********************************************************************/
static int unc90_gpio_ioctl( struct inode *inode,
                             struct file *file,
                             unsigned int cmd,
                             unsigned long arg )
{
	struct gpio_conf *gpio = (struct gpio_conf *)file->private_data;
	unc90_gpio_conf_t *portconf;
	char __user *argp = (char __user *)arg;
	unsigned int mask;
	int retval;

	portconf = &gpio->config;
	mask = 0x1 << portconf->bit;

	switch( cmd ) {

	    case GPIO_CONFIG_AS_INP:  /* config as input */
	        portconf->mode = GPIO_INPUT;
	        unc90_gpio_config_pin( portconf );
	        break;
            
	    case GPIO_CONFIG_AS_OUT:  /* config as input */
	        portconf->mode = GPIO_OUTPUT;
	        unc90_gpio_config_pin( portconf );
	        break;

	    case GPIO_CONFIG_AS_IRQ:  /* config as interrupt input */
	        if( portconf->mode != GPIO_IRQ ) {
	            if( gpio->irq_num != -1 ) 
	                return -EBUSY;
	        
	            portconf->mode = GPIO_IRQ;
	            unc90_gpio_config_pin( portconf );
	        
	            init_waitqueue_head( &gpio->wait_q );
	            gpio->triggered = 0;
#if 0	    
	            if( ( retval = request_irq( AT91C_ID_PIOA + gpio->config.port, 
	                                        unc90_gpio_irq_handler, SA_SHIRQ,
	                                        UNC90_GPIO_DRV_NAME, gpio ) ) != 0 ) {
	                printk( KERN_ERR "Unable to request irq %d, ret %d\n", 
	                                 AT91C_ID_PIOA + gpio->config.port, retval );
	                return retval;
	            }
	            gpio->irq_num = AT91C_ID_PIOA + gpio->config.port;
#endif		    
	            if( ( retval = request_irq( AT91_PIN_PA0 + gpio->config.port * 32 +
	                                        gpio->config.bit, unc90_gpio_irq_handler,
	                                        SA_SHIRQ, UNC90_GPIO_DRV_NAME, gpio ) ) != 0 ) {
	                printk( KERN_ERR "Unable to request irq %d, ret %d\n", 
	                                 AT91_PIN_PA0 + gpio->config.port * 32 + gpio->config.bit, retval );
	                return retval;
	            }
	            gpio->irq_num = AT91_PIN_PA0 + gpio->config.port * 32 + gpio->config.bit;
	        }
	        return 0;
	    
	    case GPIO_SET_INP_FILTER:  /* set input filter value */
	        
	        if( *argp == 1 )
	            portconf->flags |= GPIO_INP_GLITCH_FILTER;
	        else
	            portconf->flags &= ~GPIO_INP_GLITCH_FILTER;
	    
	        unc90_gpio_config_pin( portconf );
	        return 0;

	    case GPIO_CONFIG_PULL_UP:  /* set input pullup value */
	        
	        if( *argp == 1 )
	            portconf->flags |= GPIO_PULL_UP;
	        else
	            portconf->flags &= ~GPIO_PULL_UP;
	    
	        unc90_gpio_config_pin( portconf );
	        return 0;
	    
	    case GPIO_READ_PIN_VAL:  /* read value of the selected pin */
	        retval = unc90_gpio_get_pin_val( portconf->port, mask );
	        if( copy_to_user( argp, &retval, sizeof( int ) ) )
	            return -EINVAL;
	        return 0;
	    
	    case GPIO_WRITE_PIN_VAL:  /* write value to the selected pin */
	        unc90_gpio_set_out( portconf->port, mask, *argp );
	        return 0;

	    default:
	        return -EINVAL;
	
	} // switch( cmd )
	
	return 0; // should not reach this
}


/*******************************************************************************
 * @Function: unc90_gpio_poll
 * @Return:
 * @Descr: poll system call
 *******************************************************************************/
static unsigned int unc90_gpio_poll( struct file *file, poll_table *wait )
{
	unsigned int ret_val = 0;
	struct gpio_conf *gpio = (struct gpio_conf *)file->private_data;
	    
	if( file->f_mode & FMODE_WRITE ) 
	    ret_val |= POLLOUT | POLLWRNORM;
	   
	if( file->f_mode & FMODE_READ ) {
	    if( gpio->triggered == 1 ) {
	        ret_val |= POLLOUT | POLLRDNORM;
	    }else
	        poll_wait( file, &gpio->wait_q, wait );
	}
	return ret_val;
}

/***********************************************************************
 * @Function: unc90_gpio_write
 * @Return:   The number of bytes written or error code.
 * @Descr:    write driver function.
 ***********************************************************************/
static ssize_t unc90_gpio_write( struct file * file, const char * buf, size_t count,
                                 loff_t * ppos )
{
	struct gpio_conf *gpio = (struct gpio_conf *)file->private_data;
	int out_val = 0;
	
	if( count != sizeof( char ) )
	    return -EINVAL;

	if( gpio->config.mode != GPIO_OUTPUT )
	    return -EINVAL;
	    
	if( copy_from_user( &out_val, buf, sizeof( char ) ) )
	    return -EFAULT;
	// Write value to output
	unc90_gpio_set_out( gpio->config.port, 0x1 << gpio->config.bit, out_val );

	return ( sizeof( char ) );	
}

/***********************************************************************
 * @Function: unc90_gpio_read
 * @Return:   The number of bytes read or error code.
 * @Descr:    read driver functions.
 ***********************************************************************/
static ssize_t unc90_gpio_read( struct file * file, char * buf, size_t count,
                                loff_t * ppos )
{
	struct gpio_conf *gpio = (struct gpio_conf *)file->private_data;
	int inp_val;
	ssize_t ret = 0;
	
	if( count != sizeof( char ) )
	    return -EINVAL;
	    
	// If configured as input, return the value inmediatly
	if( gpio->config.mode == GPIO_INPUT || gpio->config.mode == GPIO_OUTPUT ) {
	     inp_val = unc90_gpio_get_pin_val( gpio->config.port, 0x1 << gpio->config.bit );
	} else if( gpio->config.mode == GPIO_IRQ ) {
	    
	    // Wait for data
	    if( !file->f_flags & O_NONBLOCK ) {
	        wait_event_interruptible( gpio->wait_q, gpio->triggered != 0 );
	        if( signal_pending( current ) )
	            return -ERESTARTSYS;
	    }
	    
	    gpio->triggered = 0;
	    inp_val = unc90_gpio_get_pin_val( gpio->config.port, 0x1 << gpio->config.bit );
	} else {
	    printk( KERN_WARNING UNC90_GPIO_DRV_NAME ": unconfigured IO, minor %d\n",
	            gpio->dev_id );
	    return -EINVAL;
	}
	
	ret = sizeof( char );
	if( copy_to_user( buf, (char *)&inp_val, sizeof( char ) ) )
	    ret = -EFAULT;
	
	return ret;
}


/***********************************************************************
 * @Function: unc90_gpio_open
 * @Return:   0 on success, error code otherwise.
 * @Descr:    open driver functions.
 ***********************************************************************/
static int unc90_gpio_open( struct inode * inode, struct file * file )
{
	struct gpio_dev *dev;
	struct gpio_conf *gpio;
	int port = MINOR( inode->i_rdev ) / UNC90_GPIO_LINES_X_PORT;
	int ret;
	
	dev = container_of( inode->i_cdev, struct gpio_dev, cdev );
	gpio = &dev->gpio[ MINOR( inode->i_rdev ) ];
	
	// Just to be sure...
	if( port == gpio->port_num ) {
	    file->private_data = gpio; // to have access from other methods 
	
	    unc90_gpio_config_pin( &gpio->config );
	    
	    if( gpio->config.mode == GPIO_IRQ ) {

	        init_waitqueue_head( &gpio->wait_q );
	        gpio->triggered = 0;
	        
		if( ( ret = request_irq( AT91_PIN_PA0 + gpio->config.port * 32 +
	                                 gpio->config.bit, unc90_gpio_irq_handler,
	                                 SA_SHIRQ, UNC90_GPIO_DRV_NAME, gpio ) ) != 0 ) {
	            printk( KERN_ERR "Unable to request irq %d, ret %d\n", 
	                             AT91_PIN_PA0 + gpio->config.port * 32 + gpio->config.bit, ret );
	            goto error_irq;
	        }
	        gpio->irq_num = AT91_PIN_PA0 + gpio->config.port * 32 + gpio->config.bit;

	    }
	} else {
	    printk( KERN_ERR UNC90_GPIO_DRV_NAME 
	            ": data structs seems to be corrupted\n" );
	    return -ENODEV;	    
	}
	
	return 0;
error_irq:
	return ret;	
}

/***********************************************************************
 * @Function: unc90_gpio_release
 * @Return:   0 on success, error code otherwise.
 * @Descr:    close driver functions.
 ***********************************************************************/
static int unc90_gpio_release( struct inode * inode, struct file * file )
{
	struct gpio_conf *gpio = (struct gpio_conf *)file->private_data;
	
	if( gpio->config.mode == GPIO_IRQ && gpio->irq_num != -1 ) {
	    
	    // Disable pin function irq
	    unc90_gpio_irq_reg( gpio->config.port, 0x1 << gpio->config.bit, 0 );
	    
	    free_irq( gpio->irq_num, gpio );
	    gpio->irq_num = -1;
	}

	return 0;
}

/***********************************************************************
 * @Function: unc90_gpio_exit
 * @Return:   Nothing.
 * @Descr:    Module unload function. 
 ***********************************************************************/
static void unc90_gpio_exit( void )
{
	int i;
	dev_t devno = MKDEV( unc90_gpio_major, unc90_gpio_minor );

	if( gpio_device ) {
	    for( i = 0; i < unc90_gpio_nr_devs; i++ ) {
	        if( gpio_device->gpio[i].irq_num != -1 )
	            free_irq( gpio_device->gpio[i].irq_num, &gpio_device->gpio[i] );

	        cdev_del( &gpio_device->cdev );
	    }
	    
	    kfree( gpio_device );
	}

#ifdef CONFIG_DEVFS_FS
	for( i = 0; i < unc90_gpio_nr_devs; i++ ) {
	    
	    devfs_remove( (i / UNC90_GPIO_LINES_X_PORT) ? 
	                  "gpio/portc/%d" : "gpio/porta/%d",
	                  i % UNC90_GPIO_LINES_X_PORT );
	}
#endif
	unregister_chrdev_region( devno, unc90_gpio_nr_devs );
}


/***********************************************************************
 * @Function: unc90_gpio_init
 * @Return:   0 on success, error code otherwise.
 * @Descr:    Module load function. 
 ***********************************************************************/
static int unc90_gpio_init( void )
{	
	int ret, i, devno;
	int port, index;
	dev_t dev;
	
	// Enable PIO controller clk (not for PIOC)
	AT91_SYS->PMC_PCER = (1 << AT91C_ID_PIOA) | 
	                     (1 << AT91C_ID_PIOB) |
	                     (1 << AT91C_ID_PIOD);
	
	if( unc90_gpio_major ) {
	    dev = MKDEV( unc90_gpio_major, unc90_gpio_minor );
	    ret = register_chrdev_region( dev, unc90_gpio_nr_devs, UNC90_GPIO_DRV_NAME );
	} else {
	    ret = alloc_chrdev_region( &dev, unc90_gpio_minor,
	                               unc90_gpio_nr_devs, UNC90_GPIO_DRV_NAME );
	    unc90_gpio_major = MAJOR( dev );
	}
	
	if( ret < 0 ) {
	    printk( KERN_WARNING UNC90_GPIO_DRV_NAME ": can't get major %d\n",
	            unc90_gpio_major );
	    return ret;
	}

        // Allocate memory for the devices
	gpio_device = kmalloc( sizeof( struct gpio_dev ), GFP_KERNEL );
	if( !gpio_device ) {
	    ret = -ENOMEM;
	    goto fail;
	}
	
	memset( gpio_device, 0, sizeof( struct gpio_dev ) );

	cdev_init( &gpio_device->cdev, &unc90_gpio_fops );
	gpio_device->cdev.owner = THIS_MODULE;
	gpio_device->cdev.ops = &unc90_gpio_fops;
	if( cdev_add( &gpio_device->cdev, dev, unc90_gpio_nr_devs ) )
	    printk( KERN_NOTICE UNC90_GPIO_DRV_NAME ": Error adding cdev\n" );
	
#ifdef CONFIG_DEVFS_FS
	for( i = 0; i < UNC90_GPIO_MAX_PORTS; i++ )
	    devfs_mk_dir( unc90_gpio_dev_names[i] );
#endif

        // Initialize each device
	for( i = 0; i < unc90_gpio_nr_devs; i++ ) {
	    devno = MKDEV( unc90_gpio_major, unc90_gpio_minor + i );

	    port = i / UNC90_GPIO_LINES_X_PORT;
	    index = i % UNC90_GPIO_LINES_X_PORT;
	    gpio_device->gpio[i].dev_id = i;
	    gpio_device->gpio[i].port_num = port;
	    gpio_device->gpio[i].irq_num = -1;
	        
	    // Copy default settings...
	    memcpy( &gpio_device->gpio[i].config, port ? 
	            &unc90_portc_default_conf[index] : &unc90_porta_default_conf[index], 
	            sizeof(unc90_gpio_conf_t) );
    	
#ifdef CONFIG_DEVFS_FS
	    devfs_mk_cdev( devno, S_IFCHR | S_IRUSR | S_IWUSR,
	                   port ? "gpio/portc/%d" : "gpio/porta/%d", index );
#endif
	}
	    
	// All ok...
	printk( KERN_INFO UNC90_GPIO_DRV_NAME ": UNC90 GPIO driver " UNC90_GPIO_VERSION "\n" );
	return 0;
fail: 
	unc90_gpio_exit();
	return ret;
}



MODULE_AUTHOR( "Pedro Perez de Heredia <pperez@embebidos.com>" );
MODULE_DESCRIPTION( "UNC90 GPIO driver" );
MODULE_LICENSE( "GPL" );

module_init(unc90_gpio_init);
module_exit(unc90_gpio_exit);

