/***********************************************************************
 *
 * Copyright (C) 2004-2005 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: i2c_ns.c,v 1.8 2006-08-09 12:05:44 uzeisberger Exp $
 * @Author: Markus Pietrek
 * @Descr: I2C driver for the NS9750. No bus driver is needed as these are
 *         cpu interal registers.
 *         Slave mode is not supported.
 * @References: [1] NS9750 Hardware Reference, March 2004, Chap.11
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *     2005/11/29 : ported to kernel 2.6.14
 *     2005/03/01 : NS9360 support added
 *     2004/08/17 : removed addr_
 ***********************************************************************/
#include <linux/config.h>	/* system configuration */

#include <asm/arch/ns9750_i2c.h> /* NS_I2C_MODULE_BASE */
#include <asm/arch/ns9750_bbus.h> /* ns_gpio_lock */

#include <asm/arch/irqs.h>	/* IRQ_BBUS_I2C */
#include <linux/module.h>	/* module_init */
#include <linux/init.h>		/* __init */
#include <linux/ioport.h>	/* request_mem_region */
#include <linux/interrupt.h>	/* request_irq */
#include <linux/wait.h>		/* DECLARE_WAIT_QUEUE_HEAD */
#include <linux/i2c.h>		/* i2c_algorithm */
#include <asm/arch/hardware.h>

#include <asm/io.h>		/* writel */


#define DEBUG_INIT		0x0001
#define DEBUG_MINOR		0x0002
#define DEBUG_RX		0x0004
#define DEBUG_TX		0x0008
#define DEBUG_INT		0x0010
#define DEBUG_ERROR		0x4000
#define DEBUG_ERROR_CRIT	0x8000

static int debug = 0xc000;

#define DEBUG_ARGS( FLG, a0, args... ) if( ( debug & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, ## args )
#define DEBUG_FN( FLG ) if( (debug & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
#define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }

#define DRIVER_NAME		"i2c_ns"

#if 0
	/* use the lowest I2C clock possible for 100kHz ([1] p. 554) */
#define NS_I2C_CFG_INIT_VAL ((((((CPU_CLK_FREQ/4)/100000)-4)/2) & \
			      NS_I2C_CFG_CLREF_MA) | 			\
			     ((1 << NS_I2C_CFG_SFW_SH) & NS_I2C_CFG_SFW_MA ) )
#endif
#if 1
/* due to worse setup timing reduced to 50KHz */
#define NS_I2C_CFG_INIT_VAL (((((CPU_CLK_FREQ/4)/100000)/2*2) & NS_I2C_CFG_CLREF_MA) | \
						 ((1 << NS_I2C_CFG_SFW_SH) & NS_I2C_CFG_SFW_MA ))
#endif
typedef enum
{
	I2C_NS_INT_AWAITING,
	I2C_NS_INT_OK,
	I2C_NS_INT_RETRY,
	I2C_NS_INT_ERROR
} i2c_ns_int_status_t;

static int i2c_ns_master_xfer( struct i2c_adapter* i2c_adap,
			       struct i2c_msg msgs[], int num );
static u32 i2c_ns_master_func( struct i2c_adapter* adap );
static irqreturn_t i2c_ns_int( int irq, void* dev_id,struct pt_regs* regs );
static int i2c_ns_read( struct i2c_adapter* i2c_adap, int count );
static int i2c_ns_write( struct i2c_adapter* i2c_adap,const char* buf,int count);
static int i2c_ns_send_cmd( struct i2c_adapter* i2c_adap, unsigned int cmd );

static int __init i2c_ns_init( void );
static void i2c_ns_exit( void );

DECLARE_WAIT_QUEUE_HEAD( wait_q );

/* no multiple instances possible, don't use struct for local variables */
static spinlock_t i2c_ns_lock;	/* locking for i2c_ns_buf etc. */
static char* i2c_ns_buf = NULL;	/* where to store the incoming character */
static i2c_ns_int_status_t i2c_ns_int_status;

static struct i2c_algorithm i2c_ns_algo = {
	.master_xfer	= i2c_ns_master_xfer,
	.functionality	= i2c_ns_master_func
};

static struct i2c_adapter i2c_ns_adapter = {
	.owner			= THIS_MODULE,
	.name			= DRIVER_NAME,
	.id				= I2C_HW_NS9750,
	.algo			= &i2c_ns_algo,
	.timeout			= HZ,
	.retries			= 1
};

/***********************************************************************
 * @Function: i2c_ns_master_xfer
 * @Return: 0 on failure otherwise number messages sends (== num)
 * @Descr: performs action described in i2c_msg
 *         accepts I2C_M_NOSTART for continous transfer, I2C_M_RD and
 *         !I2C_M_RD (write). No buffer validity check is performed.
 ***********************************************************************/

static int i2c_ns_master_xfer( struct i2c_adapter* i2c_adap,
			       struct i2c_msg msgs[], int num )
{
	int i = 0;
	int ret = 0;
	int len;
	char* buf;
	unsigned long flags;
	int retry = 100;		/* allow only some retries */

	DEBUG_FN( DEBUG_MINOR );

	i2c_ns_int_status = I2C_NS_INT_OK;
	while( i < num ) {
		if( i2c_ns_int_status == I2C_NS_INT_RETRY ) {
			/* resend complete message, stop previous access first */
			/* i is already set to 0 due to while condition */
			ret=i2c_ns_send_cmd(i2c_adap,NS_I2C_DATA_TX_CMD_M_STOP);
			if( ret || !(retry--) ) {
				DEBUG_ARGS( DEBUG_ERROR,
					    "max. number of retries failed\n" );

				break;
			}
		}

		len = msgs[ i ].len;
		buf = msgs[ i ].buf;

		spin_lock_irqsave( &i2c_ns_lock, flags );
		/* being modified by int handler */
		i2c_ns_buf = buf;
		spin_unlock_irqrestore( &i2c_ns_lock, flags );

		if( !(msgs[ i ].flags & I2C_M_NOSTART ) ) {
			unsigned int cmd;

			/* set device address */
			writel( ((msgs[i].addr << NS_I2C_MASTER_MDA_SH) &
				 NS_I2C_MASTER_MDA_MA) |
				((msgs[i].flags&I2C_M_TEN)?NS_I2C_MASTER_MAM:0),
				NS_I2C_MASTER );

			if( msgs[ i ].flags & I2C_M_RD )
				cmd = NS_I2C_DATA_TX_CMD_M_READ;
			else {
				/* we need to send first byte with the command */
				cmd = NS_I2C_DATA_TX_CMD_M_WRITE |
					NS_I2C_DATA_TX_VAL |
					*buf;
				len--;
				buf++;
			}

			/* send start, device address and read/write command */
			ret = i2c_ns_send_cmd( i2c_adap, cmd );
			if( ret ) {
				if( i2c_ns_int_status == I2C_NS_INT_RETRY ) {
					/* try the message again */
					i = 0;
					continue;
				}

				DEBUG_ARGS( DEBUG_ERROR,
					    "Setting address failed for 0x%x "
					    "with errno %i\n",
					    msgs[ i ].addr, ret  );
				break;
			}
		}
		if( len > 0 ) {
			/* either read mode or send more than one byte */
			if( msgs[ i ].flags & I2C_M_RD )
				ret = i2c_ns_read( i2c_adap, len );
			else
				ret = i2c_ns_write( i2c_adap, buf, len );

			if( ret ) {
				if( i2c_ns_int_status == I2C_NS_INT_RETRY ) {
					/* try the message again */
					i = 0;
					continue;
				}

				DEBUG_ARGS( DEBUG_ERROR,
					    "Operation 0x%x failed for 0x%x "
					    "with errno %i\n",
					    msgs[ i ].flags,msgs[ i ].addr,ret);
				break;
			}
		}

		i++;
	}

	/* message block transmitted, send stop */
	if( i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_STOP ) ) {
		/* something locked up, maybe devices thinks lost clock. */
		if( i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_NOP ) ||
		    i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_STOP ) ) {
			unsigned long flags;

			/* maybe netsilicon's I2C state machine got stuck,
			   reset it. */
			spin_lock_irqsave( &ns_gpio_lock, flags );
			rmwl( NS_BBUS_MASTER_RESET, | NS_BBUS_MASTER_RESET_I2C );
			rmwl( NS_BBUS_MASTER_RESET, & ~NS_BBUS_MASTER_RESET_I2C);
			spin_unlock_irqrestore( &ns_gpio_lock, flags );
			writel( NS_I2C_CFG_INIT_VAL, NS_I2C_CFG );
		}
	}

	i2c_ns_buf = NULL;	/* maybe we get spurios interrupts */

	return ret ? 0 : i;
}

static u32 i2c_ns_master_func( struct i2c_adapter* adap )
{
	DEBUG_FN( DEBUG_MINOR );

	return I2C_FUNC_I2C | I2C_FUNC_10BIT_ADDR | I2C_FUNC_SMBUS_QUICK | I2C_FUNC_SMBUS_BYTE |
	    I2C_FUNC_SMBUS_BYTE_DATA | I2C_FUNC_SMBUS_WORD_DATA ;
}

/***********************************************************************
 * @Function: i2c_ns_int
 * @Return: IRQ_HANDLED and i2c_ns_int_status = ERROR or OK
 * @Descr: when data read, stores it in buffer pointed to by i2c_ns_buf. Buffer
 *         is not incremented.
 ***********************************************************************/

static irqreturn_t i2c_ns_int( int irq, void* dev_id,struct pt_regs* regs )
{
	int status;

	DEBUG_FN( DEBUG_INT );

	/* read and acknowledge (by read) */
	status = readl( NS_I2C_DATA_RX );

	switch( status & NS_I2C_DATA_RX_IRQCD_MA ) {
	    case NS_I2C_DATA_RX_IRQCD_M_RX_DATA:
		spin_lock( &i2c_ns_lock );
		if( i2c_ns_buf != NULL )
			/* be paranoid for spurios interrupts? */
			*i2c_ns_buf = status & NS_I2C_DATA_RX_DATA_MA;
		spin_unlock( &i2c_ns_lock );
		/* no break, will be done by M_CMD_ACK */

	    case NS_I2C_DATA_RX_IRQCD_M_CMD_ACK:
		i2c_ns_int_status = I2C_NS_INT_OK;
		break;

	    case NS_I2C_DATA_RX_IRQCD_M_TX_DATA:
		i2c_ns_int_status = I2C_NS_INT_OK;
		break;

	    case NS_I2C_DATA_RX_IRQCD_M_NO_ACK:
		/* maybe device didn't listen closely, try again */
		i2c_ns_int_status = I2C_NS_INT_RETRY;
		break;

	    default:
		DEBUG_ARGS( DEBUG_ERROR,
			    "interrupt failed with 0x%x\n", status );
		i2c_ns_int_status = I2C_NS_INT_ERROR;
	}

	wake_up_interruptible( &wait_q );

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: i2c_ns_read
 * @Return: <0 in case of failure otherwise 0
 * @Descr: Must be called after an M_READ has been executed and the
 *         interrupt has happened. Byte has already been stored by i2c_ns_int in
 *         i2c_ns_buf but not incremented. Don't do read command in i2c_ns_int
 *         so that we still use i2c_adap->timeout.
 ***********************************************************************/

static int i2c_ns_read( struct i2c_adapter* i2c_adap, int count )
{
	int ret = 0;
	unsigned long flags;

	DEBUG_FN( DEBUG_RX );

	/* first character has already been stored in buffer by interrupt
	 * handler */
	while( count-- > 1 ) {
		spin_lock_irqsave( &i2c_ns_lock, flags );
		i2c_ns_buf++;
		spin_unlock_irqrestore( &i2c_ns_lock, flags );

		ret = i2c_ns_send_cmd( i2c_adap, NS_I2C_DATA_TX_CMD_M_NOP );
		if( ret )
			return ret;
	}

	return 0;
}

/***********************************************************************
 * @Function: i2c_ns_write
 * @Return: <0 in case of failure otherwise 0
 * @Descr: Must be called after an M_WRITE has been executed and the
 *         first character has been sent already with M_WRITE.
 ***********************************************************************/

static int i2c_ns_write( struct i2c_adapter* i2c_adap, const char* buf,int count)
{
	int ret = 0;

	DEBUG_FN( DEBUG_TX );

	while( count-- ) {
		ret = i2c_ns_send_cmd( i2c_adap,
				       (*buf & NS_I2C_DATA_TX_DATA_MA ) |
				       NS_I2C_DATA_TX_CMD_M_NOP |
				       NS_I2C_DATA_TX_VAL );
		if( ret )
			return ret;

		buf++;
	}

	return ret;
}

/***********************************************************************
 * @Function: i2c_ns_send_cmd
 * @Return: -ETIMEDOUT if no interrupt happened in i2c_adap->timeout or -EIO on
 *          an I2C failure
 * @Descr: sends command and waits signal safe for I2C interrupt acknowleding
 *         command or timeout.
 ***********************************************************************/

static int i2c_ns_send_cmd( struct i2c_adapter* i2c_adap, unsigned int cmd )
{
	int ret = 0;

	DEBUG_FN( DEBUG_INT | DEBUG_MINOR );

	/* now wait. Don't use semaphores as these don't support timeouts */
	i2c_ns_int_status = I2C_NS_INT_AWAITING;
	do {
		writel( cmd, NS_I2C_DATA_TX );
		if( !wait_event_interruptible_timeout(
			    wait_q,
			    i2c_ns_int_status != I2C_NS_INT_AWAITING,
			    i2c_adap->timeout ))
		{
			DEBUG_ARGS( DEBUG_ERROR, "timed out\n" );

			/* not interrupted */
			ret = -ETIMEDOUT;
			break;
		}
	} while( i2c_ns_int_status == I2C_NS_INT_AWAITING );

	if( i2c_ns_int_status != I2C_NS_INT_OK )
		ret = -EIO;

	return ret;
}

static int __init i2c_ns_init( void )
{
	DEBUG_FN( DEBUG_INIT );

	printk( KERN_INFO DRIVER_NAME
		": I2C CC9P9xx0 driver $Revision: 1.8 $ initializing\n");

	if( request_mem_region( NS_I2C_MODULE_BASE,
				NS_I2C_CFG + 0x4 - NS_I2C_MODULE_BASE,
				DRIVER_NAME ) == NULL ) {
		printk( DRIVER_NAME ": memory not available\n" );
		goto error;
	}

	if( request_irq( IRQ_I2C, i2c_ns_int, 0, DRIVER_NAME, NULL ) ) {
		printk( DRIVER_NAME ": irq %i not available\n", IRQ_I2C );
		goto error_irq;
	}

#ifdef CONFIG_ARCH_NS9360
	rmwl( NS_BBUS_MASTER_RESET, & 0xFFFFFF7F );
	set_gpio_cfg( 70, NS_GPIO_CFG_FUNC_2 );
	set_gpio_cfg( 71, NS_GPIO_CFG_FUNC_2 );

#endif /* CONFIG_ARCH_NS9360 */

	/* no GPIOs and interrupts to configure, 9750 has fixed pins */
	writel( NS_I2C_CFG_INIT_VAL, NS_I2C_CFG );

	i2c_add_adapter( &i2c_ns_adapter );

	return 0;

error_irq:
	release_mem_region( NS_I2C_MODULE_BASE,
			    NS_I2C_CFG + 0x4 - NS_I2C_MODULE_BASE );
error:
	return -ENODEV;
}

static void i2c_ns_exit( void )
{
	DEBUG_FN( DEBUG_INIT );

	i2c_del_adapter( &i2c_ns_adapter );

	free_irq( IRQ_I2C, NULL );
	release_mem_region( NS_I2C_MODULE_BASE,
			    NS_I2C_CFG + 0x4 - NS_I2C_MODULE_BASE );
}

subsys_initcall( i2c_ns_init );
module_exit( i2c_ns_exit );

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>" );
MODULE_DESCRIPTION( "I2C-Bus for CC9P9xx0" );
MODULE_LICENSE( "GPL" );

module_param( debug, int, 0 );
MODULE_PARM_DESC( debug, "debug level" );
