/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: 24c64.c,v 1.5 2006-08-09 11:44:56 uzeisberger Exp $
 * @Author: Markus Pietrek
 * @Descr: 24C64 EEPROM Driver
 * @References: [1] M24C64 HW Reference 4578.pdf
 * @TODO: * writing with offsets (dd of=/dev/eeprom seek=1) does not work yet,
 *          but I think this is the I2C layer.
 *        * page accesses require starting offset already aligned. Could be
 *          improved for longer block reads to start page read when next page is
 *          reached independent of starting address alignment.
 *        * big endianess not supported yet
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
 /*  @History:
 *     2005/06/28 : JD: ported to kernel 2.6.12
 ***********************************************************************/
#include <linux/config.h>
#include <linux/module.h>	/* module_init */
#include <linux/init.h>		/* __init */
#include <linux/i2c.h>		/* i2c_algorithm */

#include <asm/semaphore.h>	/* init_MUTEX */

#define DEBUG_INIT		0x0001
#define DEBUG_MINOR		0x0002
#define DEBUG_RX		0x0004
#define DEBUG_TX		0x0008
#define DEBUG_ERROR		0x4000
#define DEBUG_ERROR_CRIT	0x8000

static int debug = 0xc000;

#define DEBUG_ARGS( FLG, a0, args... ) if( ( debug & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, ## args )
#define DEBUG_FN( FLG ) if( (debug & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
#define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }

#define DRIVER_NAME		"24c64"

#define EEPROM_SIZE		8192

static int addr = 0x50;		/* eeprom's device address */

static int m24c64_attach_adapter( struct i2c_adapter* adapter );
static int m24c64_detach_client( struct i2c_client* client );
static int m24c64_attach( struct i2c_adapter* adapter, int addr, int kind );
static ssize_t m24c64_read( struct kobject* kobj, char* buf, loff_t off,
			    size_t count );
static ssize_t m24c64_write( struct kobject* kobj, char* buf, loff_t off,
			     size_t count );
static ssize_t m24c64_read_bytes( struct i2c_client* client, char* buf,
				  loff_t off, size_t count );
static ssize_t m24c64_write_bytes( struct i2c_client* client, char* buf,
				   loff_t off, size_t count );

static int __init m24c64_init( void );
static void m24c64_exit( void );

static struct semaphore m24c64_lock; /* all read and write accesses */

static struct i2c_driver m24c64_driver = {
	.driver		= {
		.owner		= THIS_MODULE,
		.name		= DRIVER_NAME,
	},
	/* register for id @ http://secure.netroedge.com/~lm78/docs.html */
	.id		= I2C_DRIVERID_24C64,
	.attach_adapter	= m24c64_attach_adapter,
	.detach_client	= m24c64_detach_client,
};

static struct bin_attribute m24c64_attr = {
	.attr = {
		.name = "eeprom",
		.mode = S_IRUGO,
	},
	.size = EEPROM_SIZE,
	.read = m24c64_read,
	.write = m24c64_write
};

static int m24c64_attach_adapter( struct i2c_adapter* adapter )
{
	DEBUG_FN( DEBUG_INIT );

	/* no SMBus, can't do probing */
	m24c64_attach( adapter, addr, 0 );
	return 0;
}

static int m24c64_detach_client( struct i2c_client* client )
{
	int res;

	DEBUG_FN( DEBUG_INIT );

	if( ( res = i2c_detach_client( client ) ) )
		return res;

	kfree( client );

	return 0;
}

static int m24c64_attach( struct i2c_adapter* adapter, int addr, int kind )
{
	struct i2c_client* client;
	int res;

	DEBUG_FN( DEBUG_INIT );

	if( ( client = (struct i2c_client *) kmalloc( sizeof(struct i2c_client),
						      GFP_KERNEL ) ) == NULL )
		return -ENOMEM;

	memset( client, 0x0, sizeof( struct i2c_client ) );

	strcpy( client->name, m24c64_driver.driver.name ); /*DRIVER_NAME<I2C_NAME_SIZE*/
	/*client->id = m24c64_driver.id; */
	client->addr = addr;
	client->adapter = adapter;
	client->driver = &m24c64_driver;

	init_MUTEX( &m24c64_lock );

	if( ( res = i2c_attach_client( client ) ) ) {
		kfree( client );
		return res;
	}

	sysfs_create_bin_file( &client->dev.kobj, &m24c64_attr );

	return 0;
}

/***********************************************************************
 * @Function: m24c64_read
 * @Return: number of bytes read or -EIO for I2C failure
 * @Descr: If possible, uses page accesses to read from eeprom. Otherwise
 *         performs random accesses.
 ***********************************************************************/

static ssize_t m24c64_read( struct kobject* kobj, char* buf, loff_t off,
			    size_t count )
{
	struct i2c_client* client = to_i2c_client(
		container_of( kobj, struct device, kobj) );
	ssize_t overall = 0;
	ssize_t trans = 0;
	ssize_t trans_in;

	if (off > EEPROM_SIZE)
		return 0;
	if (off + count > EEPROM_SIZE)
		count = EEPROM_SIZE - off;
	while( count > overall ) {
		/* page read when 32bit aligned */
		trans_in = ((off % 32) ? 1 : ((count > 32) ? 32 : count));

		trans = m24c64_read_bytes( client, buf, off, trans_in );
		if( trans != trans_in )
			return -EIO;

		overall += trans;

		/* next page */
		buf += trans;
		off += trans;
	}

	return count; //overall;
}

/***********************************************************************
 * @Function: m24c64_write
 * @Return: number of bytes written or -EIO for I2C failure
 * @Descr: If possible, uses page accesses to write to eeprom. Otherwise
 *         performs random accesses.
 ***********************************************************************/

static ssize_t m24c64_write( struct kobject* kobj, char* buf, loff_t off,
			     size_t count )
{
	struct i2c_client* client = to_i2c_client(
		container_of( kobj, struct device, kobj) );

	ssize_t overall = 0;
	ssize_t trans = 0;
	ssize_t trans_in;

	if (off > EEPROM_SIZE)
		return 0;
	if (off + count > EEPROM_SIZE)
		count = EEPROM_SIZE - off;
	while( count > overall ) {
		/* page read when 32bit aligned */
		trans_in = ((off % 32) ? 1 : ((count > 32) ? 32 : count));

		trans = m24c64_write_bytes( client, buf, off, trans_in );

		if( trans != trans_in )
			return -EIO;

		overall += trans;

		/* next page */
		buf += trans;
		off += trans;
	}

	return count; //overall;
}

/***********************************************************************
 * @Function: m24c64_read_bytes
 * @Return: # bytes read, -EINVAL for wrong size or offset or -EIO on I2C failure
 * @Descr:
 ***********************************************************************/

static ssize_t m24c64_read_bytes( struct i2c_client* client, char* buf,
				  loff_t off, size_t count )
{
	/* @TODO: not big endian safe */
	struct i2c_msg msg[] = {
                {addr:client->addr,flags:0,len: 2, buf:(char*)&off },
                {addr:client->addr,flags:I2C_M_RD,len:count,buf:buf}
        };
	int len = sizeof( msg ) / sizeof( struct i2c_msg );

	DEBUG_FN( DEBUG_RX | DEBUG_MINOR );

	if( off + count > EEPROM_SIZE ) {
        	printk( KERN_ERR "%s EEPROM read out of area: %i + %i > %i\n",
			__FUNCTION__, (int) off, count, EEPROM_SIZE );
		return -EINVAL;
	}

	/* highest address byte first on wire */
	off = ((off & 0xff) << 8) | (off >> 8);
        if( i2c_transfer( client->adapter, msg, len ) != len )
                return -EIO;

	return count;
}

static ssize_t m24c64_write_bytes( struct i2c_client* client, char* buf,
				   loff_t off, size_t count )
{
	/* @TODO: not big endian safe */
	struct i2c_msg msg[] = {
                {addr:client->addr,flags:0,len: 2, buf:(char*)&off },
                {addr:client->addr,flags:I2C_M_NOSTART,len: count, buf: buf}
        };
	int len = sizeof( msg ) / sizeof( struct i2c_msg );

	DEBUG_FN( DEBUG_TX | DEBUG_MINOR );

	if( off + count > EEPROM_SIZE ) {
        	printk( KERN_ERR "%s EEPROM write out of area: %i + %i > %i\n",
			__FUNCTION__, (int) off, count, EEPROM_SIZE );
		return -EINVAL;
	}

	/* highest address byte first on wire */
	off = ((off & 0xff) << 8) | (off >> 8);
        if( i2c_transfer( client->adapter, msg, len ) != len )
                return -EIO;

	return count;
}

static int __init m24c64_init( void )
{
	DEBUG_FN( DEBUG_INIT );

	printk( KERN_INFO DRIVER_NAME
		": EEPROM driver $Revision: 1.5 $ initializing\n");

	return i2c_add_driver( &m24c64_driver );
}

static void m24c64_exit( void )
{
	DEBUG_FN( DEBUG_INIT );

	i2c_del_driver( &m24c64_driver );
}

device_initcall( m24c64_init );
module_exit( m24c64_exit );

MODULE_AUTHOR("Markus Pietrek <mpietrek@fsforth.de>");
MODULE_DESCRIPTION("EEPROM 24C64 driver");
MODULE_LICENSE("GPL");

module_param( addr, int, 0 );
MODULE_PARM_DESC( addr, "eeprom device addr");

module_param( debug, int, 0 );
MODULE_PARM_DESC( debug, "debug level");
