/*
 * drivers/i2c/chips/ds1337.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
/*
 *  !Revision:  $Revision: 1.10 $
 *  !Author:    Pedro Perez de Heredia
 *  !Descr:     Provides Access to Date/Time via a DS1337 RTC connected to
 *              the I2C bus. Based on ds1307.c by Intrinsyc Software Inc
 *              and some other rtc drivers like s3c2410-rtc.c.
 *  !References: [1] Dallas/Maxim Datasheet DS1337/Serial RTC
 *  !History:
 *     2005/04/01 : PPH : Initial version.
 *     2005/08/02 : PPH : Fixed initialization problem.
 *                        Added Alarm support.
 *     2007/04/30 : PPH : Ported to 2.6.17.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/string.h>
#include <linux/rtc.h>		/* get the user-level API */
#include <linux/bcd.h>
#include <linux/list.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/interrupt.h>

#include <asm/rtc.h>
#include <asm/mach-types.h>


#define	I2C_RTC_DS1337_ADDR	0x68
#define DRV_NAME		"ds1337"

/* Configure */
#if !defined(CONFIG_DS1337_ALARM_IRQ)
# define	DS1337_ALARM_IRQ	NO_IRQ
#else
# define	DS1337_ALARM_IRQ	CONFIG_DS1337_ALARM_IRQ_NR
#endif

/* these addresses are for reading */
#define DS1337_ADDR_SEC		0x00
#define DS1337_ADDR_MIN		0x01
#define DS1337_ADDR_HOUR	0x02
#define DS1337_ADDR_DAY		0x03
#define DS1337_ADDR_DATE	0x04
#define DS1337_ADDR_MONTH	0x05
#define DS1337_ADDR_YEAR	0x06
#define DS1337_ADDR_AL1_SEC	0x07
#define DS1337_ADDR_AL1_MIN	0x08
#define DS1337_ADDR_AL1_HOUR	0x09
#define DS1337_ADDR_AL1_DAY	0x0a
#define DS1337_ADDR_AL1_DATE	0x0a
#define DS1337_ADDR_CTRL	0x0e
#define DS1337_ADDR_STAT	0x0f
#define DS1337_NUM_REGS		0x10

/* some register flags */
#define DS1337_CTL_BIT_RS2	0x10
#define DS1337_CTL_BIT_RS1	0x08
#define DS1337_CTL_BIT_A2IE	0x02	/* Alarm 2 Interrupt Enable */
#define DS1337_CTL_BIT_A1IE	0x01	/* Alarm 1 Interrupt Enable */
#define DS1337_CTL_BIT_DOSC	0x80	/* Disable oscilator */
#define DS1337_STAT_BIT_OSF	0x80	/* Oscilator frequency */
#define DS1337_STAT_BIT_A2F	0x01	/* Alarm 2 flag */
#define DS1337_STAT_BIT_A1F	0x01	/* Alarm 1 flag */

/* Taken from mc146818rtc.h */
#define RTC_IRQF		0x80
#define RTC_PF			0x40
#define RTC_AF			0x20
#define RTC_UF			0x10

static unsigned short normal_i2c[] = { I2C_RTC_DS1337_ADDR, I2C_CLIENT_END };

I2C_CLIENT_INSMOD_1(ds1337);

/* Function prototypes */
static int ds1337_proc( char *buf );
static int ds1337_open( void );
static void ds1337_release( void );
static int ds1337_ioctl( unsigned int cmd, unsigned long arg );
static void ds1337_reg_to_time( struct rtc_time *dt, char *buf );
static void ds1337_alarm_reg_to_time( struct rtc_time *dt, char *buf );
static int ds1337_read_time( struct rtc_time * tm );
static int ds1337_set_time( struct rtc_time * tm );
static int ds1337_read_alarm( struct rtc_wkalrm *alarm );
static int ds1337_set_alarm( struct rtc_wkalrm *alarm );
static int ds1337_set_alarm_irq( int enable );
static int ds1337_ack_alarm_irq( void );
static int ds1337_attach_adapter( struct i2c_adapter *adapter );
static int ds1337_detect( struct i2c_adapter *adapter, int address, int kind );
static void ds1337_init_client( struct i2c_client *client );
static int ds1337_detach_client( struct i2c_client *client );


static struct i2c_driver ds1337_driver = {
	.driver = {
		.name	= "ds1337",
	},
	.attach_adapter	= ds1337_attach_adapter,
	.detach_client	= ds1337_detach_client,
};

struct ds1337_data {
	struct i2c_client client;
	struct list_head list;
	int id;
};

static struct rtc_ops ds1337_ops = {
	.owner		= THIS_MODULE,
	.open		= ds1337_open,
	.release	= ds1337_release,
	.ioctl		= ds1337_ioctl,
	.read_time	= ds1337_read_time,
	.set_time	= ds1337_set_time,
	.read_alarm	= ds1337_read_alarm,
	.set_alarm	= ds1337_set_alarm,
	.proc		= ds1337_proc,
};

/* Local variables */
static int ds1337_id;
static int ds1337_alarm_irqno = NO_IRQ;
static LIST_HEAD(ds1337_clients);
static struct i2c_client *rtc_client = NULL;

#ifdef DEBUG
/***********************************************************************
 * @Function: ds1337_dump_regs
 * @Return: Nothing.
 * @Descr: displays the rtc registers.
 ***********************************************************************/
static void ds1337_dump_regs( void )
{
	struct i2c_client *client = rtc_client;
	int reg;
	s32 data;

	for( reg = 0; reg < DS1337_NUM_REGS; reg++ ) {
		data = i2c_smbus_read_byte_data( client, DS1337_ADDR_SEC + reg );
		if( data < 0 )
			printk( KERN_ERR DRV_NAME ": %s failed\n", __FUNCTION__ );
		else
			printk( "Reg[%d] = 0x%02x\n", reg, (u8)data );
	}
}
#else
# define ds1337_dump_regs()
#endif

/***********************************************************************
 * @Function: ds1337_alarm_irq
 * @Return: IRQ_HANDLED.
 * @Descr: rtc alarm interrupt handler.
 ***********************************************************************/
static irqreturn_t ds1337_alarm_irq( int irq, void *id, struct pt_regs *r )
{
	rtc_update( 1, RTC_AF | RTC_IRQF );
	return IRQ_HANDLED;
}


/***********************************************************************
 * @Function: ds1337_proc
 * @Return: error code.
 * @Descr: @TODO.
 ***********************************************************************/
static int ds1337_proc( char *buf )
{
	return 0;
}

/***********************************************************************
 * @Function: ds1337_release
 * @Return: nothing.
 * @Descr: Close system call.
 ***********************************************************************/
static int ds1337_open( void )
{
	struct i2c_client *client = rtc_client;
	struct ds1337_data *data;
	s32 reg;
	int ret;

	if( client == NULL )
		return -ENODEV;

	data = i2c_get_clientdata( client );

	if( data == NULL )
		return -ENODEV;

	/* Read RTC control register and check some known values to ensure
	 * that we have a device connected.
	 * Additionaly, check that the oscillator is running */
	reg = i2c_smbus_read_byte_data( client, DS1337_ADDR_CTRL );
	if( reg < 0 ){
		printk( KERN_ERR DRV_NAME ": %s failed\n", __FUNCTION__ );
		return -ENODEV;
	} else if( reg & 0x60 ) {
		printk( KERN_ERR DRV_NAME ": %s wrong CTRL Reg (0x%02x)\n",
		        __FUNCTION__, (u8)reg );
		return -ENODEV;
	} else if( reg & 0x80 ) {
		printk( KERN_WARNING DRV_NAME ": %s Oscillator was disabled, CTRL Reg (0x%02x)\n",
		        __FUNCTION__, (u8)reg );

		reg &= ~DS1337_CTL_BIT_DOSC;
		return i2c_smbus_write_byte_data( client, DS1337_ADDR_CTRL, (u8)reg );
	}

	if( ds1337_alarm_irqno != NO_IRQ ) {
		if( ( ret = request_irq( ds1337_alarm_irqno, ds1337_alarm_irq,
	                                 SA_INTERRUPT, "ds1337-rtc alarm", NULL ) ) != 0 ) {
			printk( KERN_ERR  "Unable to get irq %d\n", ds1337_alarm_irqno );
			return ret;
		}
	}

	return 0;
}

/***********************************************************************
 * @Function: ds1337_release
 * @Return: nothing.
 * @Descr: Close system call.
 ***********************************************************************/
static void ds1337_release( void )
{
	if( ds1337_alarm_irqno != NO_IRQ )
		free_irq( ds1337_alarm_irqno, NULL );
	return;
}

/***********************************************************************
 * @Function: ds1337_ioctl
 * @Return: error code.
 * @Descr: ioclt function
 ***********************************************************************/
static int ds1337_ioctl( unsigned int cmd, unsigned long arg )
{
	int ret = -EINVAL;

	switch( cmd ) {
		case RTC_AIE_OFF:	/* Alarm interrupt enable on/off */
		case RTC_AIE_ON:
			ret = ds1337_set_alarm_irq( ( cmd == RTC_AIE_ON ) ? 1 : 0 );
			break;

		case RTC_UIE_ON:	/* Update interrupt enable on/off */
		case RTC_UIE_OFF:
		case RTC_PIE_OFF:	/* Periodic interrupt enable on/off */
		case RTC_PIE_ON:
		case RTC_IRQP_READ:	/* Read and set irq rate */
		case RTC_IRQP_SET:
		default:
			ret = -EINVAL;	break;
	}
	return ret;
}

/***********************************************************************
 * @Function: ds1337_reg_to_time
 * @Return: nothing.
 * @Descr: Converts from rtc register time to rtc_time vals.
 ***********************************************************************/
static void ds1337_reg_to_time( struct rtc_time *dt, char *buf )
{
	dt->tm_sec = BCD2BIN( buf[DS1337_ADDR_SEC] );
	dt->tm_min = BCD2BIN( buf[DS1337_ADDR_MIN] );
	/* assume 24 hour time as this is the format we write */
	dt->tm_hour = BCD2BIN( buf[DS1337_ADDR_HOUR] );
	dt->tm_wday = (buf[DS1337_ADDR_DAY] - 1) & 0x07;
	dt->tm_mday = BCD2BIN( buf[DS1337_ADDR_DATE] );
	dt->tm_mon = ( BCD2BIN( buf[DS1337_ADDR_MONTH] & 0x1f )) - 1;
	dt->tm_year = BCD2BIN( buf[DS1337_ADDR_YEAR] );
	if (buf[DS1337_ADDR_MONTH] & 0x80)
		dt->tm_year += 100;
}

/***********************************************************************
 * @Function: ds1337_alarm_reg_to_time
 * @Return: nothing.
 * @Descr: Converts from rtc alarm register time to rtc_time vals.
 ***********************************************************************/
static void ds1337_alarm_reg_to_time( struct rtc_time *dt, char *buf )
{
	dt->tm_sec = buf[0] & 0x80 ? 0xff : BCD2BIN( buf[0] & 0x7f );
	dt->tm_min = buf[1] & 0x80 ? 0xff : BCD2BIN( buf[1] & 0x7f );
	/* assume 24 hour time as this is the format we write
	 * assume also that the alarm was programmed with date */
	dt->tm_hour = buf[2] & 0x80 ? 0xff : BCD2BIN( buf[2] & 0x7f );
	dt->tm_mday = 0xff;
	dt->tm_mon = 0xff;
	dt->tm_year = 0xffff;
}

/***********************************************************************
 * @Function: ds1337_read_time
 * @Return: 0 on success, error code otherwise.
 * @Descr: Sets tm time values into the rtc.
 ***********************************************************************/
static int ds1337_read_time( struct rtc_time * tm )
{
	struct i2c_client *client = rtc_client;
	u8 buf[DS1337_NUM_REGS], address[1] = { 0 };
	int ret = 0;
	struct i2c_msg msgs[2] = {
		{ client->addr, 0, 1, address },
		{ client->addr, I2C_M_RD, DS1337_NUM_REGS, buf }
	};

	memset( buf, 0, sizeof( buf ) );

	if( i2c_transfer( client->adapter, msgs, 2 ) == 2 )
	        ds1337_reg_to_time( tm, buf );
	else {
		printk( KERN_ERR DRV_NAME ": %s failed\n", __FUNCTION__ );
		ret = -1;
	}
	return ret;
}

/***********************************************************************
 * @Function: ds1337_set_time
 * @Return: 0 on success, error code otherwise.
 * @Descr: Sets tm time values into the rtc.
 ***********************************************************************/
static int ds1337_set_time( struct rtc_time * tm )
{
	struct i2c_client *client = rtc_client;
	u8 buf[DS1337_NUM_REGS];

	buf[ 0 ] = 0;	/* Base address reg */
	buf[DS1337_ADDR_SEC + 1] = BIN2BCD( tm->tm_sec );
	buf[DS1337_ADDR_MIN + 1] = BIN2BCD( tm->tm_min );
	buf[DS1337_ADDR_HOUR + 1] = BIN2BCD( tm->tm_hour );
	buf[DS1337_ADDR_DAY + 1] = (tm->tm_wday + 1) & 0x07;
	buf[DS1337_ADDR_DATE + 1] = BIN2BCD( tm->tm_mday );
	buf[DS1337_ADDR_MONTH + 1] = BIN2BCD( tm->tm_mon + 1 );
	buf[DS1337_ADDR_YEAR + 1] = BIN2BCD( tm->tm_year);
	if (tm->tm_year >= 100) {
		buf[DS1337_ADDR_YEAR + 1] = BIN2BCD( tm->tm_year - 100 );
		buf[DS1337_ADDR_MONTH + 1] |= 0x80;
	}

	if( i2c_master_send( client, (char *)buf, 8 ) != 8 ) {
		printk( KERN_ERR DRV_NAME ": %s failed\n", __FUNCTION__ );
		return -EIO;
	}

	return 0;
}

/***********************************************************************
 * @Function: ds1337_read_alarm
 * @Return: 0 on success, -1 otherwise.
 * @Descr: Reads the programed alarm value into the rtc.
 ***********************************************************************/
static int ds1337_read_alarm( struct rtc_wkalrm *alarm )
{
	u8 buf[4], address[1] = { DS1337_ADDR_AL1_SEC };
	int ret = 0;
	struct rtc_time *alarm_time = &alarm->time;
	struct i2c_client *client = rtc_client;
	struct i2c_msg msgs[2] = {
		{ client->addr, 0, 1, address },
		{ client->addr, I2C_M_RD, 4, buf }
	};

	memset( buf, 0, sizeof( buf ) );

	if( i2c_transfer( client->adapter, msgs, 2 ) == 2 )
	       ds1337_alarm_reg_to_time( alarm_time, buf );
	else {
		printk( KERN_ERR DRV_NAME ": %s failed\n", __FUNCTION__ );
		ret = -1;
	}
	return ret;
}

/***********************************************************************
 * @Function: ds1337_set_alarm
 * @Return: 0 on success, error code otherwise.
 * @Descr: Sets the alarm time values into the rtc ALARM1.
 ***********************************************************************/
static int ds1337_set_alarm( struct rtc_wkalrm *alarm )
{
	struct i2c_client *client = rtc_client;
	struct rtc_time *alarm_time = &alarm->time;
	u8 buf[5];

	buf[0] = DS1337_ADDR_AL1_SEC;	/* Base address reg */
	buf[1] = 0x7f & BIN2BCD( alarm_time->tm_sec );
	buf[2] = 0x7f & BIN2BCD( alarm_time->tm_min );
	buf[3] = 0x7f & BIN2BCD( alarm_time->tm_hour );
	buf[4] = 0x80;

	if( i2c_master_send( client, (char *)buf, 5 ) != 5 ) {
		printk( KERN_ERR DRV_NAME ": %s failed\n", __FUNCTION__ );
		return -EIO;
	}
	return 0;
}

/***********************************************************************
 * @Function: ds1337_ack_alarm_irq
 * @Return: 0 on success, error code otherwise.
 * @Descr: Acks the rtc alarm irq.
 ***********************************************************************/
static int ds1337_ack_alarm_irq( void )
{
	struct i2c_client *client = rtc_client;

	return i2c_smbus_write_byte_data( client, DS1337_ADDR_STAT, 0xfe );
}

/***********************************************************************
 * @Function: ds1337_set_alarm_irq
 * @Return: 0 on success, error code otherwise.
 * @Descr: Enables/disables the alarm irq.
 ***********************************************************************/
static int ds1337_set_alarm_irq( int enable )
{
	struct i2c_client *client = rtc_client;
	s32 reg;

	if( ds1337_alarm_irqno == NO_IRQ )
		return -EINVAL;

	if( ( reg = i2c_smbus_read_byte_data( client, DS1337_ADDR_CTRL ) ) < 0 )
		return -EIO;

	if( enable ) {
		/* Acknoledge pending interrupts before enable it */
		if( ds1337_ack_alarm_irq() ) {
			printk( KERN_WARNING DRV_NAME ": %s unable to ack alarm interrupt\n",
		        __FUNCTION__ );
			return -EIO;
		}
		reg |= DS1337_CTL_BIT_A1IE;
		enable_irq_wake( ds1337_alarm_irqno );
	} else {
		reg &= ~DS1337_CTL_BIT_A1IE;
		disable_irq_wake( ds1337_alarm_irqno );
	}

	return i2c_smbus_write_byte_data( client, DS1337_ADDR_CTRL, (u8)reg );
}

/***********************************************************************
 * @Function: ds1337_attach_adapter
 * @Return: 0 on success, error code otherwise.
 * @Descr: Attach the i2c adapter.
 ***********************************************************************/
static int ds1337_attach_adapter( struct i2c_adapter *adapter )
{
	return i2c_probe( adapter, &addr_data, ds1337_detect );
}

/***********************************************************************
 * @Function: ds1337_detect
 * @Return: 0 on success, error code otherwise.
 * @Descr: Tries to detect a DS1337 connected to the i2c bus.
 ***********************************************************************/
static int ds1337_detect( struct i2c_adapter *adapter, int address, int kind )
{
	struct i2c_client *new_client;
	struct ds1337_data *data;
	int err = 0;
	const char *name = "";

	if( !i2c_check_functionality( adapter, I2C_FUNC_SMBUS_BYTE_DATA ))
		goto exit;

	if( !( data = kmalloc(sizeof( struct ds1337_data ), GFP_KERNEL ) ) ) {
		err = -ENOMEM;
		goto exit;
	}
	memset( data, 0, sizeof(struct ds1337_data) );
	INIT_LIST_HEAD( &data->list );

	new_client = &data->client;
	i2c_set_clientdata( new_client, data );
	new_client->addr = address;
	new_client->adapter = adapter;
	new_client->driver = &ds1337_driver;
	new_client->flags = 0;

	/*
	 * Now we do the remaining detection. A negative kind means that
	 * the driver was loaded with no force parameter (default), so we
	 * must both detect and identify the chip. A zero kind means that
	 * the driver was loaded with the force parameter, the detection
	 * step shall be skipped. A positive kind means that the driver
	 * was loaded with the force parameter and a given kind of chip is
	 * requested, so both the detection and the identification steps
	 * are skipped.
	 */

	/* Default to an DS1337 if forced */
	if( kind == 0 )
		kind = ds1337;

	if( kind < 0 ) {
		/* detection and identification */
		u8 buf[DS1337_NUM_REGS];
		int reg;
		int data;

		/* Check that all DS1337 registers are present */
		for( reg = 0; reg < DS1337_NUM_REGS; reg++ ) {
			data = i2c_smbus_read_byte_data( new_client, DS1337_ADDR_SEC + reg );
			if( data < 0 )
				goto exit_free;
			buf[reg] = (u8)data;
		}

		/* Check that control register bits 6-5 are zero and also status bits 6-2 */
		if( (buf[DS1337_ADDR_CTRL] & 0x60) || (buf[DS1337_ADDR_STAT] & 0x7c) )
			goto exit_free;

		kind = ds1337;
	}

	if( kind == ds1337 )
		name = "ds1337";

	strlcpy( new_client->name, name, I2C_NAME_SIZE );

	if( ( err = i2c_attach_client( new_client ) ) )
		goto exit_free;

	ds1337_init_client( new_client );
	data->id = ds1337_id++;
	list_add( &data->list, &ds1337_clients );

	rtc_client = new_client;

	if( ( ds1337_alarm_irqno = DS1337_ALARM_IRQ ) != NO_IRQ )
		set_irq_type( ds1337_alarm_irqno, IRQT_FALLING );

	return register_rtc( &ds1337_ops );

exit_free:
	kfree( data );
exit:
	return err;
}

/***********************************************************************
 * @Function: ds1337_init_client
 * @Return: nothing.
 * @Descr: Client initialization.
 ***********************************************************************/
static void ds1337_init_client( struct i2c_client *client )
{
	s32 reg;

	/* Read RTC control register and check that the oscillator is running
	 * if not, enable it */
	reg = i2c_smbus_read_byte_data( client, DS1337_ADDR_CTRL );
	if( reg & 0x80 ) {
		printk( KERN_WARNING DRV_NAME ": %s Oscillator was disabled, CTRL Reg (0x%02x)\n",
		        __FUNCTION__, (u8)reg );

		reg &= ~DS1337_CTL_BIT_DOSC;
		(void)i2c_smbus_write_byte_data( client, DS1337_ADDR_CTRL, (u8)reg );
	}
}

/***********************************************************************
 * @Function: ds1337_detach_client
 * @Return: nothing.
 * @Descr: Client initialization.
 ***********************************************************************/
static int ds1337_detach_client( struct i2c_client *client )
{
	int err;
	struct ds1337_data *data = i2c_get_clientdata( client );

	unregister_rtc( &ds1337_ops );

	if( ( err = i2c_detach_client( client ) ) ) {
		printk( KERN_ERR DRV_NAME ": %s unable to detach client (err=%d)\n",
		         __FUNCTION__, err );
		return err;
	}

	list_del( &data->list );
	kfree( data );
	return 0;
}

/***********************************************************************
 * @Function: ds1337_init
 * @Return: 0 on success, error code otherwise.
 * @Descr: Performs driver load.
 ***********************************************************************/
static int __init ds1337_init( void )
{
	return i2c_add_driver( &ds1337_driver );
}

/***********************************************************************
 * @Function: ds1337_exit
 * @Return: nothing.
 * @Descr: Removes allocated resources.
 ***********************************************************************/
static void __exit ds1337_exit(void)
{
	i2c_del_driver( &ds1337_driver );
}

MODULE_AUTHOR("Digi International Inc");
MODULE_DESCRIPTION("DS1337 RTC driver");
MODULE_LICENSE("GPL");

module_init(ds1337_init);
module_exit(ds1337_exit);
