/************************************************************************
 *
 * Copyright (C) 2005 by FS Forth-Systeme GmbH && Sistemas Embebidos S.A.
 * All rights reserved.
 *
 * $Id: unc90-ide.c,v 1.1 2005-09-27 19:17:33 pperez Exp $
 * @Author: Pedro Perez de Heredia 
 * @Descr: IDE driver for the CompactFlash inteface for the UNC90 module.
 *         Based on bast-ide.c and on a previous driver version written 
 *         for the UNC20 module.
 * @References: [1] Compact Flash Interface for UNC20/UNC90 specification.
 *                  Version 1.0.
 * @TODO:    * Activate machine checking.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *  
 ************************************************************************/
 
 
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/blkdev.h>
#include <linux/errno.h>
#include <linux/ide.h>
#include <linux/init.h>

#include <asm/mach-types.h>
#include <asm/io.h>
#include <asm/arch/arch-unc90.h>	// UNC90_CS3_BASE_ADDR
#include <asm/arch/unc90utils.h>	// unc90_config_smc_cs
#include <asm/arch/irqs.h>		// AT91C_ID_IRQ0
#include <linux/delay.h>		// udelay, mdelay

#define DRV_PREFIX 			"ide-unc90"
#define	UNC90_MAX_IDE_HWIF		1
#define CF_SHIFT_ADDR			0 
#define	UNCBASCF_DEV_KIT_IDE_IRQ	AT91C_ID_IRQ0
#define	CF_CPLD_BASE_ADDR		UNC90_CS3_BASE_ADDR

// CPLD internal addresses (Base address is CS3) 
#define	CF_CPLD_CONF_REG_OF		0x200
#define	CF_CPLD_LOGIC_REV_OF		0x201
#define	CF_CPLD_LOGIC_MEM_LEN		0x1fff
#define	CF_CPLD_HWIF_OFFSET		0x2000
#define CF_CONFIG_OPTION_REG_OF		0x200	// Compact Flash Storage card Configuration Opt Reg

// CF configuration register flags
#define CF_A9_MASK			0x01
#define CF_A10_MASK			0x02
#define CF_CD_MASK			0x08
#define CF_RST_MASK			0x10
#define CF_REG_MASK			0x20
#define CF_MODE_MASK			0x80

// Common Memory/IO internal registers for Storage CF+ cards
#define CF_DATA_OFFSET			(0x00 << CF_SHIFT_ADDR)	// Data
#define CF_ERROR_OFFSET			(0x01 << CF_SHIFT_ADDR)	// Error
#define CF_NSECTOR_OFFSET		(0x02 << CF_SHIFT_ADDR)	// Number of sectors
#define CF_SECTOR_OFFSET		(0x03 << CF_SHIFT_ADDR)	// Starting sector
#define CF_LCYL_OFFSET			(0x04 << CF_SHIFT_ADDR)	// Low Cylinder
#define CF_HCYL_OFFSET			(0x05 << CF_SHIFT_ADDR)	// High Cylinder
#define CF_SELECT_OFFSET		(0x06 << CF_SHIFT_ADDR)	// Select
#define CF_STATUS_OFFSET		(0x07 << CF_SHIFT_ADDR)	// Status
#define CF_CTRL_OFFSET			(0x0e << CF_SHIFT_ADDR)	// Control
#define CF_DATA_DUP_OFFSET		(0x08 << CF_SHIFT_ADDR)	// Duplicate Data Register


/* list of registered interfaces */
static ide_hwif_t *hwifs[UNC90_MAX_IDE_HWIF];

static int __init
unc90_ide_register( unsigned int base, unsigned int aux, int irq,
		 ide_hwif_t **hwif)
{
	hw_regs_t hw;

	memset( &hw, 0, sizeof(hw) );

	// Configure addresses of IO registers
	hw.io_ports[IDE_ERROR_OFFSET] = (unsigned long)( base + CF_ERROR_OFFSET );	// Error
	hw.io_ports[IDE_NSECTOR_OFFSET] = (unsigned long)( base + CF_NSECTOR_OFFSET );	// Sector Count
	hw.io_ports[IDE_SECTOR_OFFSET] = (unsigned long)( base + CF_SECTOR_OFFSET );	// Sector No
	hw.io_ports[IDE_LCYL_OFFSET] = (unsigned long)( base + CF_LCYL_OFFSET );	// Low Cylinder
	hw.io_ports[IDE_HCYL_OFFSET] = (unsigned long)( base + CF_HCYL_OFFSET );	// High Cylinder
	hw.io_ports[IDE_SELECT_OFFSET] = (unsigned long)( base + CF_SELECT_OFFSET );	// Select
	hw.io_ports[IDE_STATUS_OFFSET] = (unsigned long)( base + CF_STATUS_OFFSET );	// Status
	hw.io_ports[IDE_CONTROL_OFFSET] = (unsigned long)( base + CF_CTRL_OFFSET );	// Control
	
	// Use duplicate data register instead of data register
	hw.io_ports[IDE_DATA_OFFSET] = (unsigned long)( base + CF_DATA_DUP_OFFSET );	// Data
	
	hw.irq = irq;

	ide_register_hw( &hw, hwif );
	
	return 0;
}


static int __init unc90_ide_init(void)
{
	unsigned long base_addr;
	unsigned short ctrl_reg;
	unsigned char logic_rev;
	int i;
	
	// TODO activate following
	//if( !(machine_is_unc90() )
	//    return 0;
	
	// Perform initial hardware configuration
	
	// Configure the CS3 signal to 8 bit
	//if( unc90_config_smc_cs( 3, 16, 1, 0, 0, 2, 0, 0, 0, 0 ) != 0 )
	if( unc90_config_smc_cs( 3, 16, 1, 6, 0, 2, 0, 0, 2, 2 ) != 0 )
	//if( unc90_config_smc_cs( 3, 16, 1, 0, 0, 2, 0, 0, 0, 0 ) != 0 )
	{
	    printk( KERN_ERR "Unable to configure CS3 signal\n" );
            return -EIO;
	}
	
	// Remap physical to virtual mem 
	base_addr = (unsigned long)ioremap( CF_CPLD_BASE_ADDR, CF_CPLD_LOGIC_MEM_LEN );
	
	logic_rev = readb( base_addr + CF_CPLD_LOGIC_REV_OF );
	printk( DRV_PREFIX ": Logic Revison: %d.%d\n", logic_rev >> 4, logic_rev & 0x0f );
	
	// Configure IRQ line
	// Configure PB29 a IRQ level triggered.
	
	//if( at91_set_a_periph( 61, 0 ) )
	//	printk( "What is going on?\n" );
	
	for( i = 0; i < UNC90_MAX_IDE_HWIF; i++ )
	{
	    // Detect if a CF is inserted on the socket
	    ctrl_reg = readb( base_addr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * i );
	    
	    if( ( ctrl_reg & CF_CD_MASK ) != CF_CD_MASK )
	    {
	        printk( DRV_PREFIX ": CF detected\n" );
	        
	        // Reset CF
	        writeb( CF_RST_MASK, base_addr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * i );
	        mdelay(1);
	        writeb( 0x00, base_addr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * i );
	        mdelay(10);	// @TODO review following delay...
	        
	        // Set A9 address line to access to the configuration option register of the CF
	        // which in a storage CF is located at address 0x200
	        writeb( CF_A9_MASK, base_addr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * i );
	
	        // Set CF in IO mode and activate CF IRQ to work in level mode
	        writeb( 0x41, base_addr + CF_CPLD_HWIF_OFFSET * i );
	    
	        //Check if the value writen can be read (some slow CF doesn't work...)
	        if( readb( base_addr + CF_CPLD_HWIF_OFFSET * i ) != 0x41 )
	        {
	            printk( DRV_PREFIX ": unable to access to Configuration Option Register\n" );
	            continue;
	        }
	    
	        // Configure CPLD to operate in IO mode
	        writeb( CF_MODE_MASK, base_addr + CF_CPLD_CONF_REG_OF + CF_CPLD_HWIF_OFFSET * i );

	        unc90_ide_register( base_addr + CF_CPLD_HWIF_OFFSET * i,
	                            base_addr + CF_CPLD_HWIF_OFFSET * i,
	                            UNCBASCF_DEV_KIT_IDE_IRQ, &hwifs[i] );
	    }
	}

	return 0;
}


module_init( unc90_ide_init );

MODULE_AUTHOR( "Pedro Perez de Heredia <pperez@embebidos.com>" );
MODULE_LICENSE( "GPL" );
MODULE_DESCRIPTION( "UNC90 IDE driver");
