/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: jtagcomm.c,v 1.5 2006-08-17 15:11:48 bwesterm Exp $
 * @Author: Markus Pietrek
 * @Descr: Provides /dev/ttyJ/
 *         Writes data to the ARM ICEBreaker COMMS channel and visa versa.
 *         It is doing polled I/O, only a limited HW handshake is used.
 *         If the JTAG debugger doesn't fetch characters
 *         within JTAGCOMM_TX_TIMEOUT us, the driver assumes that there is no
 *         connection at all and sets the timeout to 1us for further tries. This
 *         is reset on the next open or when a connection is reestablished
 *
 *
 * @References: ARM7TDMI Data Sheet ARM DDI 0029E
 *              (/net/m/NEUE_TEILE/ARM/ARM7TDMIvE.pdf) chapter 9.11.2
 *
 *          Use it with "ppjunc20 /term" or "jtagarm7 /term"
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/*******************************************************************************
 *  History:
 *     17/08/06	change flip buffer stuff (Bernd Westermann)
 *******************************************************************************/

#include <linux/config.h>
#include <linux/init.h>		/* __init */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/console.h>	/* struct console */
#include <linux/ioport.h>	/* request_mem_region */

#include <linux/serial_core.h>	/* uart_console_device */

#include <asm/semaphore.h>	/* init_MUTEX */
#include <asm/delay.h>		/* udelay */

#define DRIVER_NAME		"jtagcomm"

#define NR_PORTS		1

#define JTAGCOMM_TX_TIMEOUT 	1000000 /* us */
#define JTAGCOMM_RX_POLL_TIME 	1 /* in JIFFIES, gives us about 1000 baud */

#ifdef DEBUG
# define DEBUG_INIT		0x0001
# define DEBUG_MINOR		0x0002
# define DEBUG_TX		0x0004
# define DEBUG_RX		0x0008
# define DEBUG_ERROR_CRIT	0x8000

static int debug = DEBUG_ERROR_CRIT|DEBUG_INIT;

# define DEBUG_FN( FLG ) if( (debug & (FLG)) == (FLG) ) \
		printk("\r%s:line %d\n", __FUNCTION__, __LINE__);
#else /* DEBUG */
# define DEBUG_FN( n )
#endif /* DEBUG */

#ifdef CONFIG_MACH_CCXP
#define JTAGCOMM_READ_STATUS	"mrc p14, 0, %0, c14, c0"
#define JTAGCOMM_READ_DATA	"mrc p14, 0, %0, c9, c0"
#define JTAGCOMM_WRITE_DATA	"mcr p14, 0, %0, c8, c0"
#define JTAGCOMM_RX_RDY_MASK	0x80000000
#define JTAGCOMM_TX_RDY_MASK	0x10000000
#else /*ARM7 ARM9*/
#define JTAGCOMM_READ_STATUS	"mrc p14, 0, %0, c0, c0"
#define JTAGCOMM_READ_DATA	"mrc p14, 0, %0, c1, c0"
#define JTAGCOMM_WRITE_DATA	"mcr p14, 0, %0, c1, c0"
#define JTAGCOMM_RX_RDY_MASK	0x00000001
#define JTAGCOMM_TX_RDY_MASK	0x00000002
#endif

static int __init jc_init( void );
static void __exit jc_exit( void );

static unsigned int jc_tx_empty( struct uart_port* port );
static void jc_set_mctrl( struct uart_port* port, unsigned int mctrl );
static unsigned int jc_get_mctrl( struct uart_port* port );
static void jc_stop_tx( struct uart_port* port );
static void jc_start_tx( struct uart_port* port );
static void jc_stop_rx( struct uart_port* port );
static void jc_enable_ms( struct uart_port* port );
static void jc_break_ctl( struct uart_port* port, int break_state );
static int jc_startup( struct uart_port* port );
static void jc_shutdown( struct uart_port* port );
static void jc_set_termios( struct uart_port* port, struct termios* termios,
			    struct termios* old );
static void jc_pm( struct uart_port* port, unsigned int state,
			   unsigned int oldstate );
static const char* jc_type( struct uart_port* port );
static void jc_release_port( struct uart_port* port );
static int jc_request_port( struct uart_port* port );
static void jc_config_port( struct uart_port* port, int flags );
static int jc_verify_port( struct uart_port* port,
			   struct serial_struct* ser );
static void jc_rx_handle( unsigned long dev );
static inline void jc_send_char( char c );

#ifndef MODULE
static int __init jc_console_init( void );
static int __init jc_console_setup( struct console* co, char* options );
static void jc_console_write( struct console* co, const char* s,
			      unsigned int count );

static struct uart_driver jc_reg;
static struct console jc_console = {
	.name		= "ttyJ",
	.write		= jc_console_write,
	.device		= uart_console_device,
	.setup		= jc_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &jc_reg
};
#endif /* MODULE */

static struct uart_driver jc_reg = {
	.owner			= THIS_MODULE,
	.driver_name		= DRIVER_NAME,
	.devfs_name		= "ttyJ/",
	.dev_name		= "ttyJ",
	.major			= TTY_MAJOR,
	.minor			= 96,
	.nr			= NR_PORTS,
#ifndef MODULE
	.cons			= &jc_console
#endif /* MODULE */
};

static int jc_tx_timeout = JTAGCOMM_TX_TIMEOUT;

static struct uart_ops jc_pops = {
	.tx_empty	= jc_tx_empty,
	.set_mctrl	= jc_set_mctrl,
	.get_mctrl	= jc_get_mctrl,
	.stop_tx	= jc_stop_tx,
	.start_tx	= jc_start_tx,
	.stop_rx	= jc_stop_rx,
	.enable_ms	= jc_enable_ms,
	.break_ctl	= jc_break_ctl,
	.startup	= jc_startup,
	.shutdown	= jc_shutdown,
	.set_termios	= jc_set_termios,
	.pm		= jc_pm,
	.type		= jc_type,
	.release_port	= jc_release_port,
	.request_port	= jc_request_port,
	.config_port	= jc_config_port,
	.verify_port	= jc_verify_port,
};

static struct uart_port jc_port;
static DECLARE_MUTEX( jc_sem_write );
static DECLARE_MUTEX( jc_sem_read );
static struct timer_list jc_rx_timer; /* timer for polling input */

static int __init jc_init( void )
{
	int ret;

	DEBUG_FN( DEBUG_INIT );

	printk( KERN_INFO "Serial: JTAG driver $Revision: 1.5 $\n" );

	ret = uart_register_driver( &jc_reg );
	if( ret < 0 )
		return ret;

	jc_port.uartclk   = 0; /* clock makes no sense here*/
	jc_port.fifosize  = 1;
	jc_port.line      = 0;
	jc_port.iotype    = SERIAL_IO_MEM;
	jc_port.ops       = &jc_pops;
	jc_port.membase   = (void*) &jc_port; /* we don't need it but
				       * uart_configure_port requires it set  */
	jc_port.mapbase   = 0;
	jc_port.flags     = ASYNC_BOOT_AUTOCONF;
	jc_port.irq       = 0;	/* dummy, but must be 0 <= irq <= NR_IRQS */

	uart_add_one_port( &jc_reg, &jc_port );

	return ret;
}

static void __exit jc_exit( void )
{
	DEBUG_FN( DEBUG_INIT );

	uart_remove_one_port( &jc_reg, &jc_port );
	uart_unregister_driver( &jc_reg );
}

static unsigned int jc_tx_empty( struct uart_port* port )
{
	DEBUG_FN( DEBUG_TX | DEBUG_MINOR );

	/* no characters to process */
	return TIOCSER_TEMT;
}

static void jc_set_mctrl( struct uart_port* port, unsigned int mctrl )
{
	DEBUG_FN( DEBUG_MINOR );

	/* @TODO */
}

static unsigned int jc_get_mctrl( struct uart_port* port )
{
	DEBUG_FN( DEBUG_MINOR );

	/* @TODO */
	return 0U;
}

static void jc_stop_tx( struct uart_port* port )
{
	DEBUG_FN( DEBUG_TX | DEBUG_MINOR );

	return;
}

/***********************************************************************
 * @Function: jc_start_tx
 * @Return: n/a
 * @Descr: send all data immediately
 ***********************************************************************/

static void jc_start_tx( struct uart_port* port )
{
	struct circ_buf* xmit = &jc_port.info->xmit;

	DEBUG_FN( DEBUG_TX );

	if( jc_port.x_char ) {
		jc_send_char( jc_port.x_char );
		jc_port.icount.tx++;
		jc_port.x_char = 0;
		return;
	}

	if( uart_circ_empty( xmit ) || uart_tx_stopped( &jc_port ) )
		return;

	while( !uart_circ_empty( xmit ) ) {
		jc_send_char( xmit->buf[ xmit->tail ] );
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		jc_port.icount.tx++;
	};

	if( uart_circ_chars_pending( xmit ) < WAKEUP_CHARS )
		uart_write_wakeup( &jc_port );
}

static void jc_stop_rx( struct uart_port* port )
{
	DEBUG_FN( DEBUG_RX );

	del_timer_sync( &jc_rx_timer );
}

static void jc_enable_ms( struct uart_port* port )
{
	DEBUG_FN( DEBUG_MINOR );

	/* @TODO */
}

static void jc_break_ctl( struct uart_port* port, int break_state )
{
	DEBUG_FN( DEBUG_MINOR );

	/* @TODO */
}

/***********************************************************************
 * @Function: jc_startup
 * @Return: 0
 * @Descr: Starts receive timer
 ***********************************************************************/

static int jc_startup( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT );

	/* start receive timer */
	init_timer( &jc_rx_timer );
	jc_rx_timer.function = jc_rx_handle;
	jc_rx_timer.expires = jiffies + JTAGCOMM_RX_POLL_TIME;
	add_timer( &jc_rx_timer );

	jc_tx_timeout = JTAGCOMM_TX_TIMEOUT;
	return 0;
}

static void jc_shutdown( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT );

	/* stop timers */
	del_timer_sync( &jc_rx_timer );
}

static void jc_set_termios( struct uart_port* port, struct termios* termios,
			    struct termios* old )
{
	DEBUG_FN( DEBUG_MINOR );
}

static void jc_pm( struct uart_port* port, unsigned int state,
		   unsigned int oldstate )
{
	DEBUG_FN( DEBUG_MINOR );

	/* power management not supported yet */
}

static const char* jc_type( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	return "JTAGCOMM";
}

static void jc_release_port( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	/* do nothing */
}

static int jc_request_port( struct uart_port* port )
{
	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	/* do nothing */
	return 0;
}

static void jc_config_port( struct uart_port* port, int flags )
{
	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	if( flags & UART_CONFIG_TYPE )
		port->type = PORT_JTAGCOMM;
}

static int jc_verify_port( struct uart_port* port,
			   struct serial_struct* ser )
{
	int ret = 0;

	DEBUG_FN( DEBUG_INIT | DEBUG_MINOR );

	if( ser->type != PORT_UNKNOWN && ser->type != PORT_JTAGCOMM )
		ret = -EINVAL;

	return ret;
}

static void jc_rx_handle( unsigned long dev )
{
	int debug_comms_ctrl = JTAGCOMM_TX_RDY_MASK;
	int len;
	unsigned int data;
	struct tty_struct *tty = jc_port.info->tty;

	DEBUG_FN( DEBUG_RX );

	down( &jc_sem_read );
	__asm__ __volatile__ ( JTAGCOMM_READ_STATUS
			       : "=r"( debug_comms_ctrl ) : );

	if( ( debug_comms_ctrl & JTAGCOMM_RX_RDY_MASK ) == JTAGCOMM_RX_RDY_MASK ) {
		/* input available, read ICE debug data register */
		__asm__ __volatile__ (JTAGCOMM_READ_DATA : "=r"(data):);

		len = tty_buffer_request_room(tty, 1);
		if( len > 0 ) {
			tty_insert_flip_char(tty, data, TTY_NORMAL);
			tty_flip_buffer_push( tty );
		}
	}

	up( &jc_sem_read );

	/* try again later */
	mod_timer( &jc_rx_timer, jiffies + JTAGCOMM_RX_POLL_TIME );
}


/***********************************************************************
 * @Function: jc_send_char
 * @Return: n/a
 * @Descr: writes data to the jtag debug port. If JTAG is busy for
 * 	   JTAGCOMM_TX_TIMEOUT us, character is lost. Also, the timeout is reduced
 * 	   to 1 to speed up booting.
 ***********************************************************************/


static inline void jc_send_char( char c )
{
	int debug_comms_ctrl = JTAGCOMM_TX_RDY_MASK
;
	int timeout = jc_tx_timeout;
	unsigned int data = (unsigned int) c;

	down( &jc_sem_write );
	while( timeout-- ) {
		/* read ICEbreaker debug comms ctrl register */
		/* !!gcc will move this line before the loop without volatile */
		__asm__ __volatile__ ( JTAGCOMM_READ_STATUS
				       : "=r"( debug_comms_ctrl ) : );
		if( !( debug_comms_ctrl & JTAGCOMM_TX_RDY_MASK ) )
			break;

		udelay( 1 );	/* so JTAGCOMM_TX_TIMEOUT is calibrated */
	}

	if( timeout >= 0 ) {
		/* write ICEbreaker debug data register */
		__asm__ __volatile__ ( JTAGCOMM_WRITE_DATA :
				       : "r"( data ) );
		jc_tx_timeout = JTAGCOMM_TX_TIMEOUT; /* online again */
	} else
		jc_tx_timeout = 1; /* connection lost, be low until next open */

	up( &jc_sem_write );
}

#ifndef MODULE
static int __init jc_console_init( void )
{
	jc_tx_timeout = JTAGCOMM_TX_TIMEOUT;

	register_console( &jc_console );

	return 0;
}

static int __init jc_console_setup( struct console* co, char* options )
{
	if( co->index == -1 || co->index >= NR_PORTS )
		co->index = 0;

	/* for default consoles in kernel commandline, co is not initialized */
	return 0;
}

static void jc_console_write( struct console* co, const char* s,
			      unsigned int count )
{
	char c = *s;

	while( count ) {
		jc_send_char( c );

		if( c == '\n' )
			c = '\r';
		else {
			/* next character */
			count--;
			s++;
			c = *s;
		}
	}
}

console_initcall( jc_console_init );
#endif /* MODULE */

module_init( jc_init );
module_exit( jc_exit );

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>" );
MODULE_DESCRIPTION( "JTAG serial driver" );
MODULE_LICENSE( "GPL" );

#ifdef DEBUG
module_param(debug, int, 0);
MODULE_PARM_DESC(debug, "Debug level");
#endif
