/***********************************************************************
 *
 * Copyright (C) 2005 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: m95640.c,v 1.1 2005-12-15 16:11:54 jjaeger Exp $
 * @Author: Joachim Jaeger
 * @Descr: M95640 EEPROM Driver
 * @References: [1] M95640 HW Reference 5711.pdf
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
 /*  @History:
 *     2005/08/18 : JJ: generated from 24c64.c I2C driver
 ***********************************************************************/
 /* @TODO: expand the driver to be able to use 3bytes for addreses */
 
#include <linux/config.h>
#include <linux/module.h>	/* module_init */
#include <linux/init.h>		/* __init */
#include "../spi.h"
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-spi.h>
#include "m95640.h"

#include <asm/semaphore.h>	/* init_MUTEX */

#define DEBUG_INIT		0x0001
#define DEBUG_MINOR		0x0002
#define DEBUG_RX		0x0004
#define DEBUG_TX		0x0008
#define DEBUG_ERROR		0x4000
#define DEBUG_ERROR_CRIT	0x8000

static int debug = 0xc000;

#define DEBUG_ARGS( FLG, a0, args... ) if( ( debug & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, ## args )
#define DEBUG_FN( FLG ) if( (debug & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
#define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }

#define DRIVER_NAME		"m95640"

#define SPI_EEPROM_SIZE		8192


static int m95640_attach_adapter( struct spi_adapter* adapter );
static int m95640_detach_client( struct spi_client* client );
static int m95640_attach( struct spi_adapter* adapter, int kind );
static ssize_t m95640_read( struct kobject* kobj, char* buf, loff_t off,
			    size_t count );
static ssize_t m95640_write( struct kobject* kobj, char* buf, loff_t off,
			     size_t count );
static ssize_t m95640_read_bytes( struct spi_client* client, char* buf,
				  loff_t off, size_t count );
static ssize_t m95640_write_bytes( struct spi_client* client, char* buf,
				   loff_t off, size_t count );

static int __init m95640_init( void );
static void m95640_exit( void );

static struct semaphore m95640_lock; /* all read and write accesses */

static struct spi_driver m95640_driver = {
	.owner		= THIS_MODULE,
	.name		= DRIVER_NAME,
	/* register for id @ http://secure.netroedge.com/~lm78/docs.html */
	.id		= SPI_DRIVERID_M95640,
	.flags		= SPI_DF_NOTIFY,
	.attach_adapter	= m95640_attach_adapter,
	.detach_client	= m95640_detach_client,
};

static struct bin_attribute m95640_attr = {
	.attr = {
		.name = "eeprom",
		.mode = S_IRUGO,
	},
	.size = SPI_EEPROM_SIZE,
	.read = m95640_read,
	.write = m95640_write
};

static int m95640_attach_adapter( struct spi_adapter* adapter )
{
	DEBUG_FN( DEBUG_INIT );

	/* no SPIBus, can't do probing */
	m95640_attach( adapter, 0 );
	return 0;
}

static int m95640_detach_client( struct spi_client* client )
{
	int res;

	DEBUG_FN( DEBUG_INIT );

	if( ( res = spi_detach_client( client ) ) )
		return res;

	kfree( client );

	return 0;
}

static int m95640_attach( struct spi_adapter* adapter, int kind )
{
	struct spi_client* client;
	int res;

	DEBUG_FN( DEBUG_INIT );

	if( ( client = (struct spi_client *) kmalloc( sizeof(struct spi_client),
						      GFP_KERNEL ) ) == NULL )
		return -ENOMEM;

	memset( client, 0x0, sizeof( struct spi_client ) );

	strcpy( client->name, m95640_driver.name ); /*DRIVER_NAME<SPI_NAME_SIZE*/
	client->id = m95640_driver.id;
	client->flags = SPI_CLIENT_ALLOW_USE | SPI_CLIENT_ALLOW_MULTIPLE_USE;
	client->adapter = adapter;
	client->driver = &m95640_driver;
	client->spi_chan = 0;

	init_MUTEX( &m95640_lock );

	if( ( res = spi_attach_client( client ) ) ) {
		kfree( client );
		return res;
	}

	sysfs_create_bin_file( &client->dev.kobj, &m95640_attr );

	return 0;
}

/***********************************************************************
 * @Function: m95640_read
 * @Return: number of bytes read or -EIO for SPI failure
 * @Descr: If possible, uses page accesses to read from eeprom. Otherwise
 *         performs random accesses.
 ***********************************************************************/

static ssize_t m95640_read( struct kobject* kobj, char* buf, loff_t off,
			    size_t count )
{
	struct spi_client* client = to_spi_client(
		container_of( kobj, struct device, kobj) );
	ssize_t trans = 0;

	if (off > SPI_EEPROM_SIZE)
		return 0;
	if (off + count > SPI_EEPROM_SIZE)
		count = SPI_EEPROM_SIZE - off;
		
	trans = m95640_read_bytes( client, buf, off, count );
	if( trans != count )
		return -EIO;

	return count;
}

/***********************************************************************
 * @Function: m95640_write
 * @Return: number of bytes written or -EIO for SPI failure
 * @Descr: If possible, uses page accesses to write to eeprom. Otherwise
 *         performs random accesses.
 ***********************************************************************/

static ssize_t m95640_write( struct kobject* kobj, char* buf, loff_t off,
			     size_t count )
{

	struct spi_client* client = to_spi_client(
		container_of( kobj, struct device, kobj) );

	ssize_t overall = 0;
	ssize_t trans = 0;
	ssize_t trans_in;

	if (off > SPI_EEPROM_SIZE)
		return 0;
	if (off + count > SPI_EEPROM_SIZE)
		count = SPI_EEPROM_SIZE - off;

	while( count > overall ) {
		/* page write when 32byte aligned */
		trans_in = ((off % 32) ? 1 : ((count > 32) ? 32 : count));

		trans = m95640_write_bytes( client, buf, off, trans_in );

		if( trans != trans_in )
			return -EIO;

		overall += trans;

		/* next page */
		buf += trans;
		off += trans;
	}

	return count;
}

/***********************************************************************
 * @Function: m95640_read_bytes
 * @Return: # bytes read, -EINVAL for wrong size or offset or -EIO on SPI failure
 * @Descr:
 ***********************************************************************/

static ssize_t m95640_read_bytes( struct spi_client* client, char* buf,
				  loff_t off, size_t count )
{
        int msg_len;
	/* highest address byte first on wire */
        unsigned char cmd1[] = {SPI_READ,(off >> 8),(off & 0xff)};

	/* @TODO: not big endian safe */
	struct spi_msg msg[] = {
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len:3, wbuf: cmd1, rbuf:NULL },
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len:count, wbuf:NULL, rbuf:buf }
        };

	msg_len = sizeof( msg ) / sizeof( struct spi_msg );

	DEBUG_FN( DEBUG_RX | DEBUG_MINOR );

	if( off + count > SPI_EEPROM_SIZE ) {
        	printk( KERN_ERR "%s EEPROM read out of area: %i + %i > %i\n",
			__FUNCTION__, (int) off, count, SPI_EEPROM_SIZE );
		return -EINVAL;
	}

        if( spi_transfer( client->adapter, msg, msg_len ) != msg_len ) {
                return -EIO;
	}

	return count;
}

/***********************************************************************
 * @Function: m95640_write_bytes
 * @Return: # bytes written, -EINVAL for wrong size or offset or -EIO on SPI failure
 * @Descr:
 ***********************************************************************/

static ssize_t m95640_write_bytes( struct spi_client* client, char* buf,
				   loff_t off, size_t count )
{
        
        int msg_len;
        unsigned char cmd0[]={SPI_WREN,SPI_WREN};	/* one dummy byte */
	/* highest address byte first on wire */
        unsigned char cmd1[]={SPI_WRITE,(off >> 8),(off & 0xff)};
        unsigned char cmd2[]={SPI_RDSR};
        unsigned char rstat[1];
        
	/* @TODO: not big endian safe */
	/* Write Enable Command */
	struct spi_msg msgwe[] = {
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len: 2, wbuf: cmd0, rbuf:NULL}
        };
	/* Write Command */
	struct spi_msg msg[] = {
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len: 3, wbuf: cmd1, rbuf:NULL},
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len: count, wbuf: buf, rbuf:NULL}
        };
	/* Read Status Command */
	struct spi_msg msgrd[] = {
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len: 1, wbuf: cmd2, rbuf:NULL},
                {flags:(SPI_M_CPOL_ACTHIGH | SPI_M_CPHA_FORMATA),len: 1, wbuf: NULL, rbuf:rstat}
        };

	DEBUG_FN( DEBUG_TX | DEBUG_MINOR );

	if( off + count > SPI_EEPROM_SIZE ) {
        	printk( KERN_ERR "%s EEPROM write out of area: %i + %i > %i\n",
			__FUNCTION__, (int) off, count, SPI_EEPROM_SIZE );
		return -EINVAL;
	}

	msg_len = sizeof( msgwe ) / sizeof( struct spi_msg );

        if( spi_transfer( client->adapter, msgwe, msg_len ) != msg_len )
                return -EIO;

	msg_len = sizeof( msg ) / sizeof( struct spi_msg );

        if( spi_transfer( client->adapter, msg, msg_len ) != msg_len )
                return -EIO;

	msg_len = sizeof( msgrd ) / sizeof( struct spi_msg );

        /* Wait before calling the Status Register */
        do {
        	if( spi_transfer( client->adapter, msgrd, msg_len ) != msg_len )
			return -EIO;
	}
	while ((rstat[0] & WIP_MASK) != 0);

	return count;
}

static int __init m95640_init( void )
{
	DEBUG_FN( DEBUG_INIT );

	printk( KERN_INFO DRIVER_NAME
		": EEPROM driver $Revision: 1.1 $ initializing\n");

	return spi_add_driver( &m95640_driver );
}

static void m95640_exit( void )
{
	DEBUG_FN( DEBUG_INIT );

	spi_del_driver( &m95640_driver );
}

device_initcall( m95640_init );
module_exit( m95640_exit );

MODULE_AUTHOR("Joachim Jaeger <joachim_jaeger@digi.com>");
MODULE_DESCRIPTION("EEPROM M95640 driver");
MODULE_LICENSE("GPL");
