/*
 * Copyright (C) 2005 Stephen Street / StreetFire Sound Labs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * This is a prototype dma/interrupt driven SPI "controller" for Intel's PXA2xx 
 * series SOC. The driver plugs into the lightweight SPI framework developed by
 * David Brownell. Hardwired configuration information is provided via
 * spi_board_info structures initialized in arch/arm/mach_pxa board
 * initialization code (see include/linux/spi.h for details). 
 * 
 * This follow code snippet demostrates a sample board configuration using
 * the PXA255 NSSP port connect to a "loopback" pseudo-chip.
 * 
 *
 * static struct pxa2xx_spi_chip loopback_chip_info = {
 *	.mode = SPI_MODE_3,
 *	.tx_threshold = 12,
 *	.rx_threshold = 4,
 *	.dma_burst_size = 8,
 *	.bits_per_word = 8,
 *	.timeout_microsecs = 64,
 *	.enable_loopback = 1,
 * };
 *
 * static struct spi_board_info board_info[] __initdata = {
 *	{
 *		.modalias = "loopback",
 *		.max_speed_hz = 3686400,
 *		.bus_num = 2,
 *		.chip_select = 0,
 *		.controller_data = &loopback_chip_info,
 *	},
 * };
 *
 * static struct resource pxa_spi_resources[] = {
 *	[0] = {
 *		.start	= __PREG(SSCR0_P(2)),
 *		.end	= __PREG(SSCR0_P(2)) + 0x2c,
 *		.flags	= IORESOURCE_MEM,
 *	},
 *	[1] = {
 *		.start	= IRQ_NSSP,
 *		.end	= IRQ_NSSP,
 *		.flags	= IORESOURCE_IRQ,
 *	},
 * };
 *
 * static struct pxa2xx_spi_master pxa_nssp_master_info = {
 *	.bus_num = 2,
 *	.clock_enable = CKEN9_NSSP,
 *	.num_chipselect = 3,
 * 	.enable_dma = 1,
 * };
 *
 * static struct platform_device pxa_spi_ssp = {
 *	.name = "pxa2xx-spi-ssp",
 *	.id = -1,
 *	.resource = pxa_spi_resources,
 *	.num_resources = ARRAY_SIZE(pxa_spi_resources),
 *	.dev = {
 *		.platform_data = &pxa_nssp_master_info,
 *	},
 * }; 
 *
 * static void __init streetracer_init(void)
 * {
 *	platform_device_register(&pxa_spi_ssp);
 *	spi_register_board_info(board_info, ARRAY_SIZE(board_info));
 * } 
 *
 * The driver is built around a spi_message fifo serviced by two tasklets. The
 * first tasklet (pump_messages) is responsible for queuing SPI transactions 
 * and scheduling SPI transfers.  The second tasklet (pump_transfers) is 
 * responsible to setting up and launching the dma/interrupt driven transfers.
 * DMA transfers maybe enabled by setting the "enable_dma" flag in the 
 * "pxa2xx_spi_master" structure.  The driver will select DMA transfers if the 
 * "spi_message" transfer buffers are aligned on a 8-byte boundary.  Unaligned 
 * buffers will be sent using interrupt driven PIO. The driver user "Streaming
 * DMA mappings" to program two PXA2xx dma channels aquired during the device 
 * probing. Per transfer chip select and delay control is available.
 * 
 * This is a prototype driver, so you mileage will vary. It has only been
 * tested on the NSSP port.
 * 
 * Known Limitations:
 * 	Does not handle invert chip select polarity.
 * 	Heavy loaded systems may see transaction failures.
 * 	Wordsize support is untested.
 * 	Module hangs during unload.
 * 
 */
 
#include <linux/init.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/spi.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/delay.h>
#include <asm/dma.h>

#include <asm/arch/hardware.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/pxa2xx_spi_ssp.h>

MODULE_AUTHOR("Stephen Street");
MODULE_DESCRIPTION("PXA2xx SSP SPI Contoller");
MODULE_LICENSE("GPL");

#define MAX_SPEED_HZ 3686400
#define MAX_BUSES 3

#define DMA_INT_MASK (DCSR_ENDINTR | DCSR_STARTINTR | DCSR_BUSERR)
#define RESET_DMA_CHANNEL (DCSR_NODESC | DMA_INT_MASK)

#define IS_DMA_ALIGNED(x,y) ((((u32)(x)& 0x07)==0) && (((u32)(y)& 0x07)==0))
#define GET_IRQ_STATUS(x) (__REG(x)&(SSSR_TINT|SSSR_RFS|SSSR_TFS|SSSR_ROR))

struct master_data;

struct transfer_state {
	struct spi_msg *transfer;
	int len;
	void *tx;
	void *tx_end;
	void *rx;
	void *rx_end;
	void (*write)(u32 sssr, u32 ssdr, struct transfer_state *state);
	void (*read)(u32 sssr, u32 ssdr, struct transfer_state *state);
	irqreturn_t (*transfer_handler)(struct master_data *drv_data, 
					struct spi_msg *msg, 
					struct transfer_state *state);
	void (*cs_control)(u32 command);
};

struct master_data {
	spinlock_t lock;
	struct spi_master *master;
	int rx_channel;
	int tx_channel;
	struct list_head queue;
	struct tasklet_struct pump_messages;
	struct tasklet_struct pump_transfers;
	struct spi_msg* cur_msg;
	struct transfer_state cur_state;
	void *null_dma_buf;
	u32 trans_cnt;
	u32 dma_trans_cnt;
	u32 int_trans_cnt;
	u32 bigunalign_cnt;
	u32 buserr_cnt;
	u32 ror_cnt;
	u32 sscr0;
	u32 sscr1;
	u32 sssr;
	u32 ssitr;
	u32 ssdr;
	u32 ssto;
	u32 sspsp;
};

struct chip_data {
	u32 cr0;
	u32 cr1;
	u32 to;
	u32 psp;
	u32 timeout;
	u8 n_bytes;
	u32 dma_width;
	u32 dma_burst_size;
	u32 threshold;
	u32 dma_threshold;
	u8 enable_dma;
	void (*write)(u32 sssr, u32 ssdr, struct transfer_state *state);
	void (*read)(u32 sssr, u32 ssdr, struct transfer_state *state);
	void (*cs_control)(u32 command);
};

static inline void flush(struct master_data *drv_data)
{
	u32 sssr = drv_data->sssr;
	u32 ssdr = drv_data->ssdr;
	
	do {
		while (__REG(sssr) & SSSR_RNE) {
			(void)__REG(ssdr);
		}
	} while (__REG(sssr) & SSSR_BSY);
	__REG(sssr) = SSSR_ROR ;
}

static inline void save_state(struct master_data *drv_data, 
				struct chip_data *chip)
{
	/* Save critical register */
	chip->cr0 = __REG(drv_data->sscr0);
	chip->cr1 = __REG(drv_data->sscr1);
	chip->to = __REG(drv_data->ssto);
	chip->psp = __REG(drv_data->sspsp);
	
	/* Disable clock */
	__REG(drv_data->sscr0) &= ~SSCR0_SSE;
}

static inline void restore_state(struct master_data *drv_data, 
					struct chip_data *chip)
{
	/* Clear status and disable clock*/
	__REG(drv_data->sssr) = SSSR_ROR | SSSR_TUR | SSSR_BCE;
	__REG(drv_data->sscr0) = chip->cr0 & ~SSCR0_SSE;
	
	/* Load the registers */
	__REG(drv_data->sscr1) = chip->cr1;
	__REG(drv_data->ssto) = chip->to;
	__REG(drv_data->sspsp) = chip->psp;
	__REG(drv_data->sscr0) = chip->cr0;
}

static inline void dump_state(struct master_data* drv_data)
{
	u32 sscr0 = drv_data->sscr0;
	u32 sscr1 = drv_data->sscr1;
	u32 sssr = drv_data->sssr;
	u32 ssto = drv_data->ssto;
	u32 sspsp = drv_data->sspsp;
	
	printk("SSP dump: sscr0=0x%08x, sscr1=0x%08x, "
			"ssto=0x%08x, sspsp=0x%08x, sssr=0x%08x\n",
			__REG(sscr0), __REG(sscr1), __REG(ssto), 
			__REG(sspsp), __REG(sssr));
}

static inline void dump_message(struct spi_msg *msg)
{
/*	dev_dbg(dev, "message = %p\n", msg); TODO error dev not known*/
}

static void null_cs_control(u32 command)
{
}

static void null_writer(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_TNF) && (state->tx < state->tx_end)) {
		__REG(ssdr) = 0;
		++state->tx;
	}
}

static void null_reader(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_RNE) && (state->rx < state->rx_end)) {
		(void)(__REG(ssdr));
		++state->rx;
	}
}

static void u8_writer(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_TNF) && (state->tx < state->tx_end)) {
		__REG(ssdr) = *(u8 *)(state->tx);
		++state->tx;
	}
}

static void u8_reader(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_RNE) && (state->rx < state->rx_end)) {
		*(u8 *)(state->rx) = __REG(ssdr);
		++state->rx;
	}
}

static void u16_writer(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_TNF) && (state->tx < state->tx_end)) {
		__REG(ssdr) = *(u16 *)(state->tx);
		state->tx += 2;
	}
}

static void u16_reader(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_RNE) && (state->rx < state->rx_end)) {
		*(u16 *)(state->rx) = __REG(ssdr);
		state->rx += 2;
	}
}
static void u32_writer(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_TNF) && (state->tx < state->tx_end)) {
		__REG(ssdr) = *(u32 *)(state->tx);
		state->tx += 4;
	}
}

static void u32_reader(u32 sssr, u32 ssdr, struct transfer_state *state)
{
	while ((__REG(sssr) & SSSR_RNE) && (state->rx < state->rx_end)) {
		*(u32 *)(state->rx) = __REG(ssdr);
		state->rx += 4;
	}
}

static void dma_handler(int channel, void *data, struct pt_regs *regs)
{
	struct master_data *drv_data = (struct master_data *)data;
	u32 sssr = drv_data->sssr;
	u32 sscr1 = drv_data->sscr1;
	u32 ssto = drv_data->sscr1;
	u32 irq_status = DCSR(drv_data->tx_channel) & DMA_INT_MASK;
	struct spi_message *msg = drv_data->cur_msg;
	struct transfer_state *state = 
		(struct transfer_state *)spimsg_get_ctx(msg);
	
	if (irq_status & DCSR_BUSERR) {

		__REG(ssto) = 0;
		__REG(sssr) = SSSR_TINT | SSSR_ROR;
		__REG(sscr1) &= ~(SSCR1_TSRE | SSCR1_RSRE | SSCR1_TINTE);
		DCSR(drv_data->rx_channel) = RESET_DMA_CHANNEL;
		DCSR(drv_data->tx_channel) = RESET_DMA_CHANNEL;
		
		flush(drv_data);
		
		printk(KERN_ERR "dma_handler: bad bus address on channel %d\n",
					channel);
		
		drv_data->buserr_cnt++;
		drv_data->cur_state.transfer = PTR_ERR(-2);
		tasklet_schedule(&drv_data->pump_transfers);
	}
	
	if (channel == drv_data->rx_channel && irq_status & DCSR_ENDINTR) {
		/* Forward to transfer handler */
		state->rx = state->rx_end;
		state->transfer_handler(drv_data, msg, state);
	}
}

static irqreturn_t dma_transfer(struct master_data *drv_data, 
				struct spi_msg *msg, 
				struct transfer_state *state)
{
	u32 sssr = drv_data->sssr;
	u32 ssdr = drv_data->ssdr;
	u32 sscr1 = drv_data->sscr1;
	u32 ssto = drv_data->ssto;
	u32 irq_status = GET_IRQ_STATUS(sssr);
	u32 trailing_sssr = 0;

	if (irq_status & SSSR_ROR) {
		/* Clear and disable interrupts on SSP and DMA channels*/
		__REG(ssto) = 0;
		__REG(sssr) = SSSR_TINT | SSSR_ROR;
		__REG(sscr1) &= ~(SSCR1_TSRE | SSCR1_RSRE | SSCR1_TINTE);
		DCSR(drv_data->tx_channel) = RESET_DMA_CHANNEL;
		DCSR(drv_data->rx_channel) = RESET_DMA_CHANNEL;
		
		flush(drv_data);
		
		printk(KERN_WARNING "dma_transfer: fifo overun"
					"tx dcmd = 0x%08x rx dcmd = 0x%08x\n",
					DCMD(drv_data->tx_channel),
					DCMD(drv_data->rx_channel));

		drv_data->ror_cnt++;

		state->transfer = PTR_ERR(-2);
		tasklet_schedule(&drv_data->pump_transfers);

		return IRQ_HANDLED;
	}

	if (irq_status & SSSR_TINT || state->rx == state->rx_end) {
		
		/* Check for false positive */
		if (DCSR(drv_data->tx_channel) & DCSR_RUN) {
			__REG(sssr) = SSSR_TINT;
			return IRQ_HANDLED;
		}
		
		/* Clear and disable interrupts on SSP and DMA channels*/
		__REG(ssto) = 0;
		__REG(sssr) = SSSR_TINT | SSSR_ROR ;
		__REG(sscr1) &= ~(SSCR1_TSRE | SSCR1_RSRE | SSCR1_TINTE);
		DCSR(drv_data->tx_channel) = RESET_DMA_CHANNEL;
		DCSR(drv_data->rx_channel) = RESET_DMA_CHANNEL;
		while (!(DCSR(drv_data->rx_channel) & DCSR_STOPSTATE)
				|| (__REG(sssr) & SSSR_BSY));
		
		/* Handle trailing byte, must unmap the dma buffers */
		spimsg_put_buffer(msg, state->transfer);

		/* Calculate number of trailing bytes */
		trailing_sssr = __REG(sssr);
		if ((trailing_sssr & 0xf008) != 0xf000) {
			state->rx = state->rx_end - 
					(((trailing_sssr >> 12) & 0x0f) + 1);
			state->read(sssr, ssdr, state);
		}

		if (!(spimsg_ge_flags(state->transfer) & SPI_M_CSKEEP))
			state->cs_control(PXA2XX_CS_DEASSERT);
		
		state->transfer = spimsg_getnext(state->transfer);
		tasklet_schedule(&drv_data->pump_transfers);
		
		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}

static irqreturn_t interrupt_transfer(struct master_data *drv_data, 
					struct spi_msg *msg, 
					struct transfer_state *state)
{
	u32 sssr = drv_data->sssr;
	u32 ssdr = drv_data->ssdr;
	u32 sscr1 = drv_data->sscr1;
	u32 ssto = drv_data->ssto;
	u32 irq_status;
	
	while ((irq_status = GET_IRQ_STATUS(sssr))) {

		if (irq_status & SSSR_ROR) {

			/* Clear and disable interrupts */
			__REG(ssto) = 0;
			__REG(sssr) = SSSR_TINT | SSSR_ROR;
			__REG(sscr1) &= ~(SSCR1_TIE | SSCR1_RIE | SSCR1_TINTE);
			
			flush(drv_data);
			
			printk(KERN_WARNING "interrupt_transfer: fifo overun, "
					"transfer=%p tx_len=%d rx_len=%d\n", 
					state->transfer, 
					(state->tx_end - state->tx), 
					(state->rx_end - state->rx)); 

			drv_data->ror_cnt++;
			state->transfer = PTR_ERR(-2);
			tasklet_schedule(&drv_data->pump_transfers);

			return IRQ_HANDLED;
		}
		
		/* Look for false positive timeout */
		if ((irq_status & SSSR_TINT) && (state->rx < state->rx_end))
			__REG(sssr) = SSSR_TINT;
			
		/* Pump data */
		state->read(sssr, ssdr, state);
		state->write(sssr, ssdr, state);
		
		/* Disable tx interrupt enable rx timeout */		
		if (state->tx == state->tx_end) {
			__REG(sscr1) &= ~SSCR1_TIE;
		}
			
		if ((irq_status & SSSR_TINT) || (state->rx == state->rx_end)) {
			
			/* Clear timeout */
			__REG(ssto) = 0;
			__REG(sssr) = SSSR_TINT | SSSR_ROR ;
			__REG(sscr1) &= ~(SSCR1_TIE | SSCR1_RIE | SSCR1_TINTE);
			
			if (spimsg_get_flags(state->transfer) & SPI_M_CSKEEP)
				state->cs_control(PXA2XX_CS_DEASSERT);

			/* Schedule transfer tasklet */
			state->transfer = spimsg_getnext(state->transfer);
			tasklet_schedule(&drv_data->pump_transfers);
			
			return IRQ_HANDLED;
		}
	}
	
	return IRQ_HANDLED;
}

static irqreturn_t ssp_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct master_data *drv_data = (struct master_data *)dev_id;
	struct transfer_state *state;
	struct spi_message *msg;
	
	if (!drv_data->cur_msg || !drv_data->cur_msg->state) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad message or transfer "
				"state in interrupt handler\n");
		return IRQ_NONE;
	}
	state = (struct transfer_state *)drv_data->cur_msg->state;
	msg = drv_data->cur_msg;

	return state->transfer_handler(drv_data, msg, state);
} 

static void pump_dma_transfers(unsigned long data)
{
	struct master_data *drv_data = (struct master_data *)data;
	struct spi_message *message = drv_data->cur_msg;
	struct chip_data *chip;
	struct transfer_state * state;
	struct spi_transfer *transfer;
	u32 sssr = drv_data->sssr;
	u32 sscr1 = drv_data->sscr1;
	u32 ssto = drv_data->ssto;
	u32 tx_dcd = DCMD_INCSRCADDR | DCMD_FLOWTRG;
	u32 rx_dcd = DCMD_INCTRGADDR | DCMD_FLOWSRC | DCMD_ENDIRQEN;
	u32 len;
	unsigned long flags;
	
	if (!message) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad pump_transfers "
				"schedule\n");
		tasklet_schedule(&drv_data->pump_messages);
		return;
	}
	
	state = (struct transfer_state *)message->state;
	if (!state) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad message state\n");
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);
		return;
	}
	
	chip = spi_get_ctldata(message->dev);
	if (!chip) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad chip data\n");
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);
		return;
	}
		
	/* Handle for abort */
	if (PTR_ERR(state->transfer) == -2) {
		spimsg_complete(message, -EIO);
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);		
		return;
	}

	/* Handle end of message */
	if (spimsg_getnext(state->transfer) == NULL) {
		if (!(spimsg_getflags(state->transfer) & SPI_M_CSKEEP))
			state->cs_control(PXA2XX_CS_DEASSERT);

		spimsg_complete(message, 0);
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);		
		return;
	}
	
	/* Handle start of message */
	if (PTR_ERR(state->transfer) == -1) {
		restore_state(drv_data, chip);
		flush(drv_data);
		state->transfer = message;
	}
	
#if 0
	/* Delay if requested at end of transfer*/
	if (state->index > 1) {
		transfer = message->transfers + (state->index - 1);
		if (transfer->delay_usecs)
			udelay(transfer->delay_usecs);
	}
#endif

	/* Setup the transfer state based on the type of transfer */	
	flush(drv_data);
	drv_data->trans_cnt++;
	state->cs_control = !chip->cs_control ? 
				null_cs_control : chip->cs_control;
	state->write = spimsg_get_flags(state->transfer) & SPI_M_WR ? 
			chip->write : null_writer;
	state->read = spimsg_get_flags(state->transfer) & SPI_M_RD ? 
			chip->read : null_reader;

	len = spimsg_get_buffer(state->transfer, &state->rx);
	state->tx = state->rx;
	/* Make sure DMA is enable and buffer are aligned on 8-byte boundary */
	if (chip->enable_dma && IS_DMA_ALIGNED(state->tx, state->rx)) {
		
		/* Ensure we have the correct interrupt handler */
		state->transfer_handler = dma_transfer;
		drv_data->dma_trans_cnt++;

		
		/* Disable cache on rx buffer */
		transfer->rx_dma = 
			dma_map_single(&message->dev->dev, 
					state->rx, 
					len, 
					DMA_FROM_DEVICE);
					
		if (dma_mapping_error(transfer->rx_dma))
			goto out_mapping_error;
		
		/* Setup rx DMA Channel */
		DCSR(drv_data->rx_channel) = RESET_DMA_CHANNEL;
		DSADR(drv_data->rx_channel) = drv_data->ssdr;
		DTADR(drv_data->rx_channel) = transfer->rx_dma;
		DCMD(drv_data->rx_channel) = rx_dcd 
						| chip->dma_width 
						| chip->dma_burst_size 
						| len;
							
		/* Use local DMA if no tx buf specified */
		if (transfer->tx_buf == NULL) {
			*(u32 *)(drv_data->null_dma_buf) = 0;
			state->tx = drv_data->null_dma_buf;
			tx_map_len = 4;
			tx_dcd &= ~DCMD_INCSRCADDR;
		}
		else
			tx_map_len = transfer->len & DCMD_LENGTH;
		
		/* Disable cache on tx buffer */
		transfer->tx_dma = 
			dma_map_single(&message->dev->dev, 
					state->tx, 
					len, 
					DMA_TO_DEVICE);

		if (dma_mapping_error(transfer->tx_dma))
			goto out_mapping_error;

		/* Setup tx DMA Channel */
		DCSR(drv_data->tx_channel) = RESET_DMA_CHANNEL;
		DSADR(drv_data->tx_channel) = transfer->tx_dma;
		DTADR(drv_data->tx_channel) = drv_data->ssdr;
		DCMD(drv_data->tx_channel) = tx_dcd 
						| chip->dma_width 
						| chip->dma_burst_size 
						| len;

		/* Fix me, need to handle cs polarity */
		state->cs_control(PXA2XX_CS_ASSERT);

		/* Go baby, go */
		local_irq_save(flags);
		__REG(sssr) = SSSR_TINT | SSSR_ROR;
		DCSR(drv_data->rx_channel) |= DCSR_RUN;
		DCSR(drv_data->tx_channel) |= DCSR_RUN;
		__REG(ssto) = chip->timeout;
		__REG(sscr1) = chip->cr1 
				| chip->dma_threshold 
				| SSCR1_TSRE 
				| SSCR1_RSRE 
				| SSCR1_TINTE;
		local_irq_restore(flags);
	}
	else {
		if (len > 16)
			drv_data->bigunalign_cnt++;
		
		/* Ensure we have the correct interrupt handler	*/	
		state->transfer_handler = interrupt_transfer;
		drv_data->int_trans_cnt++;

		/* Fix me, need to handle cs polarity */
		state->cs_control(PXA2XX_CS_ASSERT);
	
		/* Go baby, go */
		local_irq_save(flags);
		__REG(sssr) = SSSR_TINT | SSSR_ROR;
		__REG(ssto) = chip->timeout;
		__REG(sscr1) = chip->cr1 
				| chip->threshold 
				| SSCR1_TIE 
				| SSCR1_RIE 
				| SSCR1_TINTE;
		local_irq_restore(flags);
	}
	
	return;
	
out_mapping_error:
	printk(KERN_ERR "problem mapping dma buf\n");
	spimsg_complete(message, -EIO);
	drv_data->cur_msg = NULL;
	tasklet_schedule(&drv_data->pump_messages);		
	return;		
}

static void pump_int_transfers(unsigned long data)
{
	struct master_data *drv_data = (struct master_data *)data;
	struct spi_msg *message = drv_data->cur_msg;
	struct chip_data *chip;
	struct transfer_state * state;
	u32 sssr = drv_data->sssr;
	u32 sscr1 = drv_data->sscr1;
	u32 ssto = drv_data->ssto;
	unsigned long flags;
	
	if (!message) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad pump_transfers "
				"schedule\n");
		tasklet_schedule(&drv_data->pump_messages);
		return;
	}
	
	state = (struct transfer_state *)message->state;
	if (!state) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad message state\n");
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);
		return;
	}
	
	chip = spi_get_ctldata(message->dev);
	if (!chip) {
		printk(KERN_ERR "pxs2xx_spi_ssp: bad chip data\n");
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);
		return;
	}
		
	/* Handle for abort */
	if (ERR_PTR(state->transfer) == -2) {
		spimsg_complete(message, -EIO);
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);		
		return;
	}

	/* Handle end of message */
	if (spimsg_getnext(state->transfer) == NULL) {
		if (spimsg_getflags(state->transfer) & SPI_M_CSKEEP)
			state->cs_control(PXA2XX_CS_DEASSERT);

		spimsg_complete(message, 0);
		drv_data->cur_msg = NULL;
		tasklet_schedule(&drv_data->pump_messages);		
		return;
	}
	
	/* Handle start of message */
	if (ERR_PTR(state->transfer) == -1) {
		restore_state(drv_data, chip);
		flush(drv_data);
		state->transfer = message;
	}

#if 0
	/* Delay if requested at end of transfer*/
	if (state->index > 1) {
		transfer = message->transfers + (state->index - 1);
		if (transfer->delay_usecs)
			udelay(transfer->delay_usecs);
	}
#endif
	
	/* Setup the transfer state based on the type of transfer */	
	flush(drv_data);
	drv_data->trans_cnt++;
	state->cs_control = chip->cs_control;
	state->write = spimsg_get_flags(state->transfer) & SPI_M_WR ? 
			chip->write : null_writer;
	state->read = spimsg_get_flags(state->transfer) & SPI_M_RD ? 
			chip->read : null_reader;

	len = spimsg_get_buffer(state->transfer, &state->rx);
	state->tx = state->rx;

	drv_data->int_trans_cnt++;

	/* Fix me, need to handle cs polarity */
	state->cs_control(PXA2XX_CS_ASSERT);

	/* Go baby, go */
	local_irq_save(flags);
	__REG(sssr) = SSSR_TINT | SSSR_ROR;
	__REG(ssto) = chip->timeout;
	__REG(sscr1) = chip->cr1 
			| chip->threshold 
			| SSCR1_TIE 
			| SSCR1_RIE 
			| SSCR1_TINTE;
	local_irq_restore(flags);
}

static void pump_messages(unsigned long data)
{
	struct master_data *drv_data = (struct master_data *)data;
	unsigned long flags;

	spin_lock_irqsave(&drv_data->lock, flags);

	/* Check for list empty */	
	if (list_empty(&drv_data->queue)) {
		spin_unlock(&drv_data->lock);
		return;
	}
	
	/* Check to see if we are already running */
	if (drv_data->cur_msg) {
		spin_unlock(&drv_data->lock);
		return;
	}		

	/* Extract head of queue and check for tasklet reschedule */
	drv_data->cur_msg = list_entry(drv_data->queue.next, 
					struct spi_msg, queue);
	list_del_init(&drv_data->cur_msg->queue);
	
	/* Setup message transfer and schedule transfer pump */
	drv_data->cur_msg->state = &drv_data->cur_state;
	drv_data->cur_state.index = -1;
	drv_data->cur_state.len = 0;
	tasklet_schedule(&drv_data->pump_transfers);
			
	spin_unlock_irqrestore(&drv_data->lock, flags);
}

static int transfer(struct spi_device *spi, struct spi_message *msg)
{
	struct master_data *drv_data = class_get_devdata(&spi->master->cdev);
	unsigned long flags;
	
	msg->actual_length = 0;
	msg->status = 0;
	
	spin_lock_irqsave(&drv_data->lock, flags);
	list_add_tail(&msg->queue, &drv_data->queue);
	spin_unlock_irqrestore(&drv_data->lock, flags);
	
	tasklet_schedule(&drv_data->pump_messages);
	
	return 0;
}

static int setup(struct spi_device *spi)
{
	struct pxa2xx_spi_chip *chip_info;
	struct chip_data *chip;
	
	chip_info = (struct pxa2xx_spi_chip *)spi->platform_data;
	
	/* Only alloc on first setup */
	chip = spi_get_ctldata(spi);
	if (chip == NULL) {
		chip = kcalloc(1, sizeof(struct chip_data), GFP_KERNEL);
		if (!chip)
			return -ENOMEM;

		spi->mode = chip_info->mode;
		spi->bits_per_word = chip_info->bits_per_word;
	}
	
	chip->cs_control = chip_info->cs_control;
	chip->cr0 = SSCR0_SerClkDiv((MAX_SPEED_HZ / spi->max_speed_hz) + 2) 
			| SSCR0_Motorola 
			| SSCR0_DataSize(spi->bits_per_word) 
			| SSCR0_SSE
			| (spi->bits_per_word > 16 ? SSCR0_EDSS : 0);
	chip->cr1 = (((spi->mode & SPI_CPHA) != 0) << 4) 
			| (((spi->mode & SPI_CPOL) != 0) << 3)
			| (chip_info->enable_loopback ? SSCR1_LBM : 0);
	chip->to = 0;
	chip->psp = 0;
	chip->timeout = (chip_info->timeout_microsecs * 10000) / 2712;
	chip->threshold = SSCR1_RxTresh(chip_info->rx_threshold) 
				| SSCR1_TxTresh(chip_info->tx_threshold);

	chip->enable_dma = chip_info->dma_burst_size != 0;
	if (chip_info->dma_burst_size <= 8) {
		chip->dma_threshold = SSCR1_RxTresh(8) | SSCR1_TxTresh(8);
		chip->dma_burst_size = DCMD_BURST8;
	}
	else if (chip_info->dma_burst_size <= 16) {
		chip->dma_threshold = SSCR1_RxTresh(16) | SSCR1_TxTresh(16);
		chip->dma_burst_size = DCMD_BURST16;
	}
	else {
		chip->dma_threshold = SSCR1_RxTresh(32) | SSCR1_TxTresh(32);
		chip->dma_burst_size = DCMD_BURST32;
	}
	
	if (spi->bits_per_word <= 8) {
		chip->n_bytes = 1;
		chip->dma_width = DCMD_WIDTH1;
		chip->read = u8_reader;
		chip->write = u8_writer;
	}
	else if (spi->bits_per_word <= 16) {
		chip->n_bytes = 2;
		chip->dma_width = DCMD_WIDTH2;
		chip->read = u16_reader;
		chip->write = u16_writer;
	}
	else if (spi->bits_per_word <= 32) {
		chip->n_bytes = 4;
		chip->dma_width = DCMD_WIDTH4;
		chip->read = u32_reader;
		chip->write = u32_writer;
	}
	else {
		printk(KERN_ERR "pxa2xx_spi_ssp: invalid wordsize\n");
		kfree(chip);
		return -ENODEV;
	}
		
	spi_set_ctldata(spi, chip);
	
	return 0;
}

static void cleanup(const struct spi_device *spi)
{
	struct chip_data *chip = spi_get_ctldata((struct spi_device *)spi);
	
	if (chip)
		kfree(chip);
	
	dev_dbg(&spi->dev, "spi_device %u.%u cleanup\n", 
				spi->master->bus_num, spi->chip_select);
}

static ssize_t trans_cnt_show(struct device *dev, char *buf)
{	
	struct spi_master *master = dev_get_drvdata(dev);
	struct master_data *drv_data = class_get_devdata(&master->cdev);
	
	return snprintf(buf, PAGE_SIZE, "%u\n", drv_data->trans_cnt);
}
DEVICE_ATTR(transfer_count, 0444, trans_cnt_show, NULL);

static ssize_t dma_trans_cnt_show(struct device *dev, char *buf)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct master_data *drv_data = class_get_devdata(&master->cdev);
	
	return snprintf(buf, PAGE_SIZE, "%u\n", drv_data->dma_trans_cnt);
}
DEVICE_ATTR(dma_transfer_count, 0444, dma_trans_cnt_show, NULL);

static ssize_t int_trans_cnt_show(struct device *dev, char *buf)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct master_data *drv_data = class_get_devdata(&master->cdev);
	
	return snprintf(buf, PAGE_SIZE, "%u\n", drv_data->int_trans_cnt);
}
DEVICE_ATTR(interrupt_transfer_count, 0444, int_trans_cnt_show, NULL);

static ssize_t bigunalign_cnt_show(struct device *dev, char *buf)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct master_data *drv_data = class_get_devdata(&master->cdev);
	
	return snprintf(buf, PAGE_SIZE, "%u\n", drv_data->bigunalign_cnt);
}
DEVICE_ATTR(big_unaligned_count, 0444, bigunalign_cnt_show, NULL);

static ssize_t buserr_cnt_show(struct device *dev, char *buf)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct master_data *drv_data = class_get_devdata(&master->cdev);
	
	return snprintf(buf, PAGE_SIZE, "%u\n", drv_data->buserr_cnt);
}
DEVICE_ATTR(bus_error_count, 0444, buserr_cnt_show, NULL);

static ssize_t ror_cnt_show(struct device *dev, char *buf)
{
	struct spi_master *master = dev_get_drvdata(dev);
	struct master_data *drv_data = class_get_devdata(&master->cdev);
	
	return snprintf(buf, PAGE_SIZE, "%u\n", drv_data->ror_cnt);
}
DEVICE_ATTR(overrun_count, 0444, ror_cnt_show, NULL);

static int probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct pxa2xx_spi_master *platform_info;
	struct master_data *drv_data = 0;
	struct resource *memory_resource;
	int irq;
	int status = 0;

	platform_info = (struct pxa2xx_spi_master *)pdev->dev.platform_data;

	if (!((drv_data = kzallloc(sizeof struct master_data + 16, SLAB_KERNEL))) != NULL) {
		status = -ENOMEM;
		goto out_error_memory;
	}

	dev_set_drvdata(dev, drv_data);

	INIT_LIST_HEAD(&drv_data->queue);
	spin_lock_init(&drv_data->lock);

	tasklet_init(&drv_data->pump_messages, 
			pump_messages, 
			(unsigned long)drv_data);

	drv_data->null_dma_buf = drv_data + sizeof(struct master_data);
	drv_data->null_dma_buf = (void *)(((u32)(drv_data->null_dma_buf)
					 & 0xfffffff8) | 8 );
	drv_data->trans_cnt = 0;
	drv_data->dma_trans_cnt = 0;
	drv_data->int_trans_cnt = 0;
	drv_data->bigunalign_cnt = 0;
	drv_data->buserr_cnt = 0;
	drv_data->ror_cnt = 0;
		
	/* Setup register addresses */
	memory_resource = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!memory_resource) {
		dev_dbg(dev, "can not find platform io memory\n");
		status = -ENODEV;
		goto out_error_memory;
	}
	
	drv_data->sscr0 = memory_resource->start + 0x00000000;
	drv_data->sscr1 = memory_resource->start + 0x00000004;
	drv_data->sssr = memory_resource->start + 0x00000008;
	drv_data->ssitr = memory_resource->start + 0x0000000c;
	drv_data->ssdr = memory_resource->start + 0x00000010;
	drv_data->ssto = memory_resource->start + 0x00000028;
	drv_data->sspsp = memory_resource->start + 0x0000002c;

	/* Attach to IRQ */
	irq = platform_get_irq(pdev, 0);
	if (irq == NO_IRQ) {
		dev_dbg(dev, "problem getting IORESOURCE_IRQ[0]\n");
		status = -ENODEV;
		goto out_error_memory;
	}
	
	status = request_irq(irq, ssp_int, SA_INTERRUPT, dev->bus_id, drv_data);
	if (status < 0) {
		dev_dbg(dev, "problem requesting IORESOURCE_IRQ %u\n", irq);
		goto out_error_memory;
	}

	/* Setup DMA is requested */	
	if (platform_info->enable_dma) {
		
		tasklet_init(&drv_data->pump_transfers, 
				pump_dma_transfers, 
				(unsigned long)drv_data);
		
		/* Get two DMA channels	(rx and tx) */
		drv_data->rx_channel = pxa_request_dma("pxa2xx_spi_ssp_rx", 
								DMA_PRIO_HIGH,
								dma_handler, 
								drv_data);							
		if (drv_data->rx_channel < 0) {
			dev_err(dev, "problem (%d) requesting rx channel\n",
					drv_data->rx_channel); 
			status = -ENODEV;
			goto out_error_irq;			
		}
		DRCMRRXSS2DR = DRCMR_MAPVLD 
				| (drv_data->rx_channel &  DRCMR_CHLNUM);
	
		drv_data->tx_channel = pxa_request_dma("pxa2xx_spi_ssp_tx", 
								DMA_PRIO_MEDIUM, 
								dma_handler, 
								drv_data);
		if (drv_data->tx_channel < 0) {
			dev_err(dev, "problem (%d) requesting tx channel\n",
					drv_data->tx_channel); 
			status = -ENODEV;
			goto out_error_rx_dma;			
		}
		DRCMRTXSS2DR = DRCMR_MAPVLD 
				| (drv_data->tx_channel &  DRCMR_CHLNUM);
	}
	else 
		/* No DMA use interrupt transfers */
		tasklet_init(&drv_data->pump_transfers, 
				pump_int_transfers, 
				(unsigned long)drv_data);
	
	/* Enable SOC clock */
	pxa_set_cken(platform_info->clock_enable, 1);
		
	/* Load default SSP configuration */
	__REG(drv_data->sscr0) = 0;
	__REG(drv_data->sscr1) = SSCR1_RxTresh(4) | SSCR1_TxTresh(12);
	__REG(drv_data->sscr0) = SSCR0_SerClkDiv(2) 
					| SSCR0_Motorola 
					| SSCR0_DataSize(8);
	__REG(drv_data->ssto) = 0;
	__REG(drv_data->sspsp) = 0;

	status = device_create_file(dev, &dev_attr_transfer_count);
	status += device_create_file(dev, &dev_attr_dma_transfer_count);
	status += device_create_file(dev, &dev_attr_interrupt_transfer_count);
	status += device_create_file(dev, &dev_attr_big_unaligned_count);
	status += device_create_file(dev, &dev_attr_bus_error_count);
	status += device_create_file(dev, &dev_attr_overrun_count);
	if (status != 0) {
		dev_err(dev, "problem creating attribute\n");
		status = -EIO;
		goto out_error_attr;
	}

	return status;

out_error_attr:

	device_remove_file(dev, &dev_attr_transfer_count);
	device_remove_file(dev, &dev_attr_dma_transfer_count);
	device_remove_file(dev, &dev_attr_interrupt_transfer_count);
	device_remove_file(dev, &dev_attr_big_unaligned_count);
	device_remove_file(dev, &dev_attr_bus_error_count);
	device_remove_file(dev, &dev_attr_overrun_count);

	pxa_free_dma(drv_data->tx_channel);
	
out_error_rx_dma:
	pxa_free_dma(drv_data->rx_channel);

out_error_irq:
	free_irq(irq, drv_data);
	
out_error_memory:

	return status;
}

static int remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct master_data *drv_data = dev_get_drvdata(dev);
	struct pxa2xx_spi_master *platform_info;
	
	int irq;
	unsigned long flags;
	
	device_remove_file(dev, &dev_attr_transfer_count);
	device_remove_file(dev, &dev_attr_dma_transfer_count);
	device_remove_file(dev, &dev_attr_interrupt_transfer_count);
	device_remove_file(dev, &dev_attr_big_unaligned_count);
	device_remove_file(dev, &dev_attr_bus_error_count);
	device_remove_file(dev, &dev_attr_overrun_count);
	
	platform_info = (struct pxa2xx_spi_master *)pdev->dev.platform_data;

	spin_lock_irqsave(&drv_data->lock, flags);

	__REG(drv_data->sscr0) = 0;
	pxa_set_cken(platform_info->clock_enable, 0);

	DRCMRRXSS2DR = 0;
	DRCMRTXSS2DR = 0;
	pxa_free_dma(drv_data->tx_channel);
	pxa_free_dma(drv_data->rx_channel);

	irq = platform_get_irq(pdev, 0);
	if (irq != 0)
		free_irq(irq, drv_data);
	
	spin_unlock_irqrestore(&drv_data->lock, flags);

	return 0;
}

static struct spi_bus_driver driver = {
	.driver = {
		.name = "pxa2xx-spi-ssp",
		.bus = &platform_bus_type,
		.owner = THIS_MODULE,
		.probe = probe,
		.remove = remove,
	},
	.queue = transfer,
};

static int pxa2xx_spi_ssp_init(void)
{
printk("MOIN SPI-PXA270.........................................\n");
	return spi_bus_driver_register(&driver);
}
module_init(pxa2xx_spi_ssp_init);

static void pxa2xx_spi_ssp_exit(void)
{
	spi_bus_driver_unregister(&driver);
}
module_exit(pxa2xx_spi_ssp_exit); 
