/* linux/drivers/spi/busses/spi-s3c2410.c
 *
 * Copyright (C) 2005
 *	Guillaume GOURAT <guillaume.gourat@nexvision.fr>
 *
 * S3C2410 SPI Controller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <linux/kernel.h>
#include <linux/module.h>

#include "../spi.h"
#include "../spi-id.h"

#include <linux/init.h>
#include <linux/time.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/device.h>

#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/io.h>

#include <asm/hardware/clock.h>
#include <asm/arch/regs-gpio.h>
#include <asm/arch/regs-spi.h>
#include <asm/arch/spi.h>

/* spi controller state */

enum s3c24xx_spi_state {
	STATE_IDLE,
	STATE_XFER,
	STATE_STOP
};

struct s3c24xx_spi {
	spinlock_t		lock;
	wait_queue_head_t	wait;

	struct spi_msg		*msg;
	unsigned int		msg_num;
	unsigned int		msg_idx;
	unsigned int		msg_ptr;

	enum s3c24xx_spi_state	state;

	void __iomem		*regs;
	struct clk		*clk;
	struct device		*dev;
	struct resource		*irq;
	struct resource		*ioarea;
	struct spi_adapter	adap;
};

/* default platform data to use if not supplied in the platform_device
*/

static struct s3c2410_platform_spi s3c24xx_spi_default_platform = {
	.flags		= 0,
	.bus_freq	= 2*1000*1000,	/* SPI-CLK */
	.min_freq	= 1*1000*1000,
	.max_freq	= 25*1000*1000,
};

/* s3c24xx_spi_get_platformdata
 *
 * get the platform data associated with the given device, or return
 * the default if there is none
*/

static inline struct s3c2410_platform_spi *s3c24xx_spi_get_platformdata(struct device *dev)
{
	if (dev->platform_data != NULL)
		return (struct s3c2410_platform_spi *)dev->platform_data;

	return &s3c24xx_spi_default_platform;
}

/* s3c24xx_spi_isbus0()
 *
 * return true if this is the bus 0 device (spi0)
*/

static inline int s3c24xx_spi_isbus0(struct s3c24xx_spi *spi)
{
	struct platform_device *pdev = to_platform_device(spi->dev);

	return pdev->id == 0;
}

/* s3c24xx_spi_master_complete
 *
 * complete the message and wake up the caller, using the given return code,
 * or zero to mean ok.
*/

static inline void s3c24xx_spi_master_complete(struct s3c24xx_spi *spi, int ret)
{
	dev_dbg(spi->dev, "master_complete %d\n", ret);

	spi->msg_ptr = 0;
	spi->msg = NULL;
	spi->msg_idx ++;
	spi->msg_num = 0;
	if (ret)
		spi->msg_idx = ret;

	wake_up(&spi->wait);
}

static inline void s3c24xx_spi_xfer_byte(struct s3c24xx_spi *spi)
{
	__u8 byte;

	if (spi->msg->wbuf)
		byte = spi->msg->wbuf[spi->msg_ptr++];
	else {
		byte = 0xff;
		spi->msg_ptr++;
	}

	do {
//		printk( KERN_INFO "s3c24xx_spi_xfer_byte: SPSTA =0x%x \n",readl(spi->regs + S3C2410_SPSTA)); /*JJ*/
	} while ((readl(spi->regs + S3C2410_SPSTA) & S3C2410_SPSTA_READY)== 0);
	writeb(byte, spi->regs + S3C2410_SPTDAT);
}


/* s3c24xx_spi_message_start
 *
 * configure the spi controller and transmit start of a message onto the bus
*/

static void s3c24xx_spi_message_start(struct s3c24xx_spi *spi)
{
	struct spi_msg *msg = spi->msg;
	unsigned long spcon;

	s3c2410_gpio_setpin(S3C2410_GPG2, 0);
	/* Wait for CS setup */

	spcon = S3C2410_SPCON_SMOD_INT | S3C2410_SPCON_ENSCK | S3C2410_SPCON_MSTR;

	if (msg->flags & SPI_M_CPOL_ACTLOW)
		spcon |= S3C2410_SPCON_CPOL_LOW;

	if (msg->flags & SPI_M_CPHA_FORMATB)
		spcon |= S3C2410_SPCON_CPHA_FMTB;

	/*if (msg->wbuf == NULL)
		spcon |= S3C2410_SPCON_TAGD;*/

	writel(spcon, spi->regs + S3C2410_SPCON);

	dev_dbg(spi->dev, "START: %02lx to SPCON\n", spcon);

	s3c24xx_spi_xfer_byte(spi);
}

static inline void s3c24xx_spi_stop(struct s3c24xx_spi *spi, int ret)
{
	unsigned long spcon;

	spcon = S3C2410_SPCON_SMOD_POLL | S3C2410_SPCON_MSTR;

	dev_dbg(spi->dev, "STOP\n");

	writel(spcon, spi->regs + S3C2410_SPCON);

	spi->state = STATE_STOP;

	s3c24xx_spi_master_complete(spi, ret);

	s3c2410_gpio_setpin(S3C2410_GPG2, 1);
}

/* helper functions to determine the current state in the set of
 * messages we are sending */

/* is_lastmsg()
 *
 * returns TRUE if the current message is the last in the set
*/

static inline int is_lastmsg(struct s3c24xx_spi *spi)
{
	return spi->msg_idx >= (spi->msg_num - 1);
}

/* is_msglast
 *
 * returns TRUE if this is the last byte in the current message
*/

static inline int is_msglast(struct s3c24xx_spi *spi)
{
	return spi->msg_ptr == spi->msg->len-1;
}

/* is_msgend
 *
 * returns TRUE if we reached the end of the current message
*/

static inline int is_msgend(struct s3c24xx_spi *spi)
{
	return spi->msg_ptr >= spi->msg->len;
}

/* spi_s3c_irq_nextbyte
 *
 * process an interrupt and work out what to do
 */

static void spi_s3c_irq_nextbyte(struct s3c24xx_spi *spi, unsigned long spsta)
{
	switch (spi->state) {

	case STATE_IDLE:
		dev_dbg(spi->dev, "%s: called in STATE_IDLE\n", __FUNCTION__);
		return;

	case STATE_STOP:
		dev_dbg(spi->dev, "%s: called in STATE_STOP\n", __FUNCTION__);
		return;

	case STATE_XFER:
		/* we are transfering data with the device... check for the
		 * end of the message, and if so, work out what to do
		 */

		/* Read incoming data if requested */
		if (spi->msg->rbuf) 
			spi->msg->rbuf[spi->msg_ptr-1] = readb(spi->regs + S3C2410_SPRDAT);
			/* spi->msg_ptr already incremented at transmission */

		if (!is_msgend(spi)) {
			/* Start next byte transfer */
			s3c24xx_spi_xfer_byte(spi);
		} else if (!is_lastmsg(spi)) {
			/* we need to go to the next spi message */
			dev_dbg(spi->dev, "XFER: Next Message\n");

			spi->msg_ptr = 0;
			spi->msg_idx ++;
			spi->msg++;

			s3c24xx_spi_message_start(spi);
		} else {
			/* stop after last byte of last message */
			s3c24xx_spi_stop(spi, 0);
		}
		break;

	}
}

/* s3c24xx_spi_irq
 *
 * top level IRQ servicing routine
*/

static irqreturn_t s3c24xx_spi_irq(int irqno, void *dev_id,
				   struct pt_regs *regs)
{
	struct s3c24xx_spi *spi = dev_id;
	unsigned long status;

	status = readl(spi->regs + S3C2410_SPSTA);

	if (status & S3C2410_SPSTA_DCOL) {
		// deal with arbitration loss
		dev_err(spi->dev, "Collision error detected\n");
	}

	/* pretty much this leaves us with the fact that we've
	 * transmitted or received whatever byte we last sent */

	if ((status & S3C2410_SPSTA_READY) == 0)
		return IRQ_HANDLED;

	spi_s3c_irq_nextbyte(spi, status);

 	return IRQ_HANDLED;
}


/* s3c24xx_spi_doxfer
 *
 * this starts an spi transfer
*/

static int s3c24xx_spi_doxfer(struct s3c24xx_spi *spi, struct spi_msg msgs[], int num)
{
	unsigned long timeout;
	int ret;

	spin_lock_irq(&spi->lock);

	spi->msg     = msgs;
	spi->msg_num = num;
	spi->msg_ptr = 0;
	spi->msg_idx = 0;
	spi->state   = STATE_XFER;

	s3c24xx_spi_message_start(spi);
	spin_unlock_irq(&spi->lock);

	timeout = wait_event_timeout(spi->wait, spi->msg_num == 0, HZ * 1); /*JJ was 5*/

	ret = spi->msg_idx;

	if (timeout == 0)
		dev_err(spi->dev, "timeout\n");
	else if (ret != num)
		dev_err(spi->dev, "incomplete xfer (%d)\n", ret);

 	return ret;
}

/* s3c24xx_spi_xfer
 *
 * first port of call from the spi bus code when an message needs
 * transfering across the spi bus.
*/

static int s3c24xx_spi_xfer(struct spi_adapter *adap,
			struct spi_msg msgs[], int num)
{
	struct s3c24xx_spi *spi = (struct s3c24xx_spi *)adap->algo_data;
	int retry;
	int ret;

	for (retry = 0; retry < adap->retries; retry++) {

		ret = s3c24xx_spi_doxfer(spi, msgs, num);

		if (ret != -EAGAIN)
			return ret;

		dev_dbg(spi->dev, "Retrying transmission (%d)\n", retry);

		udelay(100);
	}

	return -EREMOTEIO;
}

/* spi bus registration info */

static struct spi_algorithm s3c24xx_spi_algorithm = {
	.name			= "s3c2410-spi-algorithm",
	.master_xfer		= s3c24xx_spi_xfer,
};

static struct s3c24xx_spi s3c24xx_spi = {
	.lock	= SPIN_LOCK_UNLOCKED,
	.wait	= __WAIT_QUEUE_HEAD_INITIALIZER(s3c24xx_spi.wait),
	.adap	= {
		.name			= "s3c2410-spi",
		.owner			= THIS_MODULE,
		.algo			= &s3c24xx_spi_algorithm,
		.retries		= 2,
	},
};

/* s3c24xx_spi_calcdivisor
 *
 * return the divisor settings for a given frequency
*/

static int s3c24xx_spi_calcdivisor(unsigned long clkin, unsigned int wanted,
				   unsigned int *divs)
{
	unsigned int calc_divs = clkin / wanted;

	calc_divs /= 2;
	calc_divs -= 1;

	if (calc_divs > 256)
		calc_divs = 256;

	*divs = calc_divs;

	return clkin / ((calc_divs + 1) * 2);
}

/* freq_acceptable
 *
 * test wether a frequency is within the acceptable range of error
*/

static inline int freq_acceptable(unsigned int freq, unsigned int wanted)
{
	int diff = freq - wanted;

	return (diff >= -200 && diff <= 200);
}

/* s3c24xx_spi_getdivisor
 *
 * work out a divisor for the user requested frequency setting,
 * either by the requested frequency, or scanning the acceptable
 * range of frequencies until something is found
*/

static int s3c24xx_spi_getdivisor(struct s3c24xx_spi *spi,
				  struct s3c2410_platform_spi *pdata,
				  unsigned long *pre,
				  unsigned int *got)
{
	unsigned long clkin = clk_get_rate(spi->clk);
	unsigned int divs;
	int freq;
	int start, end;

	clkin /= 1000;		/* clkin now in KHz */

	dev_dbg(spi->dev,  "pdata %p, freq %lu %lu..%lu\n",
		 pdata, pdata->bus_freq, pdata->min_freq, pdata->max_freq);

	if (pdata->bus_freq != 0) {
		freq = s3c24xx_spi_calcdivisor(clkin, pdata->bus_freq/1000,
					       &divs);
		if (freq_acceptable(freq, pdata->bus_freq/1000))
			goto found;
	}

	/* ok, we may have to search for something suitable... */

	start = (pdata->max_freq == 0) ? pdata->bus_freq : pdata->max_freq;
	end = pdata->min_freq;

	start /= 1000;
	end /= 1000;

	/* search loop... */

	for (; start > end; start--) {
		freq = s3c24xx_spi_calcdivisor(clkin, start, &divs);
		if (freq_acceptable(freq, start))
			goto found;
	}

	/* cannot find frequency spec */

	return -EINVAL;

 found:
	*got = freq;
	*pre = divs;
	return 0;
}

/* s3c24xx_spi_init
 *
 * initialise the controller, set the IO lines and frequency
*/

static int s3c24xx_spi_init(struct s3c24xx_spi *spi)
{
	unsigned long sppin, sppre;
	struct s3c2410_platform_spi *pdata;
	unsigned int freq;

	/* get the platform data */

	pdata = s3c24xx_spi_get_platformdata(spi->adap.dev.parent);

	/* initialise the gpio */
	if (s3c24xx_spi_isbus0(spi))
	{
		/* spi0 */
		s3c2410_gpio_cfgpin(S3C2410_GPE11, S3C2410_GPE11_SPIMISO0);
		s3c2410_gpio_cfgpin(S3C2410_GPE12, S3C2410_GPE12_SPIMOSI0);
		s3c2410_gpio_cfgpin(S3C2410_GPE13, S3C2410_GPE13_SPICLK0);
		s3c2410_gpio_setpin(S3C2410_GPG2, 1);
		s3c2410_gpio_cfgpin(S3C2410_GPG2, S3C2410_GPG2_OUTP);
		s3c2410_gpio_pullup(S3C2410_GPG2, 0);
		s3c2410_gpio_pullup(S3C2410_GPE11, 1);
		s3c2410_gpio_pullup(S3C2410_GPE12, 0);
		s3c2410_gpio_pullup(S3C2410_GPE13, 0);
	}
	else
	{
		/* spi1 */
		s3c2410_gpio_cfgpin(S3C2410_GPG5, S3C2410_GPG5_SPIMISO1);
		s3c2410_gpio_cfgpin(S3C2410_GPG6, S3C2410_GPG6_SPIMOSI1);
		s3c2410_gpio_cfgpin(S3C2410_GPG7, S3C2410_GPG7_SPICLK1);
	}

	/* we need to work out the divisors for the clock... */

	if (s3c24xx_spi_getdivisor(spi, pdata, &sppre, &freq) != 0) {
		dev_err(spi->dev, "cannot meet bus frequency required\n");
		return -EINVAL;
	}

	dev_info(spi->dev, "bus frequency set to %d KHz\n", freq);

	/* initialise spi registers */
	sppin = 2;
	sppin |= S3C2410_SPPIN_KEEP;
	writel(sppin, spi->regs + S3C2410_SPPIN);
	writel(sppre, spi->regs + S3C2410_SPPRE);

	return 0;
}

static void s3c24xx_spi_free(struct s3c24xx_spi *spi)
{
	if (spi->clk != NULL && !IS_ERR(spi->clk)) {
		clk_disable(spi->clk);
		clk_unuse(spi->clk);
		clk_put(spi->clk);
		spi->clk = NULL;
	}

	if (spi->regs != NULL) {
		iounmap(spi->regs);
		spi->regs = NULL;
	}

	if (spi->ioarea != NULL) {
		release_resource(spi->ioarea);
		kfree(spi->ioarea);
		spi->ioarea = NULL;
	}
}

/* s3c24xx_spi_probe
 *
 * called by the bus driver when a suitable device is found
*/

static int s3c24xx_spi_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct s3c24xx_spi *spi;/* = &s3c24xx_spi;*/
	struct resource *res;
	int ret;

	/* find the clock and enable it */
	spi = kmalloc(sizeof(*spi), GFP_KERNEL);
	if (!spi)
		return -ENOMEM;

	memset(spi, 0, sizeof(*spi));
	spi->lock = SPIN_LOCK_UNLOCKED;
	init_waitqueue_head(&spi->wait);
	spi->adap.algo = &s3c24xx_spi_algorithm;
	spi->adap.retries = 2;

	spi->dev = dev;
	spi->clk = clk_get(dev, "spi");
	if (IS_ERR(spi->clk)) {
		dev_err(dev, "cannot get clock\n");
		ret = -ENOENT;
		goto out;
	}

	dev_dbg(dev, "clock source %p\n", spi->clk);

	clk_use(spi->clk);
	clk_enable(spi->clk);

	/* map the registers */

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res == NULL) {
		dev_err(dev, "cannot find IO resource\n");
		ret = -ENOENT;
		goto out;
	}

	spi->ioarea = request_mem_region(res->start, (res->end-res->start)+1,
					 pdev->name);

	if (spi->ioarea == NULL) {
		dev_err(dev, "cannot request IO\n");
		ret = -ENXIO;
		goto out;
	}

	spi->regs = ioremap(res->start, (res->end-res->start)+1);

	if (spi->regs == NULL) {
		dev_err(dev, "cannot map IO\n");
		ret = -ENXIO;
		goto out;
	}

	dev_dbg(dev, "registers %p (%p, %p)\n", spi->regs, spi->ioarea, res);

	/* setup info block for the spi core */

	spi->adap.algo_data = spi;
	spi->adap.dev.parent = dev;

	/* initialise the spi controller */

	ret = s3c24xx_spi_init(spi);
	if (ret != 0)
		goto out;

	/* find the IRQ for this unit (note, this relies on the init call to
	 * ensure no current IRQs pending
	 */

	res = platform_get_resource(pdev, IORESOURCE_IRQ, 0);
	if (res == NULL) {
		dev_err(dev, "cannot find IRQ\n");
		ret = -ENOENT;
		goto out;
	}

	ret = request_irq(res->start, s3c24xx_spi_irq, SA_INTERRUPT,
			  pdev->name, spi);

	if (ret != 0) {
		dev_err(dev, "cannot claim IRQ\n");
		goto out;
	}

	spi->irq = res;

	dev_dbg(dev, "irq resource %p (%ld)\n", res, res->start);

	ret = spi_add_adapter(&spi->adap);
	if (ret < 0) {
		dev_err(dev, "failed to add bus to spi core\n");
		goto out;
	}

	dev_set_drvdata(dev, spi);

	dev_info(dev, "%s: S3C SPI adapter\n", spi->adap.dev.bus_id);

 out:
	if (ret < 0) {
		s3c24xx_spi_free(spi);
		kfree(spi);/*JJ*/
	}

	return ret;
}

/* s3c24xx_spi_remove
 *
 * called when device is removed from the bus
*/

static int s3c24xx_spi_remove(struct device *dev)
{
	struct s3c24xx_spi *spi = dev_get_drvdata(dev);

	if (spi != NULL) {
		s3c24xx_spi_free(spi);
		dev_set_drvdata(dev, NULL);
		kfree(spi); /*JJ*/
	}

	return 0;
}

#ifdef CONFIG_PM
static int s3c24xx_spi_resume(struct device *dev, u32 level)
{
	struct s3c24xx_spi *spi = dev_get_drvdata(dev);

	if (spi != NULL && level == RESUME_ENABLE) {
		dev_dbg(dev, "resume: level %d\n", level);
		s3c24xx_spi_init(spi);
	}

	return 0;
}

#else
#define s3c24xx_spi_resume NULL
#endif

/* device driver for platform bus bits */

static struct device_driver s3c2410_spi_driver = {
	.name		= "s3c2410-spi",
	.bus		= &platform_bus_type,
	.probe		= s3c24xx_spi_probe,
	.remove		= s3c24xx_spi_remove,
	.resume		= s3c24xx_spi_resume,
};

static struct device_driver s3c2440_spi_driver = {
	.name		= "s3c2440-spi",
	.bus		= &platform_bus_type,
	.probe		= s3c24xx_spi_probe,
	.remove		= s3c24xx_spi_remove,
	.resume		= s3c24xx_spi_resume,
};

static char banner[] __initdata = KERN_INFO "Samsung s3c24xx SPI drivers, (c) 2005 Guillaume GOURAT / NexVision\n";

static int __init spi_adap_s3c_init(void)
{
	int ret;

	printk(banner);

	ret = driver_register(&s3c2410_spi_driver);
	if (ret == 0)
		ret = driver_register(&s3c2440_spi_driver);

	return ret;

}

static void __exit spi_adap_s3c_exit(void)
{
	driver_unregister(&s3c2410_spi_driver);
	driver_unregister(&s3c2440_spi_driver);
}

module_init(spi_adap_s3c_init);
module_exit(spi_adap_s3c_exit);

MODULE_DESCRIPTION("S3C24XX SPI Bus driver");
MODULE_AUTHOR("Guillaume GOURAT, <guillaume.gourat@nexvision.fr>");
MODULE_LICENSE("GPL");
