/*
 *  drivers/spi/spi-thread.c
 *
 *  Copyright (C) 2005 MontaVista Software, Inc <sources@mvista.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/proc_fs.h>
#include <linux/kmod.h>
#include <linux/init.h>
#include <linux/wait.h>
#include <linux/kthread.h>
#include <linux/spi.h>
#include <asm/atomic.h>
#include "spi-core.h"

static int spi_thread(void *context);

struct threaded_async_data {
	atomic_t exiting;
	struct device *dev;
	struct task_struct *thread;
	wait_queue_head_t wq;
};

/**
 * __spi_start_async - start the thread
 * @dev: device which the thread is related to
 * @return: abstract pointer to the thread context
 */
void *__spi_start_async (struct device *dev)
{
	struct threaded_async_data *td = kmalloc (sizeof (struct threaded_async_data), GFP_KERNEL);

	if (!td)
		return NULL;

	td->dev = dev;
	atomic_set(&td->exiting, 0);
	td->thread = kthread_run(spi_thread, td, "%s-work", dev->bus_id);
	init_waitqueue_head(&td->wq);
	return td;
}

/**
 * __spi_stop_async - stop the thread
 * @dev: device which the thread is related to
 * @ctx: abstract pointer to the thread context
 */
void __spi_stop_async (struct device *dev, void *ctx)
{
	struct threaded_async_data *td = ctx;

	if (ctx) {
		atomic_inc (&td->exiting);
		kthread_stop(td->thread);
		kfree(td);
	}
}

/**
 * spi_thread_awake - function that called to determine if thread needs to process any messages
 * @td: pointer to struct threaded_async_data
 * Description:
 * 	Thread wakes up if there is signal to exit (bd->exiting is set)
 * 	or there are any messages in bus' queue.
 */
static int spi_thread_awake(struct threaded_async_data *td)
{
	int ret = -EINVAL;
	struct spi_bus_data *bd = td->dev->platform_data;

	if (atomic_read(&td->exiting)) {
		return 1;
	}

	if (bd) {
		spin_lock_irq(&bd->lock);
		ret = !list_empty(&bd->msgs);
		spin_unlock_irq(&bd->lock);
	}
	return ret;
}

/**
 * spi_bus_next_msg - retrieve the next message
 * @this: spi_bus_driver that needs to retrieve next message from queue
 * @data: pointer to spi_bus_data structure associated with spi_bus_driver
 */
static struct spi_msg *spi_bus_next_msg(struct spi_bus_driver *this, struct spi_bus_data *data)
{
	return list_entry(data->msgs.next, struct spi_msg, link);
}

/**
 * spi_thread - the thread that calls bus functions to perform actual transfers
 * @context: pointer to struct spi_bus_data with bus-specific data
 * Description:
 * 	This function is started as separate thread to perform actual
 * 	transfers on SPI bus
 **/
static int spi_thread(void *context)
{
	struct threaded_async_data *td = context;
	struct spi_bus_data *bd = td->dev->platform_data;
	struct spi_msg *cmsg = NULL;
	int xfer_status;

	while (!kthread_should_stop()) {

		wait_event_interruptible(td->wq, spi_thread_awake(td));

		if (atomic_read(&td->exiting))
			goto thr_exit;

		spin_lock_irq(&bd->lock);
		while (!list_empty(&bd->msgs)) {
			/*
			 * this part is locked by bus_data->lock,
			 * to protect spi_msg extraction
			 */
			cmsg = spi_bus_next_msg(bd->bus, bd);
			list_del (&cmsg->link);

			spin_unlock_irq(&bd->lock);

			/*
			 * and this part is locked by device's lock;
			 * spi_queue will be able to queue new
			 * messages
			 *
			 * note that bd->selected_device is locked,
			 * not msg->device
			 * they are the same, but msg can be freed in
			 * msg->status function
			 */
			bd->selected_device = spimsg_get_spidev(cmsg);
			spi_device_lock(bd->selected_device);
			if (bd->bus->set_clock && cmsg->clock)
				bd->bus->set_clock(cmsg->device->dev.parent,
						cmsg->clock);
			xfer_status = bd->bus->xfer(cmsg);
			if (cmsg->status)
				cmsg->status(cmsg, xfer_status);

			spi_device_unlock(bd->selected_device);

			/* lock the bus_data again... */
			spin_lock_irq(&bd->lock);
		}
		spin_unlock_irq(&bd->lock);
	}

thr_exit:
	return 0;
}

/**
 * __spi_queue - (internal) queue the message to be processed asynchronously
 * @msg: message to be sent
 * Description:
 * 	This function queues the message to spi bus driver's queue.
 */
int __spi_queue(struct spi_msg *msg)
{
	struct device *dev = &msg->device->dev;
	struct spi_bus_data *pd = dev->parent->platform_data;
	struct threaded_async_data *td = pd->async_data;

	spin_lock_irq(&pd->lock);
	list_add_tail(&msg->link, &pd->msgs);
	dev_dbg(dev->parent, "message has been queued\n");
	spin_unlock_irq(&pd->lock);
	wake_up_interruptible(&td->wq);
	return 0;
}


