/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: ns9750fb.c,v 1.6 2006-11-02 14:28:15 mpietrek Exp $
 * @Author: Markus Pietrek
 * @Descr: LCD driver for the NS9750
 * @Usage: add video=ns9750fb:<LCD display> with
 *             LCD_display = sharp_lq57
 * @References: [1] NS9750 Hardware Reference, March 2004, Chap. 12
 *              [2] h1940fb.c
 *              [3] ModARM9/appNote_NS9750_TFTb.pdf
 * @TODO: * add ns_fb_blank (powersave)
 *        * cross-check with skeletonfb.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************
 * @History:
 *   30/10/2006: MP:    Added palette configuration
 *   27/07/2005: PPH:   Fixed how the PCD fild is computed in timing
 *                      register 2.
 ***********************************************************************/
#include <linux/config.h>	/* system configuration */

#include <linux/module.h>	/* MODULE_* */
#include <linux/init.h>		/* __init */
#include <linux/fb.h>		/* fb_info */
#include <linux/ioport.h>	/* request_mem_region */
#include <asm/arch/ns9750_lcd.h> /* NS_LCD_* */
#include <asm/arch/ns9750_bbus.h> /* set_gpio_cfg */


#include <asm/arch/hardware.h>	/* request_gpio */

#define DEBUG_INIT		0x0001
#define DEBUG_MINOR		0x0002
#define DEBUG_ERROR		0x4000
#define DEBUG_ERROR_CRIT	0x8000

static int debug = 0xc000;

#define DEBUG_ARGS( FLG, a0, args... ) if( ( debug & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, ## args )
#define DEBUG_FN( FLG ) if( (debug & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
#define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }

#ifdef CONFIG_ARCH_NS9750
#define DRIVER_NAME		"ns9750fb"
#endif

#ifdef CONFIG_ARCH_NS9360
#define DRIVER_NAME		"ns9360fb"
#endif

#define SHARP_LQ57_WIDTH	320
#define SHARP_LQ57_HEIGHT	240

/* general configuration for multiple TFT displays */
struct lcd_cfg
{
	const char* name;	   /* name used in kernel command line */
	int width;
	int height;
	int clock;		   /* LCD clock in Hz */
	unsigned long control;	   /* register value */
	unsigned long timing[ 4 ]; /* register value */
};

/* supporting only RGB565 with wiring for RGB555 TFT displays */
static struct lcd_cfg lcd_cfg[] = {
	{
		name: "sharp_lq57",
		width: SHARP_LQ57_WIDTH,
		height: SHARP_LQ57_HEIGHT,
		clock: 7000000,
		control: (NS_LCD_CONTROL_WATERMARK |
			  NS_LCD_CONTROL_PWR |
			  NS_LCD_CONTROL_BGR |
			  NS_LCD_CONTROL_TFT |
			  NS_LCD_CONTROL_BPP_16 |
			  NS_LCD_CONTROL_EN),
		timing: {
			(NS_LCD_TIMING_0_HBP( 8 ) |
			 NS_LCD_TIMING_0_HFP( 37 ) |
			 NS_LCD_TIMING_0_HFP( 2 ) |
			 NS_LCD_TIMING_0_PPL( (SHARP_LQ57_WIDTH/16)-1 )),
			(NS_LCD_TIMING_1_VBP( 5 ) |
			 NS_LCD_TIMING_1_VFP( 5 ) |
			 NS_LCD_TIMING_1_VSW( 0 ) |
			 NS_LCD_TIMING_1_LPP( SHARP_LQ57_HEIGHT-1)),
			(NS_LCD_TIMING_2_CPL( SHARP_LQ57_WIDTH-1 ) |
			 NS_LCD_TIMING_2_IHS |
			 NS_LCD_TIMING_2_IVS), /* PCD is added later */
			0
		}
	}
};

struct ns_fb_info {
	struct fb_info		fb;
	struct device		*dev;

	/* raw memory addresses */
	dma_addr_t		map_dma;	/* physical */
	u_char *		map_cpu;	/* virtual */
	u_int			map_size;

	/* addresses of pieces placed in raw buffer */
	u_char *		screen_cpu;	/* virtual address of frame buffer */
	dma_addr_t		screen_dma;	/* physical address of frame buffer */
	u32 pseudo_pal[16];
};

static struct ns_fb_info info;

static struct lcd_cfg* lcd_active = NULL;
static char ns_fb_lcd_name[ 20 ] = "";

/* [3], Chap.3.1, Column "NS9750 Signal" */
static struct ns_fb_gpio {
	int gpio;
	int function;
} ns_fb_gpios[] = { { 20, 1 },
	 { 18, NS_GPIO_CFG_FUNC_GPIO | NS_GPIO_CFG_OUTPUT },
	 { 19, NS_GPIO_CFG_FUNC_1 },
	 { 21, NS_GPIO_CFG_FUNC_1 },
	 { 25, NS_GPIO_CFG_FUNC_1 },
	 { 26, NS_GPIO_CFG_FUNC_1 },
	 { 27, NS_GPIO_CFG_FUNC_1 },
	 { 28, NS_GPIO_CFG_FUNC_1 },
	 { 29, NS_GPIO_CFG_FUNC_1 },
	 { 30, NS_GPIO_CFG_FUNC_1 },
	 { 31, NS_GPIO_CFG_FUNC_1 },
	 { 32, NS_GPIO_CFG_FUNC_2 },
	 { 33, NS_GPIO_CFG_FUNC_2 },
	 { 34, NS_GPIO_CFG_FUNC_2 },
	 { 35, NS_GPIO_CFG_FUNC_2 },
	 { 37, NS_GPIO_CFG_FUNC_2 },
	 { 38, NS_GPIO_CFG_FUNC_2 },
	 { 39, NS_GPIO_CFG_FUNC_2 },
	 { 40, NS_GPIO_CFG_FUNC_2 },
	 { 41, NS_GPIO_CFG_FUNC_2 },
	 { 22, NS_GPIO_CFG_FUNC_1 } };

extern void *dma_alloc_writecombine(struct device *dev, size_t size, dma_addr_t *handle, int gfp);

static int ns_fb_set_par( struct fb_info* info );
static int ns_fb_setcolreg(
        unsigned regno,
	unsigned red, unsigned green, unsigned blue,
	unsigned transp,
        struct fb_info* info );
static int __init ns_fb_lcd_locate( const char* name );
int __devinit ns_fb_init( void );
static void __exit ns_fb_cleanup( void );
int __init ns_fb_setup( char *options );

static struct fb_ops ns_fb_ops = {
	.owner		= THIS_MODULE,
	.fb_set_par	= ns_fb_set_par,
	.fb_setcolreg	= ns_fb_setcolreg,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
};


/* Fake monspecs to fill in fbinfo structure */
/* Don't know if the values are important    */
static struct fb_monspecs monspecs __initdata = {
	.hfmin	= 30000,
	.hfmax	= 70000,
	.vfmin	= 50,
	.vfmax	= 65,
};

static int ns_fb_set_par( struct fb_info* info )
{
	DEBUG_FN( DEBUG_INIT|DEBUG_MINOR );

	/* We support only 16BPP true color */
	info->fix.visual	  = FB_VISUAL_TRUECOLOR;
	info->fix.line_length     = lcd_active->width*16/8;

	return 0;
}

/***********************************************************************
 * @Function: ns_fb_setcolreg
 * @Return: return 0 if color could be set, otherwise -EINVAL
 * @Descr: fills palette
 ***********************************************************************/

static int ns_fb_setcolreg(
        unsigned regno,
	unsigned red, unsigned green, unsigned blue,
	unsigned transp,
        struct fb_info* info )
{
	struct ns_fb_info* fbi = (struct ns_fb_info *) info;
	int bpp;
        int m;

	DEBUG_ARGS( DEBUG_INIT | DEBUG_MINOR,
                    "setcolreg(reg=%d, RGBA=%d,%d,%d,%d info=%p)\n",
                    regno, red, green, blue, transp, info );

	bpp = fbi->fb.var.bits_per_pixel;
	m   = 1 << bpp;
	if( regno >= m )
		return -EINVAL;

	switch( bpp ) {
	case 16:
		/* RGB 555 */
		fbi->pseudo_pal[ regno ] = ( (   red   & 0xF800 ) |
                                             ( ( green & 0xF800  ) >> 5 ) |
                                             ( ( blue  & 0xF800 ) >> 10 ) );
		break;
	case 8:
                DEBUG_ARGS( DEBUG_ERROR_CRIT, " 8bit Not yet implemented\n" );
                break;
	}
        
	return 0;
}

/***********************************************************************
 * @Function: ns_fb_lcd_locate
 * @Return: 1 if an LCD display with name is registered otherwise 0
 * @Descr: lcd_active is set to
 ***********************************************************************/

static int __init ns_fb_lcd_locate( const char* name )
{
	int i;

	DEBUG_FN( DEBUG_INIT|DEBUG_MINOR );

	for( i = 0; i < ARRAY_SIZE( lcd_cfg ); i++ )
		if( !strcmp( name, lcd_cfg[ i ].name ) ) {
			lcd_active = &lcd_cfg[ i ];
			return 1;
		}

	return 0;		/* not found */
}


/***********************************************************************
 * @Function: ns_fb_map_video_memory
 * @Return: -ENOMEM on failure
 * @Descr: Allocated non-cached memory suitable for framebuffers
 ***********************************************************************/

static int __init ns_fb_map_video_memory( struct ns_fb_info* fbi )
{
	DEBUG_FN( DEBUG_INIT|DEBUG_MINOR );

	fbi->map_size = PAGE_ALIGN( fbi->fb.fix.smem_len + PAGE_SIZE );
	fbi->map_cpu = dma_alloc_writecombine( fbi->dev, fbi->map_size,
					       &fbi->map_dma, GFP_KERNEL );

	if( fbi->map_cpu ) {
		/* prevent initial garbage on screen */

		memset( fbi->map_cpu, 0x0, fbi->map_size );

		fbi->fb.screen_base = fbi->map_cpu + PAGE_SIZE;
		fbi->screen_dma = fbi->map_dma + PAGE_SIZE;
		fbi->fb.fix.smem_start = fbi->screen_dma;
		DEBUG_ARGS( DEBUG_INIT|DEBUG_MINOR,
			    "Video area mapped to 0x%08x\n", fbi->screen_dma );
	}

	return fbi->map_cpu ? 0 : -ENOMEM;
}

/***********************************************************************
 * @Function: ns_fb_init_registers
 * @Return: -EBUSY when gpio's are reserved
 * @Descr: Initializes GPIO pins and setup LCD registers
 ***********************************************************************/

int __devinit ns_fb_init_registers( struct ns_fb_info* fbi )
{
	int i;

	/* configure GPIO */
	for( i = 0; i < ARRAY_SIZE( ns_fb_gpios ); i++ )
		if( !request_gpio( ns_fb_gpios[ i ].gpio, 1, DRIVER_NAME ) )
			goto error;

	/* all gpios belong to us, configure them */
	for( i = 0; i < ARRAY_SIZE( ns_fb_gpios ); i++ )
		set_gpio_cfg( ns_fb_gpios[ i ].gpio, ns_fb_gpios[ i ].function );

	set_gpio_ctrl( 18, 0 );	/* power enable */

	lcd_active->timing[ 2 ] |=
		/* LCD Clock assumed to be set to AHB clock. Rounded up...
		 * should be subtracted 2) */
		NS_LCD_TIMING_2_PCD((AHB_CLK_FREQ / lcd_active->clock) - 1);

	for( i = 0; i < 5; i++ )
		writel( lcd_active->timing[ i ], NS_LCD_TIMING_0 + i * 4 );

	writel( fbi->fb.fix.smem_start, NS_LCD_UPBASE );

	/* Enable video */
	writel( lcd_active->control, NS_LCD_CONTROL );

	return 0;

error:
	while( i-- )
		release_gpio( ns_fb_gpios[ i ].gpio, 1 );

	return -EBUSY;
}

/***********************************************************************
 * @Function: ns_fb_init
 * @Return: -EBUSY when gpios are reserved, -ENOMEM when mapping not possible
 * @Descr: registers and initialized framebuffer driver
 ***********************************************************************/

int __devinit ns_fb_init( void )
{
	int ret = 0;
	const char* lcd_name = ns_fb_lcd_name;
	int i;

	/* set default driver for platform */
	if( !lcd_name[ 0 ] )
		lcd_name = lcd_cfg[ 0 ].name;

	if( !ns_fb_lcd_locate( lcd_name ) ) {
		printk( KERN_WARNING DRIVER_NAME
			": Couldn't locate LCD display %s\n",
			lcd_name );
		return -ENODEV;
	}

	printk( KERN_INFO DRIVER_NAME ": Using LCD display %s\n",
		lcd_active->name );

	strcpy( info.fb.fix.id, DRIVER_NAME );

	if(request_mem_region(NS_LCD_MODULE_BASE, 0x400, DRIVER_NAME) == NULL)
		goto error_region;

	info.fb.fix.type	    = FB_TYPE_PACKED_PIXELS;
	info.fb.fix.type_aux	    = 0;
	info.fb.fix.xpanstep	    = 0;
	info.fb.fix.ypanstep	    = 0;
	info.fb.fix.ywrapstep	    = 0;
	info.fb.fix.accel	    = FB_ACCEL_NONE;

	info.fb.var.nonstd	    = 0;
	info.fb.var.activate	    = FB_ACTIVATE_NOW;
	info.fb.var.height	    = lcd_active->height;
	info.fb.var.width	    = lcd_active->width;
	info.fb.var.accel_flags     = 0;
	info.fb.var.vmode	    = FB_VMODE_NONINTERLACED;

	info.fb.fbops		    = &ns_fb_ops;
	info.fb.flags		    = FBINFO_FLAG_DEFAULT;
	info.fb.monspecs	    = monspecs;
//	info.fb.currcon	    = 0;
	info.fb.pseudo_palette      = &info.pseudo_pal;

	info.fb.var.xres	    = lcd_active->width;
	info.fb.var.xres_virtual    = lcd_active->width;
	info.fb.var.yres	    = lcd_active->height;
	info.fb.var.yres_virtual    = lcd_active->height;
	info.fb.var.bits_per_pixel  = 16;

	info.fb.var.red.offset    = 10;
	info.fb.var.green.offset  = 5;
	info.fb.var.blue.offset   = 0;
	info.fb.var.red.length    = 5;
	info.fb.var.green.length  = 5;
	info.fb.var.blue.length   = 5;
	info.fb.var.transp.length = 0;

	info.fb.fix.smem_len	    = info.fb.var.xres * info.fb.var.yres *
				      info.fb.var.bits_per_pixel / 8;

	/* Initialize video memory */
	ret = ns_fb_map_video_memory( &info );
	if( ret ) {
		printk( KERN_WARNING DRIVER_NAME
			": Failed to allocate video RAM: %d\n", ret );
		ret = -ENOMEM;
		goto error_gpio;
	}

	ret = ns_fb_init_registers( &info );

	if( ret < 0 ) {
		printk( KERN_WARNING DRIVER_NAME ": Failed to reserve gpios\n" );
		goto error_gpio;
	}

	ret = register_framebuffer( &info.fb );
	if( ret < 0 ) {
		printk( KERN_WARNING DRIVER_NAME
			": Failed to register framebuffer device: %d\n", ret );
		goto error;
	}

	printk( KERN_INFO "fb%d: %s frame buffer device\n",
		info.fb.node, info.fb.fix.id );

	return 0;

error:
	for( i = 0; i < ARRAY_SIZE( ns_fb_gpios ); i++ )
		release_gpio( ns_fb_gpios[ i ].gpio, 1 );
error_gpio:
	release_mem_region(NS_LCD_MODULE_BASE, 0x400);
error_region:
	return ret;
}

static void __exit ns_fb_cleanup( void )
{
	int i;

	unregister_framebuffer( &info.fb );

	/* release GPIO but let them configured for LCD */
	for( i = 0; i < ARRAY_SIZE( ns_fb_gpios ); i++ )
		release_gpio( ns_fb_gpios[ i ].gpio, 1 );

	release_mem_region(NS_LCD_MODULE_BASE, 0x400);
}

#ifndef MODULE
int __init ns_fb_setup( char *options )
{
	strncpy( ns_fb_lcd_name, options, sizeof( ns_fb_lcd_name ) - 1 );
	ns_fb_lcd_name[ sizeof( ns_fb_lcd_name ) - 1 ] = 0;

	return 0;
}
#endif

module_init( ns_fb_init);
module_exit( ns_fb_cleanup);

MODULE_AUTHOR( "Markus Pietrek <mpietrek@fsforth.de>" );
MODULE_DESCRIPTION( "Framebuffer driver for the NetSilicon NS9750" );
MODULE_LICENSE( "GPL" );
