/*
 * drivers/video/ns9xxxfb.c
 *
 * Copyright (C) 2006 by Digi International Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <linux/config.h>

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fb.h>
#include <linux/ioport.h>
#include <asm/arch/regs-lcd.h>
#include <asm/arch/regs-bbus.h>
#include <asm/arch/regs-sys.h>
#include <asm/arch/system.h>
#include <asm/arch/hardware.h>	

#include "./displays/displays.h"

#define DEBUG_INIT		0x0001
#define DEBUG_MINOR		0x0002
#define DEBUG_ERROR		0x4000
#define DEBUG_ERROR_CRIT	0x8000

static int debug = (DEBUG_ERROR | DEBUG_ERROR_CRIT);

#define DEBUG_ARGS( FLG, a0, args... ) if( ( debug & (FLG) ) == (FLG)) \
		printk( KERN_INFO "%s: " a0, __FUNCTION__, ## args )
#define DEBUG_FN( FLG ) if( (debug & (FLG)) == (FLG) ) \
		printk( KERN_INFO "\r%s:line %d\n",__FUNCTION__,__LINE__)
#define ASSERT( expr, func ) if( !( expr ) ) { \
        	printk( KERN_ERR "Assertion failed! %s:line %d %s\n", \
			__FUNCTION__,__LINE__,#expr);		      \
        	func }

#define DRIVER_NAME		"ns9xxxfb"

static struct lcd_cfg lcd_cfg = {
	.name         =	DISP_NAME,
	.width        = DISP_XSIZE,
	.height	      = DISP_YSIZE,
	.clock	      =	DISP_CLK,
	.control      =	DISP_CONTROL,
	.timing	      =	{DISP_TIMING0, DISP_TIMING1, DISP_TIMING2, DISP_TIMING3},
};

struct ns9xxxfb_info {
	struct fb_info		fb;
	struct device		*dev;

	/* raw memory addresses */
	dma_addr_t		map_dma;	/* physical */
	u_char *		map_cpu;	/* virtual */
	u_int			map_size;

	/* addresses of pieces placed in raw buffer */
	u_char *		screen_cpu;	/* virtual address of frame buffer */
	dma_addr_t		screen_dma;	/* physical address of frame buffer */
	u32 pseudo_pal[16];
};

static struct ns9xxxfb_info info;

static struct lcd_cfg* lcd_active = NULL;


extern void *dma_alloc_writecombine(struct device *dev, size_t size, dma_addr_t *handle, int gfp);
extern void ns9xxxfb_lcd_power_enable(void);
extern void ns9xxxfb_lcd_power_disable(void);


static int ns9xxxfb_set_par( struct fb_info* info );
static int ns9xxxfb_setcolreg(
        unsigned regno,
	unsigned red, unsigned green, unsigned blue,
	unsigned transp,
        struct fb_info* info );
int __devinit ns9xxxfb_init( void );
static void __exit ns9xxxfb_cleanup( void );
int __init ns9xxxfb_setup( char *options );

static struct fb_ops ns_fb_ops = {
	.owner		= THIS_MODULE,
	.fb_set_par	= ns9xxxfb_set_par,
	.fb_setcolreg	= ns9xxxfb_setcolreg,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
};


/* Fake monspecs to fill in fbinfo structure */
/* Don't know if the values are important    */
static struct fb_monspecs monspecs __initdata = {
	.hfmin	= 30000,
	.hfmax	= 70000,
	.vfmin	= 50,
	.vfmax	= 65,
};

static int ns9xxxfb_set_par( struct fb_info* info )
{
	DEBUG_FN( DEBUG_INIT|DEBUG_MINOR );

	/* We support only 16BPP true color */
	info->fix.visual	  = FB_VISUAL_TRUECOLOR;
	info->fix.line_length     = lcd_active->width*16/8;

	return 0;
}

/***********************************************************************
 * @Function: ns9xxxfb_setcolreg
 * @Return: return 0 if color could be set, otherwise -EINVAL
 * @Descr: fills palette
 ***********************************************************************/

static int ns9xxxfb_setcolreg(unsigned regno, unsigned red, unsigned green,
			      unsigned blue, unsigned transp, struct fb_info* info )
{
	struct ns9xxxfb_info* fbi = (struct ns9xxxfb_info *)info;
	int bpp, m;

	DEBUG_ARGS( DEBUG_INIT | DEBUG_MINOR,
                    "setcolreg(reg=%d, RGBA=%d,%d,%d,%d info=%p)\n",
                    regno, red, green, blue, transp, info );

	bpp = fbi->fb.var.bits_per_pixel;
	m   = 1 << bpp;
	if( regno >= m )
		return -EINVAL;

	switch( bpp ) {
	case 16:
		/* RGB 555 */
		fbi->pseudo_pal[ regno ] = ( (   red   & 0xF800 ) |
                                             ( ( green & 0xF800  ) >> 5 ) |
                                             ( ( blue  & 0xF800 ) >> 10 ) );
		break;
	case 8:
                DEBUG_ARGS( DEBUG_ERROR_CRIT, " 8bit Not yet implemented\n" );
                break;
	}
        
	return 0;
}

/***********************************************************************
 * @Function: ns9xxxfb_map_video_memory
 * @Return: -ENOMEM on failure
 * @Descr: Allocated non-cached memory suitable for framebuffers
 ***********************************************************************/
static int __init ns9xxxfb_map_video_memory( struct ns9xxxfb_info* fbi )
{
	DEBUG_FN( DEBUG_INIT|DEBUG_MINOR );

	fbi->map_size = PAGE_ALIGN( fbi->fb.fix.smem_len + PAGE_SIZE );
	fbi->map_cpu = dma_alloc_writecombine( fbi->dev, fbi->map_size,
					       &fbi->map_dma, GFP_KERNEL );

	if( fbi->map_cpu ) {
		/* prevent initial garbage on screen */
		memset( fbi->map_cpu, 0x0, fbi->map_size );

		fbi->fb.screen_base = fbi->map_cpu + PAGE_SIZE;
		fbi->screen_dma = fbi->map_dma + PAGE_SIZE;
		fbi->fb.fix.smem_start = fbi->screen_dma;
		DEBUG_ARGS( DEBUG_INIT|DEBUG_MINOR,
			    "Video area mapped to 0x%08x\n", fbi->screen_dma );
	}

	return fbi->map_cpu ? 0 : -ENOMEM;
}

/***********************************************************************
 * @Function: ns9xxxfb_init_registers
 * @Return: -EBUSY when gpio's are reserved
 * @Descr: Initializes GPIO pins and setup LCD registers
 ***********************************************************************/
int __devinit ns9xxxfb_init_registers( struct ns9xxxfb_info* fbi )
{
	int i;
	u32 clcdclk;

	/* configure GPIO */
	for( i = 0; i < ARRAY_SIZE( ns9xxxfb_gpios ); i++ )
		if( ns9xxx_request_gpio( ns9xxxfb_gpios[i].gpio, DRIVER_NAME ) != 0 )
			goto error;

	/* all gpios belong to us, configure them */
	for( i = 0; i < ARRAY_SIZE( ns9xxxfb_gpios ); i++ )
		ns9xxx_gpio_cfgpin( ns9xxxfb_gpios[i].gpio, ns9xxxfb_gpios[i].function );

	ns9xxxfb_lcd_power_enable();

	if( lcd_active->clock == DISP_USE_EXT_CLK ) {
		/* Configure LCD clock for external clock */
		SYS_RSTSLEEP &= ~SYS_RST_LCDC;
		SYS_CCR = (SYS_CCR & ~SYS_CLK_LPCS) | SYS_CLK_LPCS_EXT;
		SYS_RSTSLEEP |= SYS_RST_LCDC;
	}

	if( !( lcd_active->timing[2] & LCD_TIMING_2_BCD ) ) {
		switch( SYS_CCR & SYS_CLK_LPCS ) {
			case SYS_CLK_LPCS_AHB:		clcdclk = ns9xxx_ahbclock();	break;
			case SYS_CLK_LPCS_AHB_DIV2:	clcdclk = ns9xxx_ahbclock()/2;	break;
			case SYS_CLK_LPCS_AHB_DIV4:	clcdclk = ns9xxx_ahbclock()/4;	break;
			case SYS_CLK_LPCS_AHB_DIV8:	clcdclk = ns9xxx_ahbclock()/8;	break;
			default:			clcdclk = lcd_active->clock;	break;
		}
		lcd_active->timing[2] |= LCD_TIMING_2_PCD((clcdclk/lcd_active->clock) - 1);	/* Rounded up, should substract 2 */
	}

	for( i = 0; i < 4; i++ )
		LCD_TIMINGX(i) = lcd_active->timing[i];

	LCD_UPBASE = fbi->fb.fix.smem_start;

	/* Enable video */
	LCD_CONTROL = lcd_active->control;

	return 0;

error:
	while( i-- )
		ns9xxx_release_gpio( ns9xxxfb_gpios[i].gpio );

	return -EBUSY;
}

/***********************************************************************
 * @Function: ns9xxxfb_init
 * @Return: -EBUSY when gpios are reserved, -ENOMEM when mapping not possible
 * @Descr: registers and initialized framebuffer driver
 ***********************************************************************/
int __devinit ns9xxxfb_init( void )
{
	int ret = 0, i;

	/* Active display is selected through kernel configuration */
	lcd_active = &lcd_cfg;

	if (lcd_active == NULL || lcd_active->name == NULL) {
		printk(KERN_WARNING DRIVER_NAME
		       ": Display not defined, use xconfig to select a display\n");
		return -ENODEV;
	}

	printk( KERN_INFO DRIVER_NAME ": Using LCD display %s\n",
		lcd_active->name );

	strcpy( info.fb.fix.id, DRIVER_NAME );

	if(request_mem_region(LCD_MODULE_BASE_VA, 0x400, DRIVER_NAME) == NULL) {
		ret = -EBUSY;
		goto error_region;
	}

	info.fb.fix.type	    = FB_TYPE_PACKED_PIXELS;
	info.fb.fix.type_aux	    = 0;
	info.fb.fix.xpanstep	    = 0;
	info.fb.fix.ypanstep	    = 0;
	info.fb.fix.ywrapstep	    = 0;
	info.fb.fix.accel	    = FB_ACCEL_NONE;

	info.fb.var.nonstd	    = 0;
	info.fb.var.activate	    = FB_ACTIVATE_NOW;
	info.fb.var.height	    = lcd_active->height;
	info.fb.var.width	    = lcd_active->width;
	info.fb.var.accel_flags     = 0;
	info.fb.var.vmode	    = FB_VMODE_NONINTERLACED;

	info.fb.fbops		    = &ns_fb_ops;
	info.fb.flags		    = FBINFO_FLAG_DEFAULT;
	info.fb.monspecs	    = monspecs;
	info.fb.pseudo_palette      = &info.pseudo_pal;

	info.fb.var.xres	    = lcd_active->width;
	info.fb.var.xres_virtual    = lcd_active->width;
	info.fb.var.yres	    = lcd_active->height;
	info.fb.var.yres_virtual    = lcd_active->height;
	info.fb.var.bits_per_pixel  = 16;

	info.fb.var.red.offset      = 10;
	info.fb.var.green.offset    = 5;
	info.fb.var.blue.offset     = 0;
	info.fb.var.red.length      = 5;
	info.fb.var.green.length    = 5;
	info.fb.var.blue.length     = 5;
	info.fb.var.transp.length   = 0;

	info.fb.fix.smem_len	    = info.fb.var.xres * info.fb.var.yres *
				      info.fb.var.bits_per_pixel / 8;

	/* Initialize video memory */
	ret = ns9xxxfb_map_video_memory( &info );
	if( ret ) {
		printk( KERN_WARNING DRIVER_NAME
			": Failed to allocate video RAM: %d\n", ret );
		ret = -ENOMEM;
		goto error_gpio;
	}

	ret = ns9xxxfb_init_registers( &info );
	if( ret < 0 ) {
		printk( KERN_WARNING DRIVER_NAME ": Failed to reserve gpios\n" );
		goto error_gpio;
	}

	ret = register_framebuffer( &info.fb );
	if( ret < 0 ) {
		printk( KERN_WARNING DRIVER_NAME
			": Failed to register framebuffer device: %d\n", ret );
		goto error;
	}

	printk( KERN_INFO "fb%d: %s frame buffer device\n",
		info.fb.node, info.fb.fix.id );

	return 0;

error:
	for( i = 0; i < ARRAY_SIZE( ns9xxxfb_gpios ); i++ )
		ns9xxx_release_gpio( ns9xxxfb_gpios[i].gpio );
error_gpio:
	release_mem_region(LCD_MODULE_BASE_VA, 0x400);
error_region:
	return ret;
}

static void __exit ns9xxxfb_cleanup( void )
{
	int i;

	ns9xxxfb_lcd_power_disable();
	unregister_framebuffer(&info.fb);

	/* release GPIO but let them configured for LCD */
	for( i = 0; i < ARRAY_SIZE( ns9xxxfb_gpios ); i++ )
		ns9xxx_release_gpio( ns9xxxfb_gpios[i].gpio );

	release_mem_region(LCD_MODULE_BASE_VA, 0x400);
}

module_init( ns9xxxfb_init);
module_exit( ns9xxxfb_cleanup);

MODULE_AUTHOR( "Digi International Inc" );
MODULE_DESCRIPTION( "Framebuffer driver for the NetSilicon NS9360/NS9750" );
MODULE_LICENSE( "GPL" );
