/*
 *  board/ccw9c/spi_ver.h
 *
 *  Copyright (C) 2006 by Digi International Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version2  as published by
 *  the Free Software Foundation.
*/
/*
 *  !Revision:   $Revision: 1.1 $
 *  !Author:     Markus Pietrek
 *  !Descr:      Prints the SPI version and it's variant to console.
 *               The SPI bootloader places two pointer/32bit length to the
 *               version and variant string in the Tx Descriptor RAM.
 *               This way we don't need to determine the address of the SPI
 *               bootloader and the strings in it ourself.
 *               These 4 dwords are secured by a 32bit 2-complement checksum
 *               as only 32bit accesses are allowed to the Tx RAM.
*/

#include <common.h>

#ifdef CONFIG_HAVE_SPI_LOADER

#include <ns9750_eth.h>

#define MAX_SIZE 64             /* this is save within the Tx RAM of 1KiB */

#define barrier() __asm__ __volatile__("": : :"memory")

void ns9xxx_print_spi_version( void )
{
#pragma pack(push)
        struct spi {
                struct {
                        const char* pcStr;
                        uint32_t    uiLen;
                } xVersion, xVariant;
                uint32_t uiChecksum;
        };
#pragma pack(pop)

        const struct spi* pSPI = (const struct spi*) get_eth_reg_addr( NS9750_ETH_TXBD );
        const uint32_t* puiSPI = NULL;

        uint32_t uiEGCR       = 0;
        uint32_t uiChecksum   = 0;
        char     bValid       = 0;

        /* enable TX RAM. If not done, we receive a data abort */
        uiEGCR = *get_eth_reg_addr( NS9750_ETH_EGCR1 );
        *get_eth_reg_addr( NS9750_ETH_EGCR1 ) |= NS9750_ETH_EGCR1_ETX;

        barrier();
        
        /* validate it. We must do 32bit accesses. */
        for( uiChecksum = 0, puiSPI = (const uint32_t*) pSPI;
             puiSPI < (const uint32_t*) (pSPI + 1);
             puiSPI++ )
                uiChecksum += *puiSPI;
        
        bValid = ( 0xffffffff == uiChecksum );

        if( bValid ) {
                /* strings must be in RAM, SDRAM is limited to 512 MB */
                bValid &= (((uint32_t) pSPI->xVersion.pcStr & 0x1fffffff ) != 0 );
                bValid &= (((uint32_t) pSPI->xVariant.pcStr & 0x1fffffff ) != 0 );
                bValid &= ( pSPI->xVersion.uiLen <= MAX_SIZE );
                bValid &= ( pSPI->xVariant.uiLen <= MAX_SIZE );
        }

        if( bValid )
                printf( "SPI ID:%.*s, %.*s\n",
                        pSPI->xVersion.uiLen, pSPI->xVersion.pcStr,
                        pSPI->xVariant.uiLen, pSPI->xVariant.pcStr );
        else
                printf( "SPI ID:not available or debug download\n" );

        barrier();
        
        /* restore EGCR */
        *get_eth_reg_addr( NS9750_ETH_EGCR1 ) = uiEGCR;
}

/* we have SPI bootloader */
#endif  /* CONFIG_HAVE_SPI_LOADER */
